Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.
Require Import List.
Require Import Setoid.
Require Import Morphisms.

Require Import LibTactics.
Require Import DomType.
Require Import DomBasic.
Require Import DomSet.
Require Import DomMap.

Require Import Syntax.

Require Import DomAbsValue.
Require Import DomAcc.
Require Import DomAbsMem.
Require Import DomPidxSen.
Require Import DomPidxInsen.

Require Import SemAbsBlock.
Require Import SemAccSen.
Require Import SemAccInsen.

Require Import SemSenLocal.
Require Import SemInsenLocal.

Require Import TrsSen.

Require Import CorAccBasic.
Require Import CorAccExpr.

Set Implicit Arguments.

Section Accs.
  Variables (prog:progT) (acc:accT) (state:stateT).

  Definition tstate' := tstate prog state.

  Lemma tacc_correct_acclocs_libcall_ret_to_some_value :
    forall fid mem mem' v retlv
           (Hlv: acc_correct_acclocs_lv acc fid retlv mem)
           (Hret: let alocs := (acclocs acc) [fid ]! in
                  match sem_expr_lv fid retlv mem with
                    | Elt locs => SetInterface.subset locs alocs
                    | Top => false
                  end
           ),
      memory_update
        acc
        (sem_expr_lv
           fid retlv
           (mrestrict_join (set_to_powerset (coerce (acclocs acc) [fid]))
                           mem
                           mem'))
        v
        (mrestrict_join (set_to_powerset (coerce (acclocs acc) [fid])) mem mem')
      ≡
      mrestrict_join (set_to_powerset (coerce (acclocs acc) [fid]))
      (memory_update acc (sem_expr_lv fid retlv mem)
                     v mem) mem'.
  Proof.
    intros.
    case: retlv Hlv Hret => [x|e|e idx|e pos] Hlv Hret.

    (* lv_id *)
    { simpl.
      rewrite/acc_correct_acclocs_var in Hlv.
      apply/eq_sym/memory_update_mrestrict_join_commutative.
      rewrite/sem_expr_lv in Hret.
      apply SetFacts.subset_iff in Hret.
      apply SetFacts.mem_iff.
      apply Hret.
      by apply SetFacts.singleton_iff.
    }

    (* lv_deref *)
    { simpl.
      pose (tacc_correct_acclocs_expr acc e fid mem Hlv) as Htexpr.
      case: mem' => [mem'|]; last first.
      { rewrite Top_top.
        rewrite [X in eq _ X]mem_mrestrict_join_top_rhs.
        rewrite [X in memory_update _ _ _ X]mem_mrestrict_join_top_rhs.
        exact:eq_refl.
      }
      { set llocs := (locs_of _).
        set rlocs := (locs_of _).
        assert (eq llocs rlocs) as Heqlocs.
        { apply locs_of_mor.
          apply tacc_correct_acclocs_expr; done.
        }
        rewrite Heqlocs.
        set larrv := (arrv_of _).
        set rarrv := (arrv_of _).
        assert (eq larrv rarrv) as Heqarrv.
        { apply arrv_of_mor.
          apply tacc_correct_acclocs_expr; done.
        }
        rewrite Heqarrv.
        clear llocs Heqlocs larrv Heqarrv.
        case Heqjlocs: (rlocs ⊔ locs_from_arrvT rarrv) => [locs|].
        { rewrite -> memory_update_mrestrict_join_commutative_subset;
          first by exact:eq_refl.
          apply/set_to_powerset_le_mor.
          simpl in Hret.
          rewrite Heqjlocs in Hret.
          apply SetFacts.subset_iff; done.
        }
        { rewrite/memory_update//; clear -mem; case: mem => [mem|] //. }
      }
    }

    (* lv_array *)
    { simpl.
      inversion Hlv as [Heidx].
      move/andP in Heidx; inversion Heidx as [He Hidx].
      pose (tacc_correct_acclocs_expr acc e fid mem He) as Hte.
      pose (tacc_correct_acclocs_expr acc idx fid mem Hidx) as Htidx.
      case: mem' => [mem'|]; last first.
      { rewrite Top_top.
        rewrite [X in eq _ X]mem_mrestrict_join_top_rhs.
        rewrite [X in memory_update _ _ _ X]mem_mrestrict_join_top_rhs.
        exact:eq_refl.
      }
      { set larrv := (arrv_of _).
        set rarrv := (arrv_of _).
        assert (eq larrv rarrv) as Heqarrv.
        { apply arrv_of_mor.
          apply tacc_correct_acclocs_expr; done.
        }
        rewrite Heqarrv.
        case Heqlocs: (locs_from_arrvT rarrv) => [locs|].
        { rewrite -> memory_update_mrestrict_join_commutative_subset;
          first by exact:eq_refl.
          apply set_to_powerset_le_mor.
          simpl in Hret.
          rewrite Heqlocs in Hret.
          apply SetFacts.subset_iff; done.
        }
        { rewrite/memory_update//; clear -mem; case: mem => [mem|] //. }
      }
    }

    (* lv_struct *)
    { simpl.
      simpl in Hlv.
      pose (tacc_correct_acclocs_lv _ _ _ _ Hlv) as Hsexpr.
      case: mem' => [mem'|]; last first.
      { rewrite Top_top.
        rewrite [X in eq _ X]mem_mrestrict_join_top_rhs.
        rewrite [X in memory_update _ _ _ X]mem_mrestrict_join_top_rhs.
        exact:eq_refl.
      }
      { simpl in Hret; move: Hret.
        set rlocs := sem_expr_lv _ _ _.
        set llocs := sem_expr_lv _ _ _.
        assert (eq llocs rlocs) as Hlocs.
        { rewrite/llocs/rlocs.
          by apply Hsexpr.
        }
        move: llocs rlocs Hlocs=> [llocs|] [rlocs|] Hlocs //=.
        move=> Hret.
        rewrite/eq/eq_/withtop_setoid/withtop_eq in Hlocs.
        rewrite Hlocs.
        apply/eq_sym/memory_update_mrestrict_join_commutative_subset.
        apply/set_to_powerset_le_mor.
        rewrite/le/le_/set_poset.
        by apply SetFacts.subset_iff.
      }
    }
  Qed.

  Lemma tacc_correct_acclocs_libcall_rand :
    forall (fid:fidT) a
           (mem:memT) ol le mem'
           (Hret : match ol with
                     | Some lv =>
                       let alocs := (acclocs acc) [fid ]! in
                       match sem_expr_lv fid lv mem with
                         | Elt locs => SetInterface.subset locs alocs
                         | Top => false
                       end
                     | None => true
                   end)
           (Hlv : match ol with
                    | Some lv => acc_correct_acclocs_lv acc fid lv mem
                    | None => true
                  end),
      libcall acc fid a ol rand le
              (mrestrict_join
                 (E:=valueT) (set_to_powerset (acclocs acc) [fid ]!) mem mem')
      ⊑
      mrestrict_join (E:=valueT) (set_to_powerset (acclocs acc) [fid ]!)
      (libcall acc fid a ol rand le mem) mem'.
  Proof.
    intros.
    apply/DomType.eq_le.
    case: ol Hlv Hret => [retlv|] Hlv Hret; last by apply/eq_refl.
    eapply tacc_correct_acclocs_libcall_ret_to_some_value; eauto.
  Qed.

  Lemma tacc_correct_acclocs_libcall_strlen :
    forall (fid:fidT) a
           (mem:memT) ol le mem'
           (Hret : match ol with
                     | Some lv =>
                       let alocs := (acclocs acc) [fid ]! in
                       match sem_expr_lv fid lv mem with
                         | Elt locs => SetInterface.subset locs alocs
                         | Top => false
                       end
                     | None => true
                   end)
           (Hlv : match ol with
                    | Some lv => acc_correct_acclocs_lv acc fid lv mem
                    | None => true
                  end),
      libcall acc fid a ol strlen le
              (mrestrict_join
                 (E:=valueT) (set_to_powerset (acclocs acc) [fid ]!) mem mem')
      ⊑
      mrestrict_join (E:=valueT) (set_to_powerset (acclocs acc) [fid ]!)
      (libcall acc fid a ol strlen le mem) mem'.
  Proof.
    intros.
    apply/DomType.eq_le.
    case: ol Hlv Hret => [retlv|] Hlv Hret; last by apply/eq_refl.
    eapply tacc_correct_acclocs_libcall_ret_to_some_value; eauto.
  Qed.

  Lemma tacc_correct_acclocs_libcall_strcpy :
    forall (fid:fidT) a
           (mem:memT) ol le mem'
           (Hret : let dstLocs := sem_expr fid (hd (expr_const lit_bot)
                                                   le) mem in
                   let dst := locs_of dstLocs in
                   let srcLocs := sem_expr fid (hd (expr_const lit_bot)
                                                   (tl le)) mem in
                   let src := locs_of srcLocs in
                   let srcValue := memory_lookup src mem in
                   let newMem := memory_update acc dst srcValue mem in
                   match ol with
                     | Some lv =>
                       let alocs := (acclocs acc) [fid ]! in
                       match sem_expr_lv fid lv newMem with
                         | Elt locs => SetInterface.subset locs alocs
                         | Top => false
                       end
                     | None => true
                   end)
           (Hlv : let dstLocs := sem_expr fid (hd (expr_const lit_bot)
                                                  le) mem in
                  let dst := locs_of dstLocs in
                  let srcLocs := sem_expr fid (hd (expr_const lit_bot)
                                                  (tl le)) mem in
                  let src := locs_of srcLocs in
                  let srcValue := memory_lookup src mem in
                  let newMem := memory_update acc dst srcValue mem in
                  match ol with
                    | Some lv => acc_correct_acclocs_lv acc fid lv newMem
                    | None => true
                  end)
           (Hexpr : acc_correct_acclocs_exprs acc fid le mem)
           (Hargs : acc_correct_acclocs_libcall_args acc fid le mem),
      libcall acc fid a ol strcpy le
              (mrestrict_join
                 (E:=valueT) (set_to_powerset (acclocs acc) [fid ]!) mem mem')
      ⊑
      mrestrict_join (E:=valueT) (set_to_powerset (acclocs acc) [fid ]!)
      (libcall acc fid a ol strcpy le mem) mem'.
  Proof.
    intros.
    apply/DomType.eq_le.
    case: mem' => [mem'|]; last first.

    { simpl.
      case: ol Hret Hlv => [retlv|] Hret Hlv.
      { rewrite [X in eq _ X]mem_mrestrict_join_top_rhs.
        remember (
            memory_update
              acc
              (locs_of
                 (sem_expr fid (hd (expr_const lit_bot) le)
                           (mrestrict_join
                              (set_to_powerset (coerce (acclocs acc) [fid]))
                              mem (Top Map! [locT, valueT]))))
              (memory_lookup
                 (locs_of
                    (sem_expr fid (hd (expr_const lit_bot) (tl le))
                              (mrestrict_join
                                 (set_to_powerset (coerce (acclocs acc) [fid]))
                                 mem
                                 (Top Map! [locT, valueT]))))
                 (mrestrict_join
                    (set_to_powerset (coerce (acclocs acc) [fid])) mem
                    (Top Map! [locT, valueT])))
              (mrestrict_join (set_to_powerset (coerce (acclocs acc) [fid])) mem
                              (Top Map! [locT, valueT]))) as mem_top.
        assert (eq mem_top top) as Htop.
        { subst.
          rewrite [X in memory_update _ _ _ X]mem_mrestrict_join_top_rhs.
          reflexivity.
        }
        rewrite Htop.
        reflexivity.
      }
      { rewrite [X in eq _ X]mem_mrestrict_join_top_rhs.
        rewrite [X in memory_update _ _ _ X]mem_mrestrict_join_top_rhs.
        reflexivity.
      }
    }

    { simpl.
      case: ol Hret Hlv => [retlv|] Hret Hlv.
      { set llocs := (locs_of _).
        set ltlocs := (locs_of _).
        set rlocs := (locs_of _).
        set rtlocs := (locs_of _).
        have Hrlocs: rlocs = (locs_of (sem_expr 
                                         fid (hd (expr_const lit_bot) le) mem))
          by done.
        have Hrtlocs:
          rtlocs = locs_of (sem_expr fid (hd (expr_const lit_bot) (tl le)) mem)
          by done.
        assert (eq llocs rlocs) as Hlocs.
        { apply locs_of_mor.
          clear -Hexpr.
          case: le Hexpr => [|hd tl] Hexpr //=.
          simpl in Hexpr; move/andP in Hexpr; inversion Hexpr as [Hhd Htl];
          clear Hexpr Htl.
          move: (tacc_correct_acclocs_expr _ _ _ _ Hhd); done.
        }
        assert (eq ltlocs rtlocs) as Htlocs.
        { apply locs_of_mor; clear -Hexpr.
          case: le Hexpr => [|hd [|thd ttl]] Hexpr //=.
          simpl in Hexpr; move/andP in Hexpr; inversion Hexpr as [Hhd Htl];
          clear Hexpr Hhd.
          move/andP in Htl; inversion Htl as [Hthd Httl];
          clear Htl Httl.
          move: (tacc_correct_acclocs_expr _ _ _ _ Hthd); done.
        }
        rewrite Htlocs; clear Htlocs ltlocs.
        set llv := (sem_expr_lv _ _ _).
        set rlv := (sem_expr_lv _ _ _).
        have Hrlv: rlv = (sem_expr_lv
                            fid retlv
                            (memory_update
                               acc rlocs
                               (memory_lookup rtlocs mem) mem)) by done.
        assert (eq llv rlv) as Heqlv.

        { move: (tacc_correct_acclocs_lv _ _ _ _ Hlv).
          rewrite/SemAccSen.acc_correct_acclocs_lv; move=> Htlv.
          pose (Htlv mem') as Htlvmem.
          eapply eq_trans; last by exact Htlvmem.
          apply sem_expr_lv_mor; try reflexivity.
          rewrite Hlocs.
          rewrite <- Hrlocs.
          rewrite <- Hrtlocs.
          case: rlocs Hrlocs Hlocs rlv Hrlv => [rlocs|] Hrlocs Hlocs rlv Hrlv.
          { rewrite memory_update_mrestrict_join_commutative_subset.
            { apply memory_update_mor; try reflexivity.
              clear -mem Hargs.
              eapply memory_lookup_mrestrict_join_valid_2; eauto.
            }
            { clear -Hargs Hrlocs.
              case: le Hargs Hrlocs => [|hd tl] Hargs Hrlocs //=.
              { inversion Hrlocs; subst; done. }
              { simpl in *; move/andP in Hargs;
                inversion Hargs as [Hhd Htl]; clear Hargs Htl.
                rewrite <- Hrlocs in Hhd.
                apply set_to_powerset_le_mor.
                apply SetFacts.subset_iff; done.
              }
            }
          }
          { rewrite/memory_update.
            clear -mem; case: (mrestrict_join _ _ _); case: mem => //.
          }
        }
        rewrite Heqlv; clear Heqlv llv.

        case: rlv Hrlv => [rlv|] Hrlv.
        { rewrite memory_update_mrestrict_join_commutative_subset.
          { apply memory_update_mor; try reflexivity.
            { apply locs_to_valueT_mor; done. }
            { rewrite Hlocs.
              case: rlocs Hrlocs Hlocs rlv Hrlv
              => [rlocs|] Hrlocs Hlocs rlv Hrlv.
              { rewrite memory_update_mrestrict_join_commutative_subset.
                { apply memory_update_mor; try reflexivity.
                  clear -mem Hargs.
                  eapply memory_lookup_mrestrict_join_valid_2; eauto.
                }
                { clear -Hargs Hrlocs.
                  case: le Hargs Hrlocs => [|hd tl] Hargs Hrlocs //=.
                  { inversion Hrlocs; subst; done. }
                  { simpl in *; move/andP in Hargs;
                    inversion Hargs as [Hhd Htl]; clear Hargs Htl.
                    rewrite <- Hrlocs in Hhd.
                    apply set_to_powerset_le_mor.
                    apply SetFacts.subset_iff; done.
                  }
                }
              }
              { rewrite/memory_update.
                clear -mem; case: (mrestrict_join _ _ _); case: mem => //.
              }
            }
          }
          { simpl in Hret.
            rewrite <- Hrlocs in Hret.
            rewrite <- Hrtlocs in Hret.
            rewrite <- Hrlv in Hret.
            apply set_to_powerset_le_mor.
            apply SetFacts.subset_iff; done.
          }
        }

        { rewrite{1}/memory_update.
          case: (memory_update _ _ _ _);
          rewrite{1}/memory_update;
          case: (memory_update _ _ _ _); intros;
          rewrite -> mem_mrestrict_join_top_lhs; reflexivity.
        }
      }

      { set llocs := (locs_of _).
        set ltlocs := (locs_of _).
        set rlocs := (locs_of _).
        set rtlocs := (locs_of _).
        have Hrlocs:
          rlocs = (locs_of (sem_expr fid (hd (expr_const lit_bot) le) mem))
          by done.
        have Hrtlocs:
          rtlocs = (locs_of (sem_expr fid (hd (expr_const lit_bot)
                                              (tl le)) mem))
          by done.
        assert (eq llocs rlocs) as Hlocs.
        { apply locs_of_mor.
          clear -Hexpr.
          case: le Hexpr => [|hd tl] Hexpr //=.
          simpl in Hexpr; move/andP in Hexpr; inversion Hexpr as [Hhd Htl];
          clear Hexpr Htl.
          move: (tacc_correct_acclocs_expr _ _ _ _ Hhd); done.
        }
        rewrite Hlocs; clear Hlocs llocs.

        assert (eq ltlocs rtlocs) as Htlocs.
        { apply locs_of_mor; clear -Hexpr.
          case: le Hexpr => [|hd [|thd ttl]] Hexpr //=.
          simpl in Hexpr; move/andP in Hexpr; inversion Hexpr as [Hhd Htl];
          clear Hexpr Hhd.
          move/andP in Htl; inversion Htl as [Hthd Httl];
          clear Htl Httl.
          move: (tacc_correct_acclocs_expr _ _ _ _ Hthd); done.
        }
        rewrite Htlocs; clear Htlocs ltlocs.

        case: rlocs Hrlocs => [rlocs|] Hrlocs.
        { rewrite memory_update_mrestrict_join_commutative_subset.
          { apply memory_update_mor; try reflexivity.
            eapply memory_lookup_mrestrict_join_valid_2; eauto.
          }
          { clear -Hargs Hrlocs.
            case: le Hargs Hrlocs => [|hd tl] Hargs Hrlocs //=.
            { inversion Hrlocs; subst; done. }
            { simpl in *; move/andP in Hargs;
              inversion Hargs as [Hhd Htl]; clear Hargs Htl.
              rewrite <- Hrlocs in Hhd.
              apply set_to_powerset_le_mor.
              apply SetFacts.subset_iff; done.
            }
          }
        }
        { rewrite/memory_update.
          clear -mem; case: (mrestrict_join _ _ _); case: mem => //.
        }
      }
    }
  Qed.

  Lemma tacc_correct_acclocs_libcall_scanf :
    forall (fid:fidT) a
           (mem:memT) ol le mem'
           (Hexpr : acc_correct_acclocs_exprs acc fid le mem)
           (Hargs : acc_correct_acclocs_libcall_args acc fid le mem),
      libcall acc fid a ol scanf le
              (mrestrict_join
                 (E:=valueT) (set_to_powerset (acclocs acc) [fid ]!) mem mem')
      ⊑
      mrestrict_join (E:=valueT) (set_to_powerset (acclocs acc) [fid ]!)
      (libcall acc fid a ol scanf le mem) mem'.
  Proof.
    intros.
    apply/DomType.eq_le.
    case: mem' => [mem'|]; last first.

    { rewrite [X in eq _ X]mem_mrestrict_join_top_rhs.
      simpl.
      elim: le Hexpr Hargs => [|hd tl IH] Hexpr Hargs.
      { simpl; rewrite mem_mrestrict_join_top_rhs; reflexivity. }
      { simpl in *.
        move/andP in Hexpr; inversion Hexpr as [Hehd Hetl];
        clear Hexpr Hehd.
        move/andP in Hargs; inversion Hargs as [Hahd Hatl];
        clear Hargs Hahd.
        move: (IH Hetl Hatl).
        set folder := (fold_right _ _ _); move=> IHapp.
        case: folder IHapp => [folder|] IHapp //.
      }
    }

    { simpl.
      elim: le Hexpr Hargs => [|lhd ltl IH] Hexpr Hargs; first by reflexivity.
      simpl.
      move/andP in Hexpr; inversion Hexpr as [Hehd Hetl]; clear Hexpr.
      move/andP in Hargs; inversion Hargs as [Hahd Hatl]; clear Hargs.
      set llocs := (locs_of _).
      set rlocs := (locs_of _).
      have Hrlocs: rlocs = (locs_of (sem_expr fid lhd mem)) by done.
      assert (eq llocs rlocs) as Heqlocs.
      { apply/locs_of_mor.
        apply tacc_correct_acclocs_expr; done.
      }
      rewrite Heqlocs; clear Heqlocs llocs.
      case: rlocs Hrlocs => [rlocs'|] Hrlocs.
      { rewrite memory_update_mrestrict_join_commutative_subset.
        { apply memory_update_mor; try reflexivity.
          apply (IH Hetl Hatl); done.
        }
        { rewrite <- Hrlocs in Hahd; simpl in Hahd.
          apply set_to_powerset_le_mor.
          apply SetFacts.subset_iff; done.
        }
      }
      { rewrite{1}/memory_update; case: (fold_right _ _ _);
        rewrite{1}/memory_update; case: (fold_right _ _ _); intros;
        rewrite mem_mrestrict_join_top_lhs; reflexivity.
      }
    }
  Qed.

  Lemma tacc_correct_acclocs_libcall_errno_location :
    forall (fid:fidT) a
           (mem:memT) ol le mem'
           (Hret : match ol with
                     | Some lv =>
                       let alocs := (acclocs acc) [fid ]! in
                       match sem_expr_lv fid lv mem with
                         | Elt locs => SetInterface.subset locs alocs
                         | Top => false
                       end
                     | None => true
                   end)
           (Hlv : match ol with
                    | Some lv => acc_correct_acclocs_lv acc fid lv mem
                    | None => true
                  end),
      libcall acc fid a ol errno_location le
              (mrestrict_join
                 (E:=valueT) (set_to_powerset (acclocs acc) [fid ]!) mem mem')
      ⊑
      mrestrict_join (E:=valueT) (set_to_powerset (acclocs acc) [fid ]!)
      (libcall acc fid a ol errno_location le mem) mem'.
  Proof.
    intros.
    apply/DomType.eq_le.
    case: ol Hlv Hret => [retlv|] Hlv Hret; last by apply/eq_refl.
    eapply tacc_correct_acclocs_libcall_ret_to_some_value; eauto.
  Qed.

  Lemma tacc_correct_acclocs_libcall_builtin_alloca :
    forall (fid:fidT) a
           (mem:memT) ol le mem'
           (Hret : match ol with
                     | Some lv =>
                       let alocs := (acclocs acc) [fid ]! in
                       match sem_expr_lv fid lv mem with
                         | Elt locs => SetInterface.subset locs alocs
                         | Top => false
                       end
                     | None => true
                   end)
           (Hlv : match ol with
                    | Some lv => acc_correct_acclocs_lv acc fid lv mem
                    | None => true
                  end)
           (Hexpr : acc_correct_acclocs_exprs acc fid le mem),
      libcall acc fid a ol builtin_alloca le
              (mrestrict_join
                 (E:=valueT) (set_to_powerset (acclocs acc) [fid ]!) mem mem')
      ⊑
      mrestrict_join (E:=valueT) (set_to_powerset (acclocs acc) [fid ]!)
      (libcall acc fid a ol builtin_alloca le mem) mem'.
  Proof.
    intros.
    apply/DomType.eq_le.
    case: ol Hlv Hret => [retlv|] Hlv Hret; last by apply/eq_refl.
    case: mem' => [mem'|]; last first.

    { simpl; rewrite [X in eq _ X]mem_mrestrict_join_top_rhs.
      rewrite [X in memory_update _ _ _ X]mem_mrestrict_join_top_rhs.
      reflexivity.
    }

    { case: retlv Hlv Hret => [x|e|e idx|e pos] Hlv Hret.
      { simpl.
        set lval := (arrv_to_valueT _).
        set rval := (arrv_to_valueT _).

        assert (eq lval rval) as Heqval.
        { apply arrv_to_valueT_mor; dom_tac.
          rewrite/eq/eq_/map_setoid; move=> key.
          rewrite 2! map_mlookup_spec; rewrite 2! map_madd_spec.
          case: (eqb a key); last by reflexivity.
          rewrite/eq/eq_/pair_setoid/pair_eq/=; split; last by reflexivity.
          rewrite/eq/eq_/=; split; first by reflexivity.
          apply itv_of_mor.
          clear -le Hexpr.
          case: le Hexpr => [|lhd ltl] Hexpr; first by reflexivity.
          simpl in Hexpr; move/andP in Hexpr; inversion Hexpr as [Hhd Htl];
          clear Htl Hexpr.
          exact: (tacc_correct_acclocs_expr acc _ _ _ Hhd).
        }

        rewrite Heqval.
        apply eq_sym.
        apply memory_update_mrestrict_join_commutative.
        simpl in Hret; rewrite <- is_true_iff in Hret.
        rewrite <- (SetFacts.subset_iff (elt:=locT)) in Hret.
        rewrite/SetInterface.Subset in Hret.
        simpl; rewrite <- is_true_iff.
        rewrite <- (SetFacts.mem_iff (elt:=locT)).
        eapply Hret; eauto.
        rewrite (SetFacts.singleton_iff (elt:=locT)); done.
      }

      { simpl.
        set llocs := (locs_of _).
        set rlocs := (locs_of _).
        assert (eq llocs rlocs) as Heqlocs.
        { apply locs_of_mor.
          simpl in Hlv.
          move: (tacc_correct_acclocs_expr acc _ _ _ Hlv).
          rewrite/SemAccSen.acc_correct_acclocs_expr; move=> Htexpr.
          exact:Htexpr.
        }
        rewrite Heqlocs; clear Heqlocs llocs.
        set llav := (locs_from_arrvT _).
        set rlav := (locs_from_arrvT _).
        assert (eq llav rlav) as Heqlav.
        { apply/locs_from_arrvT_mor/arrv_of_mor.
          simpl in Hlv.
          move: (tacc_correct_acclocs_expr acc _ _ _ Hlv).
          rewrite/SemAccSen.acc_correct_acclocs_expr; move=> Htexpr.
          exact:Htexpr.
        }
        rewrite Heqlav; clear Heqlav llav.
        set lv := (arrv_to_valueT _).
        set rv := (arrv_to_valueT _).
        assert (eq lv rv) as Heqv.
        { apply/arrv_to_valueT_mor; dom_tac.
          rewrite/eq/eq_/map_setoid; move=> key.
          rewrite 2! map_mlookup_spec; rewrite 2! map_madd_spec.
          case: (eqb a key); last by reflexivity.
          rewrite/eq/eq_/pair_setoid/pair_eq/=; split; last by reflexivity.
          rewrite/eq/eq_/=; split; first by reflexivity.
          apply itv_of_mor.
          clear -le Hexpr.
          case: le Hexpr => [|lhd ltl] Hexpr; first by reflexivity.
          simpl in Hexpr; move/andP in Hexpr; inversion Hexpr as [Hhd Htl];
          clear Htl Hexpr.
          exact: (tacc_correct_acclocs_expr acc _ _ _ Hhd).
        }
        rewrite Heqv; clear Heqv lv.

        case Heqjlocs: (rlocs ⊔ rlav) => [locs|].
        { rewrite -> memory_update_mrestrict_join_commutative_subset;
          first by exact:eq_refl.
          apply/set_to_powerset_le_mor.
          simpl in Hret.
          rewrite Heqjlocs in Hret.
          apply SetFacts.subset_iff; done.
        }
        { rewrite/memory_update//; clear -mem; case: mem => [mem|] //. }
      }

      { simpl; simpl in Hret; move: Hret.
        set llav := (locs_from_arrvT _).
        set rlav := (locs_from_arrvT _).
        assert (eq llav rlav) as Heqlav.
        { apply/locs_from_arrvT_mor/arrv_of_mor.
          simpl in Hlv; move/andP in Hlv; inversion Hlv as [He Hidx];
          clear Hlv.
          move: (tacc_correct_acclocs_expr acc _ _ _ He).
          rewrite/SemAccSen.acc_correct_acclocs_expr; move=> Htexpr.
          apply/eq_sym/Htexpr.
        }
        move=> Hret.
        rewrite Heqlav.
        set lv := (arrv_to_valueT _).
        set rv := (arrv_to_valueT _).
        assert (eq lv rv) as Heqv.
        { apply/arrv_to_valueT_mor; dom_tac.
          rewrite/eq/eq_/map_setoid; move=> key.
          rewrite 2! map_mlookup_spec; rewrite 2! map_madd_spec.
          case: (eqb a key); last by reflexivity.
          rewrite/eq/eq_/pair_setoid/pair_eq/=; split; last by reflexivity.
          rewrite/eq/eq_/=; split; first by reflexivity.
          apply itv_of_mor.
          clear -le Hexpr.
          case: le Hexpr => [|lhd ltl] Hexpr; first by reflexivity.
          simpl in Hexpr; move/andP in Hexpr; inversion Hexpr as [Hhd Htl];
          clear Htl Hexpr.
          exact: (tacc_correct_acclocs_expr acc _ _ _ Hhd).
        }
        rewrite Heqv; clear Heqv lv.
        
        case Heqjlocs: rlav => [locs|].
        { rewrite -> memory_update_mrestrict_join_commutative_subset;
          first by exact:eq_refl.
          apply/set_to_powerset_le_mor.
          simpl in Hret.
          case: llav Heqlav Hret => [llav|] Heqlav Hret.
          { rewrite Heqjlocs in Heqlav.
            rewrite/eq/eq_/withtop_setoid/withtop_eq in Heqlav.
            rewrite <- Heqlav.
            by apply SetFacts.subset_iff.
          }
          { rewrite Heqjlocs in Heqlav.
            contradict Heqlav.
          }
        }
        { rewrite/memory_update//; clear -mem; case: mem => [mem|] //. }
      }

      { move: Hret; simpl.
        simpl in Hlv.
        move: (tacc_correct_acclocs_lv acc _ _ _ Hlv)=> {Hlv} Hlv.
        set rlocs := (sem_expr_lv _ _ _).
        set llocs := (sem_expr_lv _ _ _).
        assert (eq llocs rlocs) as Hlocs.
        { by apply Hlv. }
        move=> Hret.
        set lval := (arrv_to_valueT _).
        set rval := (arrv_to_valueT _).
        assert (eq lval rval) as Heqval.
        { apply arrv_to_valueT_mor; dom_tac.
          rewrite/eq/eq_/map_setoid; move=> key.
          rewrite 2! map_mlookup_spec; rewrite 2! map_madd_spec.
          case: (eqb a key); last by reflexivity.
          rewrite/eq/eq_/pair_setoid/pair_eq/=; split; last by reflexivity.
          rewrite/eq/eq_/=; split; first by reflexivity.
          apply itv_of_mor.
          clear -le Hexpr.
          case: le Hexpr => [|lhd ltl] Hexpr; first by reflexivity.
          simpl in Hexpr; move/andP in Hexpr; inversion Hexpr as [Hhd Htl];
          clear Htl Hexpr.
          exact: (tacc_correct_acclocs_expr acc _ _ _ Hhd).
        }
        rewrite Heqval.
        move: llocs rlocs Hlocs Hret=> [llocs|] [rlocs|] Hlocs Hret //.
        rewrite/eq/eq_/withtop_setoid/withtop_eq in Hlocs.
        rewrite Hlocs.
        apply/eq_sym/memory_update_mrestrict_join_commutative_subset.
        apply/set_to_powerset_le_mor.
        by apply SetFacts.subset_iff.
      }
    }
  Qed.

  Lemma tacc_correct_acclocs_libcall_builtin_choose_expr :
    forall (fid:fidT) a
           (mem:memT) ol le mem'
           (Hret : match ol with
                     | Some lv =>
                       let alocs := (acclocs acc) [fid ]! in
                       match sem_expr_lv fid lv mem with
                         | Elt locs => SetInterface.subset locs alocs
                         | Top => false
                       end
                     | None => true
                   end)
           (Hlv : match ol with
                    | Some lv => acc_correct_acclocs_lv acc fid lv mem
                    | None => true
                  end)
           (Hexpr : acc_correct_acclocs_exprs acc fid le mem),
      libcall acc fid a ol builtin_choose_expr le
              (mrestrict_join
                 (E:=valueT) (set_to_powerset (acclocs acc) [fid ]!) mem mem')
      ⊑
      mrestrict_join (E:=valueT) (set_to_powerset (acclocs acc) [fid ]!)
      (libcall acc fid a ol builtin_choose_expr le mem) mem'.
  Proof.
    intros.
    apply/DomType.eq_le.
    case: ol Hlv Hret => [retlv|] Hlv Hret; last by apply/eq_refl.
    case: mem' => [mem'|]; last first.

    { simpl; rewrite [X in eq _ X]mem_mrestrict_join_top_rhs.
      case: (eqb _ _).
      { rewrite [X in memory_update _ _ _ X]mem_mrestrict_join_top_rhs.
        reflexivity.
      }
      { case: (leb _ _);
        rewrite [X in memory_update _ _ _ X]mem_mrestrict_join_top_rhs;
        reflexivity.
      }
    }

    { simpl.
      set leqb := (eqb _ _).
      set reqb := (eqb _ _).
      assert (eq leqb reqb) as Heqb.
      { apply eqb_mor; last by reflexivity.
        apply itv_of_mor.
        case: le Hexpr leqb reqb => [|lhd ltl] Hexpr leqb reqb;
        first by reflexivity.
        simpl in Hexpr; move/andP in Hexpr; inversion Hexpr as [Hhd Htl];
        clear Hexpr Htl.
        simpl.
        eapply tacc_correct_acclocs_expr; eauto.
      }
      rewrite Heqb; clear Heqb leqb.

      case: reqb.
      { move: Hret.
        set llv := (sem_expr_lv _ _ _).
        set rlv := (sem_expr_lv _ _ _).
        move=> Hret.
        assert (eq rlv llv) as Heqlv.
        { move: (tacc_correct_acclocs_lv acc _ _ _ Hlv).
          rewrite/SemAccSen.acc_correct_acclocs_lv; move=> Htlv.
          apply Htlv.
        }
        rewrite Heqlv; clear Heqlv rlv.
        set lv := (sem_expr _ _ _).
        set rv := (sem_expr _ _ _).
        assert (eq lv rv) as Heqv.
        { clear -le Hexpr lv rv.
          destruct le as [|lhd [|lthd [|ltthd ltttl]]]; try reflexivity.
          simpl in *.
          move/andP in Hexpr; inversion Hexpr as [Hhd Htl]; clear Hexpr Hhd.
          move/andP in Htl; inversion Htl as [Hthd Httl]; clear Htl Hthd.
          move/andP in Httl; inversion Httl as [Htthd Htttl]; clear Httl Htttl.
          eapply tacc_correct_acclocs_expr; eauto.
        }
        rewrite Heqv; clear Heqv lv.

        case Hllv: llv Hret => [llv'|] Hret.
        { apply eq_sym.
          apply memory_update_mrestrict_join_commutative_subset.
          simpl in Hret.
          apply set_to_powerset_le_mor.
          by apply SetFacts.subset_iff.
        }
        { rewrite/memory_update/=.
          clear -mem.
          case: (mrestrict_join _ _ _); case: mem; intros;
          rewrite mem_mrestrict_join_top_lhs;
          reflexivity.
        }
      }

      { set lleb := (leb _ _).
        set rleb := (leb _ _).
        assert (eq lleb rleb) as Hleb.
        { apply leb_eq_mor; first by reflexivity.
          apply itv_of_mor.
          case: le Hexpr lleb rleb => [|lhd ltl] Hexpr lleb rleb;
          first by reflexivity.
          simpl in Hexpr; move/andP in Hexpr; inversion Hexpr as [Hhd Htl];
          clear Hexpr Htl.
          simpl.
          eapply tacc_correct_acclocs_expr; eauto.
        }
        rewrite Hleb; clear Hleb lleb.

        case: rleb.
        { move: Hret.
          set llv := (sem_expr_lv _ _ _).
          set rlv := (sem_expr_lv _ _ _).
          move=> Hret.
          assert (eq rlv llv) as Heqlv.
          { move: (tacc_correct_acclocs_lv acc _ _ _ Hlv).
            rewrite/SemAccSen.acc_correct_acclocs_lv; move=> Htlv.
            apply Htlv.
          }
          rewrite Heqlv; clear Heqlv rlv.
          set ltv := (sem_expr _ _ _).
          set lttv := (sem_expr _ _ _).
          set rtv := (sem_expr _ _ _).
          set rttv := (sem_expr _ _ _).
          assert (eq ltv rtv) as Heqtv.
          { clear -le Hexpr ltv rtv.
            destruct le as [|lhd [|lthd lttl]]; try reflexivity.
            simpl in *.
            move/andP in Hexpr; inversion Hexpr as [Hhd Htl]; clear Hexpr Hhd.
            move/andP in Htl; inversion Htl as [Hthd Httl]; clear Htl Httl.
            eapply tacc_correct_acclocs_expr; eauto.
          }
          rewrite Heqtv; clear Heqtv ltv.
          assert (eq lttv rttv) as Heqttv.
          { clear -le Hexpr lttv rttv.
            destruct le as [|lhd [|lthd [|ltthd ltttl]]]; try reflexivity.
            simpl in *.
            move/andP in Hexpr; inversion Hexpr as [Hhd Htl]; clear Hexpr Hhd.
            move/andP in Htl; inversion Htl as [Hthd Httl]; clear Htl Hthd.
            move/andP in Httl; inversion Httl as [Htthd Htttl];
            clear Httl Htttl.
            eapply tacc_correct_acclocs_expr; eauto.
          }
          rewrite Heqttv; clear Heqttv lttv.
          case Hllv: llv Hret => [llv'|] Hret.
          { apply eq_sym.
            apply memory_update_mrestrict_join_commutative_subset.
            simpl in Hret.
            apply set_to_powerset_le_mor.
              by apply SetFacts.subset_iff.
          }
          { rewrite/memory_update/=.
                   clear -mem.
            case: (mrestrict_join _ _ _); case: mem; intros;
            rewrite mem_mrestrict_join_top_lhs;
            reflexivity.
          }
        }
        { move: Hret.
          set llv := (sem_expr_lv _ _ _).
          set rlv := (sem_expr_lv _ _ _).
          move=> Hret.
          assert (eq rlv llv) as Heqlv.
          { move: (tacc_correct_acclocs_lv acc _ _ _ Hlv).
            rewrite/SemAccSen.acc_correct_acclocs_lv; move=> Htlv.
            apply Htlv.
          }
          rewrite Heqlv; clear Heqlv rlv.
          set lv := (sem_expr _ _ _).
          set rv := (sem_expr _ _ _).
          assert (eq lv rv) as Heqv.
          { clear -le Hexpr lv rv.
            destruct le as [|lhd [|lthd lttl]]; try reflexivity.
            simpl in *.
            move/andP in Hexpr; inversion Hexpr as [Hhd Htl]; clear Hexpr Hhd.
            move/andP in Htl; inversion Htl as [Hthd Httl]; clear Htl Httl.
            eapply tacc_correct_acclocs_expr; eauto.
          }
          rewrite Heqv; clear Heqv lv.

          case Hllv: llv Hret => [llv'|] Hret.
          { apply eq_sym.
            apply memory_update_mrestrict_join_commutative_subset.
            simpl in Hret.
            apply set_to_powerset_le_mor.
              by apply SetFacts.subset_iff.
          }
          { rewrite/memory_update/=.
                   clear -mem.
            case: (mrestrict_join _ _ _); case: mem; intros;
            rewrite mem_mrestrict_join_top_lhs;
            reflexivity.
          }
        }
      }
    }
  Qed.

  Lemma tacc_correct_acclocs_libcall :
    forall fid ol lc le a mem
           (Hinstr: acc_correct_acclocs_instr
                      acc fid (instr_libcall ol lc le a) mem),
      SemAccSen.acc_correct_acclocs_instr
        acc fid (instr_libcall ol lc le a) mem.
  Proof.
    intros; destruct lc;
    unfold acc_correct_acclocs_instr in Hinstr;
    unfold SemAccSen.acc_correct_acclocs_instr;
    move/andP in Hinstr;
    move: Hinstr=> [Hlvexprret Hargs];
    move/andP in Hlvexprret;
    move: Hlvexprret=> [Hlvexpr Hret];
    move/andP in Hlvexpr;
    move: Hlvexpr=> [Hlv Hexpr];
    rewrite/SemAccSen.acc_correct_acclocs_instr/sem_instr => mem'.
    { apply/DomType.eq_le/eq_refl. }
    { apply/DomType.eq_le/eq_refl. }
    { eapply tacc_correct_acclocs_libcall_rand; eauto. }
    { eapply tacc_correct_acclocs_libcall_strlen; eauto. }
    { eapply tacc_correct_acclocs_libcall_strcpy; eauto. }
    { eapply tacc_correct_acclocs_libcall_scanf; eauto. }
    { eapply tacc_correct_acclocs_libcall_errno_location; eauto. }
    { eapply tacc_correct_acclocs_libcall_builtin_alloca; eauto. }
    { apply/DomType.eq_le/eq_refl. }
    { eapply tacc_correct_acclocs_libcall_builtin_choose_expr; eauto. }
    { apply/DomType.eq_le/eq_refl. }
  Qed.
End Accs.
