Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.
Require Import Setoid.
Require Import Morphisms.

Require Import DomType.
Require Import DomUtil.
Require Import DomBasic.

Class Proper2 {A B} (R : A -> B -> Prop) (a : A) (b : B) : Prop :=
  proper2_prf : R a b.

Notation "'Proper2' x" := (Proper2 (x)%signature) (at level 0).

Definition respectful2 {A A' B B' : Type}
  (RA : A -> A' -> Prop)
  (RB : B -> B' -> Prop) :
  ((A -> B) -> (A' -> B') -> Prop) :=
  Eval compute in @respectful_hetero A A' (fun _ => B) (fun _ => B') RA (fun _ _ => RB).

Definition existful2 {A A' B B' : Type}
  (RA : A -> A' -> Prop)
  (RB : B -> B' -> Prop) :
  ((A -> B) -> (A' -> B') -> Prop) :=
  fun HA HA' =>
    forall (a:A), exists (a':A'), RB (HA a) (HA' a') /\ RA a a'.

Definition existful2' {A A' : Type}
  (RA : A -> A' -> Prop) :
  ((A -> Prop) -> A' -> Prop) :=
  fun HA a' =>
    forall (a:A), HA a -> RA a a'.

Delimit Scope signature_scope with signature.

Module ProperNotations.
  Notation " R '===>' R' " := (@respectful2 _ _ _ _ (R%signature) (R'%signature))
    (right associativity, at level 55) : signature_scope.

  Notation " R '===|' " := (@existful2 _ _ _ _ (R%signature))
    (right associativity, at level 55) : signature_scope.

  Notation " R '===|'' " := (@existful2' _ _ (R%signature))
    (right associativity, at level 55) : signature_scope.
End ProperNotations.

Export ProperNotations.
