Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.

Require Import LibTactics.
Require Import DomBridge.
Require Import DomType.
Require Import DomBasic.
Require Import DomMap.

Require Import Syntax.

Require Import SemCon.
Require Import DomAbsValue.
Require Import DomAcc.
Require Import DomAbsMem.
Require Import DomPidxSen.

Require Import SemAbsBlock.
Require Import SemAccSen.
Require Import SemSen.

Require Import DomZ.
Require Import DomSet.
Require Import CorBasic.
Require Import CorAbs.
Require Import CorItv.
Require Import CorConExpr.

Set Implicit Arguments.

Obligation Tactic := Simpless.tactic.

Section Correctness.
Variable (prog:progT) (acc:accT).
  
  Lemma memory_update_value_le_proper : forall (v w:valueT) (l:locsT) (m:memT)
    (Hvw: v ⊑ w),
    memory_update acc l v m ⊑ memory_update acc l w m.
  Proof.
    intros. rewrite/le/le_/memT_poset/withtop_poset/withtop_le.
    remember (memory_update acc l v m) as vm; destruct vm as [vm|];
    remember (memory_update acc l w m) as wm; destruct wm as [wm|].
    { assert (Helt: Elt vm ⊑ Elt wm); last apply Helt.
      rewrite Heqvm. rewrite Heqwm.
      apply mem_le_spec. 
      rewrite <- Heqvm. intro Hevm. inversion Hevm.
      rewrite <- Heqwm. intro Hewm. inversion Hewm.
      intros. do 2 rewrite memory_update_spec.
      destruct m; last (apply DomType.eq_le; reflexivity).
      destruct l; last (apply DomType.eq_le; reflexivity).
      destruct (loc !\in s); last (apply DomType.eq_le; reflexivity).
      destruct (eqb (SetInterface.cardinal s) 1 && is_strong acc loc);
        first auto.
      apply join_lub; first apply join_lhs.
      apply DomType.le_trans with w; auto. apply join_rhs.
    }
    { auto. }
    { 
      symmetry in Heqvm. symmetry in Heqwm. move:Heqvm Heqwm.
      rewrite/memory_update.
      destruct m. destruct l.
      destruct (SetInterface.elements s).
      { intros. inversion Heqvm. }
      { intros. inversion Heqvm. }
      { intros. inversion Heqwm. }
      { intros. inversion Heqwm. }
    }
    { auto. }
  Qed.

  Lemma sound_libcall : forall ret lib args site m m' am
    (Hacc_prog: acc_correct_callfroms_prog prog acc)
    (Hacc_star: acc_correct_callfroms_star acc)
    (pidx:pidxT) (Hpidx_con: pidx_consistent prog (DomPidx.pifid pidx) pidx.2)
    (Hlcall : SemCon.libcall pidx ret lib args site m m')
    (Hmem : m \abs am)
    (Hwf_newmem : DomConMem.mem_wf pidx m'),
    (m' \abs sem_instr acc (DomPidx.pifid pidx)
        (instr_libcall ret lib args site) am).
  Proof. 
    destruct lib; intros.
    
  (* airac_print *)
    { rewrite/sem_instr/libcall.
      inversion Hlcall; subst. rewrite <- Heq. auto. }

  (* airac_dump *)
    { rewrite/sem_instr/libcall.
      inversion Hlcall; subst. rewrite <- Heq. auto. }

  (* rand *)
    { rewrite/sem_instr/libcall. destruct ret;
      last (inversion Hlcall; subst; rewrite <- Heq; auto).
      inversion Hlcall; subst. rewrite <- Heq.
      apply memory_update_sound with prog pidx; auto.
      apply sem_expr_lv_proper2_ with m; auto.
      constructor; auto.
      rewrite Heq; auto. }

  (* strlen *)
    { rewrite/sem_instr/libcall. destruct ret;
      last (inversion Hlcall; subst; rewrite <- Heq; auto).
      inversion Hlcall; subst. rewrite <- Heq.
      apply memory_update_sound with prog pidx; auto.
      apply sem_expr_lv_proper2_ with m; auto.
      constructor; auto.
      rewrite Heq; auto. }

  (* strcpy *)
    { rewrite/sem_instr/libcall. destruct ret.
      { inversion Hlcall; subst. rewrite <- Heq.
        assert ((DomConValue.vloc dst_loc) \abs
          (sem_expr (DomPidx.pifid pidx) (hd (expr_const lit_bot) args) am))
          as Hdst_abs.
        { apply sem_expr_proper2_ with m; auto. }
        assert ((DomConValue.vloc src_loc) \abs
          (sem_expr (DomPidx.pifid pidx) (hd (expr_const lit_bot)
                                             (tl args)) am)) as Hsrc_abs.
        { apply sem_expr_proper2_ with m; auto. }

        apply memory_update_sound with prog pidx; auto.
        { rewrite <-Hcopy. 
          apply memory_update_sound with prog pidx; auto.
          { rewrite/memory_lookup.
            remember (locs_of 
              (sem_expr (DomPidx.pifid pidx) (hd (expr_const lit_bot)
                                                 (tl args)) am)).
            destruct l0; last apply abs_option_value_top.
            rewrite/abs_relation/option_abs.
            rewrite SetInterface.fold_1.
            apply sound_memory_lookup_ with m src_loc; auto.
            apply existsb_exists.

            rewrite/abs_relation/value_abs in Hsrc_abs. 
            rewrite/abs_relation/loc_abs in Hsrc_abs.
            rewrite/DomConValue.vloc in Hsrc_abs.
            rewrite <- Heql0 in Hsrc_abs. rewrite/smem/set_setlike in Hsrc_abs.
            apply SetInterface.mem_2 in Hsrc_abs.
            apply SetFacts.elements_iff in Hsrc_abs.
            apply InA_alt in Hsrc_abs.
            destruct Hsrc_abs as [x [Heq_abs Hin]].
            exists x. split; auto.
            rewrite/_eq/order_OrderedType in Heq_abs. apply/eqbP. auto. }
          { rewrite Hcopy. auto. }
        }
        { apply sem_expr_lv_proper2_ with mem'; auto.
          rewrite <- Hcopy.
          apply memory_update_sound with prog pidx; auto.
          { rewrite/memory_lookup.
            remember (locs_of 
              (sem_expr (DomPidx.pifid pidx) (hd (expr_const lit_bot) (tl args)) am)).
            destruct l0; last apply abs_option_value_top.
            rewrite/abs_relation/option_abs.
            remember ((m) [src_loc ]!). destruct y; auto; try discriminate.
            rewrite SetInterface.fold_1.
            apply sound_memory_lookup_ with m src_loc; auto.
            rewrite <- Hsrc0. rewrite Heqy. auto.
            apply existsb_exists.

            rewrite/abs_relation/value_abs in Hsrc_abs. 
            rewrite/abs_relation/loc_abs in Hsrc_abs.
            rewrite/DomConValue.vloc in Hsrc_abs.
            rewrite <- Heql0 in Hsrc_abs. rewrite/smem/set_setlike in Hsrc_abs.
            apply SetInterface.mem_2 in Hsrc_abs.
            apply SetFacts.elements_iff in Hsrc_abs.
            apply InA_alt in Hsrc_abs. destruct Hsrc_abs as [x [Heq_abs Hin]].
            exists x. split; auto.
            rewrite/_eq/order_OrderedType in Heq_abs. apply/eqbP. auto. }
          { rewrite Hcopy. auto. }
        }
        { rewrite Heq. auto. }
      }
      { inversion Hlcall; subst. rewrite <- Hcopy.
        assert ((DomConValue.vloc dst_loc) \abs
          (sem_expr (DomPidx.pifid pidx) (hd (expr_const lit_bot) args) am))
          as Hdst_abs.
        { apply sem_expr_proper2_ with m; auto. }
        assert ((DomConValue.vloc src_loc) \abs
          (sem_expr (DomPidx.pifid pidx) (hd (expr_const lit_bot)
                                             (tl args)) am)) as Hsrc_abs.
        { apply sem_expr_proper2_ with m; auto. }
        apply memory_update_sound with prog pidx; auto.
        { rewrite/memory_lookup.
          remember (locs_of (sem_expr (DomPidx.pifid pidx)
                                      (hd (expr_const lit_bot) (tl args)) am)).
          destruct l; last apply abs_option_value_top.
          rewrite/abs_relation/option_abs.
          remember ((m) [src_loc ]!). destruct y; auto; try discriminate.
          rewrite SetInterface.fold_1.
          apply sound_memory_lookup_ with m src_loc; auto.
          rewrite <- Hsrc0. rewrite Heqy. auto.
          apply existsb_exists.

          rewrite/abs_relation/value_abs in Hsrc_abs. 
          rewrite/abs_relation/loc_abs in Hsrc_abs.
          rewrite/DomConValue.vloc in Hsrc_abs.
          rewrite <- Heql in Hsrc_abs. rewrite/smem/set_setlike in Hsrc_abs.
          apply SetInterface.mem_2 in Hsrc_abs.
          apply SetFacts.elements_iff in Hsrc_abs.
          apply InA_alt in Hsrc_abs. destruct Hsrc_abs as [x [Heq_abs Hin]].
          exists x. split; auto.
          rewrite/_eq/order_OrderedType in Heq_abs. apply/eqbP. auto. }
        { rewrite Hcopy. auto. }
      }
    }

  (* scanf *)
    { move:m m' Hwf_newmem Hlcall am Hmem. induction args; intros.
      { inversion Hlcall; subst. rewrite <- Heq. 
        rewrite/sem_instr/libcall/fold_left. auto. }
      { inversion Hlcall; subst. rewrite <- Heq. 
        rewrite/sem_instr/libcall.
        replace (List.map (λ e : exprT, locs_of (sem_expr (DomPidx.pifid pidx) e am))
                          (a :: args)) with
        (locs_of (sem_expr (DomPidx.pifid pidx) a am)
                 :: List.map (λ e : exprT, locs_of (sem_expr (DomPidx.pifid pidx) e am)) args); last by auto.
        replace (fold_right
                   (λ (locs : locsT) (m0 : memT), memory_update acc locs (top, bot, bot) m0)
                   am
                   (locs_of (sem_expr (DomPidx.pifid pidx) a am)
                            :: List.map (λ e : exprT, locs_of (sem_expr (DomPidx.pifid pidx) e am)) args)) with
        (memory_update acc (locs_of (sem_expr (DomPidx.pifid pidx) a am)) 
                       (top, bot, bot)
                       (fold_right
                          (λ (locs : locsT) (m0 : memT), memory_update acc locs (top, bot, bot) m0)
                          am     
                          (List.map (λ e : exprT, locs_of (sem_expr (DomPidx.pifid pidx) e am)) args))); last by auto.
        apply memory_update_sound with prog pidx; auto.
        { rewrite/sem_instr/libcall in IHargs.
          apply IHargs with m; auto. }
        { assert ((DomConValue.vloc x_loc)
                    \abs
                    (sem_expr (DomPidx.pifid pidx) a am)).
          { eapply sem_expr_proper2_; eauto. }
          apply H.
        }
        { constructor. }
        { rewrite Heq. auto. }
      }
    }

  (* errno_location *)
    { rewrite/sem_instr/libcall. destruct ret.
      { inversion Hlcall; subst. rewrite <- Heq.
        apply memory_update_sound with prog pidx; auto.
        { apply sem_expr_lv_proper2_ with m; auto. }
        { rewrite/abs_relation/option_abs.
          rewrite/abs_relation/value_abs/arrv_to_valueT/arrv_of/fst/snd.
          rewrite/mlookup/withtop_maplike.
          rewrite/mlookup/map_maplike.
          rewrite/DomConValue.vheap.
          rewrite MapFacts.add_o. rewrite is_compare_eqb.
          rewrite eqb_refl. rewrite/coerce.
          constructor. constructor.
          constructor; rewrite/fst; omega.
          constructor; rewrite/fst/snd; omega.
          constructor; rewrite/snd; omega. }
        { rewrite Heq. auto. }
      }
      { inversion Hlcall; subst. rewrite <- Heq. auto. }
    }

  (* builtin_alloca *)
    { rewrite/sem_instr/libcall. destruct ret.
      { inversion Hlcall; subst. rewrite <- Heq.
        apply memory_update_sound with prog pidx; auto.
        { apply sem_expr_lv_proper2_ with m; auto. }
        { rewrite/abs_relation/option_abs.
          rewrite/abs_relation/value_abs/arrv_to_valueT/arrv_of/fst/snd.
          rewrite/mlookup/withtop_maplike.
          rewrite/mlookup/map_maplike.
          rewrite/DomConValue.vheap.
          rewrite MapFacts.add_o. rewrite is_compare_eqb.
          rewrite eqb_refl. rewrite/coerce.
          constructor. constructor.
          constructor; rewrite/fst; omega.
          rewrite/fst/snd.
          assert ((DomConValue.vint sz) \abs
            (sem_expr (DomPidx.pifid pidx) (hd (expr_const lit_bot) args) am)).
          apply sem_expr_proper2_ with m; auto. apply H.
          constructor; rewrite/snd; omega. }
        { rewrite Heq. auto. }
      }
      { inversion Hlcall; subst. rewrite <- Heq. auto. }
    }

  (* realloc *)
    { rewrite/sem_instr/libcall. inversion Hlcall. }

  (* builtin_choose_expr *)
    { rewrite/sem_instr/libcall. destruct ret.
      { remember (eqb (itv_of (sem_expr (DomPidx.pifid pidx)
                                        (hd (expr_const lit_bot) args) am))
        (itv_const 0)). destruct b.
        { inversion Hlcall; subst.
          { assert ((DomConValue.vint cnd) \abs
            (sem_expr (DomPidx.pifid pidx) (hd (expr_const lit_bot) args) am)).
            apply sem_expr_proper2_ with m; auto.
            rewrite/abs_relation/value_abs in H.
            symmetry in Heqb. move/eqbP:Heqb=>Heqb.
            rewrite/DomConValue.vint in H.
            rewrite -> Heqb in H.
            inversion H. elim Hcnd. omega. }
          { rewrite <- Heq.
            apply memory_update_sound with prog pidx; auto.
            { apply sem_expr_lv_proper2_ with m; auto. }
            { apply sem_expr_proper2_ with m; auto. }
            { rewrite Heq; auto. }
          }
        }
        remember (DomType.leb (itv_const 0)
          (itv_of (sem_expr (DomPidx.pifid pidx)
                            (hd (expr_const lit_bot) args) am))).
        destruct b.
        { inversion Hlcall; subst.
          { assert (m' \abs
            memory_update acc (sem_expr_lv (DomPidx.pifid pidx) l am)
            (sem_expr (DomPidx.pifid pidx) (hd (expr_const lit_bot) (tl args)) am) am).
            rewrite <- Heq.
            apply memory_update_sound with prog pidx; auto.
            apply sem_expr_lv_proper2_ with m; auto.
            apply sem_expr_proper2_ with m; auto.
            rewrite Heq. auto.
            move:H. apply abs_relation_proper. reflexivity.
            apply memory_update_value_le_proper. apply join_lhs.
          }
          { assert (m' \abs
            memory_update acc (sem_expr_lv (DomPidx.pifid pidx) l am)
            (sem_expr (DomPidx.pifid pidx)
                      (hd (expr_const lit_bot) (tl (tl args))) am) am).
            rewrite <- Heq.
            apply memory_update_sound with prog pidx; auto.
            apply sem_expr_lv_proper2_ with m; auto.
            apply sem_expr_proper2_ with m; auto.
            rewrite Heq. auto.
            move:H. apply abs_relation_proper. reflexivity.
            apply memory_update_value_le_proper. apply join_rhs.
          }
        }
        { inversion Hlcall; subst.
          { rewrite<- Heq.
            apply memory_update_sound with prog pidx; auto.
            apply sem_expr_lv_proper2_ with m; auto.
            apply sem_expr_proper2_ with m; auto.
            rewrite Heq. auto. }
          { symmetry in Heqb0.
            move/lebP:Heqb0=>Heqb0. elim Heqb0.
            assert ((DomConValue.vint 0%Z) \abs 
              (sem_expr (DomPidx.pifid pidx)
                        (hd (expr_const lit_bot) args) am)).
            apply sem_expr_proper2_ with m; auto. 
            rewrite/abs_relation/value_abs in H.
            rewrite/DomConValue.vint in H.
            rewrite/le/le_/option_poset/option_le/itv_const/itv_of/fst.
            remember (sem_expr (DomPidx.pifid pidx)
                               (hd (expr_const lit_bot) args) am).
            destruct v0. destruct p. destruct i.
            rewrite/itv_of/fst/abs_relation/Z_itvT_abs in H.
            rewrite/le/le_/DomItv.itvT'_poset. destruct i.
            { auto. }
            { inversion H. apply H1. }
            { inversion H. apply H1. }
            { inversion H. split. apply H3. apply H6. }
            inversion H.
          }
        }
      }
      { inversion Hlcall; subst. rewrite <- Heq. auto. }
    }

  (* unknown *)
    { rewrite/sem_instr/libcall.
      inversion Hlcall; subst. rewrite<- Heq; auto. }
  Qed.
End Correctness.
