Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.

Require Import LibTactics.
Require Import DomType.
Require Import DomBasic.

Require Import Containers.OrderedType.

Section Orders.
  Context {A} `{setoid A, @poset A _, @order A _ _}.

  Global Program Instance order_StrictOrder : Containers.OrderedType.StrictOrder (@lt A _ _) eq.
  Next Obligation.
    rewrite /lt /neq => x y z [Hle Hneq] [Hle' Hneq'].
    split; first by apply: le_trans; eauto.
    contradict Hneq; apply: le_antisym; first by [].
    rewrite -> Hle'.
    exact/DomType.eq_le/eq_sym.
  Qed.
  Next Obligation.
    rewrite /lt in H2 => Hequiv.
      by case: H2 => Hle; apply.
  Qed.

  Context `{@setoidDec A _, @posetDec A _ _ _}.
  Global Program Instance order_OrderedType : Containers.OrderedType.OrderedType A := {
    _eq := eq;
    _lt := @lt A _ _;
    _cmp x y := if leb x y
                    then if leb y x
                         then Eq
                         else Lt
                    else Gt
  }.
  Next Obligation.
    case (lebP x y); case (lebP y x); rewrite /lt /neq; constructor.
    exact: le_antisym.
      by split; first by []; contradict n; apply DomType.eq_le; apply: eq_sym.
        by split; first by []; contradict n; apply DomType.eq_le; apply: eq_sym.
        case (le_total x y) => Htotal.
          by contradict n0.
            by contradict n.
  Qed.

  Lemma is_compare_eqb (x y : A) : (x == y) = (eqb x y).
  Proof.
    by case: (eq_dec x y) => Hxy; symmetry; apply/eqbP.
  Qed.
End Orders.

Section OrderedTypes.
  Context {A} `{OrderedType.OrderedType A}.

  Local Program Instance OrderedType_setoid : setoid A := {
    eq_ := OrderedType._eq
  }.

  Local Program Instance OrderedType_poset : @poset A _ := {
    le_ x y := (OrderedType._lt x y) \/ (OrderedType._eq x y)
  }.
  Next Obligation.
    by move=> x y Heq; right.
  Qed.
  Next Obligation.
    rewrite/eq/eq_/OrderedType_setoid.
    move=> /= x y [Hxy|Hxy] [Hyx|Hyx] //=; last by rewrite -> Hyx.
    case OrderedType.OT_StrictOrder => [lt_trans lt_asym].
    elim (lt_asym x x); last by [].
    by apply/lt_trans; eauto.
  Qed.
  Next Obligation.
    move=> //= x y z [] Hxy; last by rewrite -> Hxy.
    move=> [] Hyz; last by rewrite <- Hyz; left.
    left.
    case OrderedType.OT_StrictOrder => [le_trans _].
    by apply le_trans with y.
  Qed.

  Local Program Instance OrderedType_order : @order A _ _.
  Next Obligation.
    by case: (OrderedType._compare_spec x y); firstorder.
  Qed.

  Local Program Instance OrderedType_setoidDec : @setoidDec A _ := {
    eqb_ x y := OrderedType.is_compare x y Eq
  }.
  Next Obligation.
    case OrderedType.OT_StrictOrder => [le_trans le_asym].
    rewrite /OrderedType.is_compare.
    case Hcmp: (OrderedType.compare x y); constructor.
    exact/OrderedType.compare_2.
    exact/(le_asym x y)/OrderedType.compare_1.
    move=> Heq; apply/(le_asym y x); last by apply/eq_sym.
    exact/OrderedType.compare_3.
  Qed.

  Local Program Instance OrderedType_posetDec : @posetDec A _ _ _ := {
    leb_ x y := ~~ (OrderedType.is_compare x y Gt)
  }.
  Next Obligation.
    case OrderedType.OT_StrictOrder => [le_trans le_asym].
    rewrite /OrderedType.is_compare.
    case Hcmp: (OrderedType.compare x y); constructor.
    by right; apply/OrderedType.compare_2.
    by left; apply/OrderedType.compare_1.
    apply OrderedType.compare_3 in Hcmp.
    move=> [] Hlr.
    apply/(le_asym x x); last by [].
    by apply/le_trans; eauto.
    rewrite <- Hlr in Hcmp.
    by apply/(le_asym x x).
  Qed.
End OrderedTypes.
