Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.
Require Import Setoid.
Require Import ZArith.

Require Import DomType.
Require Import DomUtil.
Require Import DomZ.
Require Import Morphisms.

Inductive itvT' : Set :=
| itv'_top
| itv'_inf_z (u:Z)
| itv'_z_inf (l:Z)
| itv'_z_z (l u:Z) of (l ⊑ u).

Definition zero_le_one : 0%Z ⊑ 1%Z.
Proof. unfold le,le_,Z_poset. omega. Qed.

Definition itv'_btop := itv'_z_z 0 1 zero_le_one.

Program Instance itvT'_setoid : setoid itvT' := {
  eq_ := nosimpl (fun lhs y =>
    match lhs, y with
      | itv'_top, itv'_top => True
      | itv'_top, itv'_inf_z _ => False
      | itv'_top, itv'_z_inf _ => False
      | itv'_top, itv'_z_z _ _ _ => False

      | itv'_inf_z _, itv'_top => False
      | itv'_inf_z zlhs_ub, itv'_inf_z zy_ub => zlhs_ub ≡ zy_ub
      | itv'_inf_z _, itv'_z_inf _ => False
      | itv'_inf_z _, itv'_z_z _ _ _ => False

      | itv'_z_inf _, itv'_top => False
      | itv'_z_inf _, itv'_inf_z _ => False
      | itv'_z_inf zlhs_lb, itv'_z_inf zy_lb => zlhs_lb ≡ zy_lb
      | itv'_z_inf _, itv'_z_z _ _ _ => False

      | itv'_z_z _ _ _, itv'_top => False
      | itv'_z_z _ _ _, itv'_inf_z _ => False
      | itv'_z_z _ _ _, itv'_z_inf _ => False
      | itv'_z_z zlhs_lb zlhs_ub _, 
        itv'_z_z zy_lb zy_ub _ => 
        zlhs_lb ≡ zy_lb ∧ zlhs_ub ≡ zy_ub
    end)
}.
Next Obligation. by move=> x; case: x; firstorder. Qed.
Next Obligation. by move=> x y; case: x; case: y; firstorder. Qed.
Next Obligation.
  by move=> x y z; case: x; case: y; case: z;
    unfold eq in *; simpl in *; firstorder.
Qed.

Program Instance itvT'_poset : poset := {
  le_ := nosimpl (fun lhs y =>
    match lhs, y with
      | itv'_top, itv'_top => True
      | itv'_top, itv'_inf_z _ => False
      | itv'_top, itv'_z_inf _ => False
      | itv'_top, itv'_z_z _ _ _ => False

      | itv'_inf_z _, itv'_top => True
      | itv'_inf_z zlhs_ub, itv'_inf_z zy_ub => zlhs_ub ⊑ zy_ub
      | itv'_inf_z _, itv'_z_inf _ => False
      | itv'_inf_z _, itv'_z_z _ _ _ => False

      | itv'_z_inf _, itv'_top => True
      | itv'_z_inf _, itv'_inf_z _ => False
      | itv'_z_inf zlhs_lb, itv'_z_inf zy_lb => zy_lb ⊑ zlhs_lb
      | itv'_z_inf _, itv'_z_z _ _ _ => False

      | itv'_z_z _ _ _, itv'_top => True
      | itv'_z_z _ zlhs_ub _, itv'_inf_z zy_ub => zlhs_ub ⊑ zy_ub
      | itv'_z_z zlhs_lb _ _, itv'_z_inf zy_lb => zy_lb ⊑ zlhs_lb
      | itv'_z_z zlhs_lb zlhs_ub _, 
        itv'_z_z zy_lb zy_ub _ => 
        zy_lb ⊑ zlhs_lb ∧ zlhs_ub ⊑ zy_ub
    end)
}.
Next Obligation.
  rewrite/eq/le; move=> x y; case: x; case: y; simpl in *; firstorder.
  by rewrite/eq/le; simpl in *; firstorder.
  by rewrite/eq/le; simpl in *; firstorder.
Qed.
Next Obligation.
  rewrite/eq/le; move=> x y; case: x; case: y; simpl in *; firstorder.
  by rewrite/eq/le; simpl in *; firstorder.
  by rewrite/eq/le; simpl in *; firstorder.
  by rewrite/eq/le; simpl in *; firstorder.
  by rewrite/eq/le; simpl in *; firstorder.
Qed.
Next Obligation.
  by rewrite/eq/le; move=> x y z; case: x; case: y; case: z; simpl in *; firstorder.
Qed.

Program Instance itvT'_hasTop : hasTop := {
  top_ := itv'_top
}.
Next Obligation. by case: x. Qed.

Require Import Omega.
Program Instance itvT'_hasJoin : hasJoin := { 
  join_ := nosimpl (fun lhs y =>
    match lhs, y with
      | itv'_top, itv'_top => itv'_top
      | itv'_top, itv'_inf_z _ => itv'_top
      | itv'_top, itv'_z_inf _ => itv'_top
      | itv'_top, itv'_z_z _ _ _ => itv'_top

      | itv'_inf_z _, itv'_top => itv'_top
      | itv'_inf_z zlhs_ub, itv'_inf_z zy_ub => 
        itv'_inf_z (Z.max zlhs_ub zy_ub)
      | itv'_inf_z _, itv'_z_inf _ => itv'_top
      | itv'_inf_z zlhs_ub, itv'_z_z _ zy_ub _ => 
        itv'_inf_z (Z.max zlhs_ub zy_ub)

      | itv'_z_inf _, itv'_top => itv'_top
      | itv'_z_inf _, itv'_inf_z _ => itv'_top
      | itv'_z_inf zlhs_lb, itv'_z_inf zy_lb => 
        itv'_z_inf (Z.min zlhs_lb zy_lb)
      | itv'_z_inf zlhs_lb, itv'_z_z zy_lb _ _ => 
        itv'_z_inf (Z.min zlhs_lb zy_lb)

      | itv'_z_z _ _ _, itv'_top => itv'_top
      | itv'_z_z _ zlhs_ub _, itv'_inf_z zy_ub => 
        itv'_inf_z (Z.max zlhs_ub zy_ub)
      | itv'_z_z zlhs_lb _ _, itv'_z_inf zy_lb => 
        itv'_z_inf (Z.min zlhs_lb zy_lb)
      | itv'_z_z zlhs_lb zlhs_ub _, 
        itv'_z_z zy_lb zy_ub _ => 
        itv'_z_z (Z.min zlhs_lb zy_lb) (Z.max zlhs_ub zy_ub) _
    end)
}.
Next Obligation.
  eapply Z.le_trans; first by eapply Z.min_le_compat; eauto.
  by eauto using Z.le_trans, Z.le_min_l, Z.le_max_l.
Qed.
Next Obligation.
  rewrite/eq/le; case: x; case: y; simpl in *; firstorder using Z.le_max_l.
  by rewrite/eq/le; simpl in *; firstorder.
  by rewrite/eq/le; simpl in *; firstorder.
  by rewrite/eq/le; simpl in *; firstorder.
  by rewrite/eq/le; simpl in *; firstorder.
Qed.
Next Obligation.
  rewrite/eq/le; case: x; case: y; simpl in *; firstorder using Z.le_max_r.
  by rewrite/eq/le; simpl in *; firstorder.
  by rewrite/eq/le; simpl in *; firstorder.
  by rewrite/eq/le; simpl in *; firstorder.
  by rewrite/eq/le; simpl in *; firstorder.
Qed.
Next Obligation.
  rewrite/eq/le; simpl in *; case:ub Hx Hy; case:x; case:y; firstorder;
  rewrite/eq/le; simpl in *; auto using Z.max_lub, Z.min_glb.
Qed.

Program Instance itvT'_setoidDec : setoidDec := { 
  eqb_ := nosimpl (fun x y =>
    match x, y with
      | itv'_top, itv'_top => true
      | itv'_top, itv'_inf_z _ => false
      | itv'_top, itv'_z_inf _ => false
      | itv'_top, itv'_z_z _ _ _ => false

      | itv'_inf_z _, itv'_top => false
      | itv'_inf_z zx_ub, itv'_inf_z zy_ub => eqb zx_ub zy_ub
      | itv'_inf_z _, itv'_z_inf _ => false
      | itv'_inf_z _, itv'_z_z _ _ _ => false

      | itv'_z_inf _, itv'_top => false
      | itv'_z_inf _, itv'_inf_z _ => false
      | itv'_z_inf zx_lb, itv'_z_inf zy_lb => eqb zx_lb zy_lb
      | itv'_z_inf _, itv'_z_z _ _ _ => false

      | itv'_z_z _ _ _, itv'_top => false
      | itv'_z_z _ _ _, itv'_inf_z _ => false
      | itv'_z_z _ _ _, itv'_z_inf _ => false
      | itv'_z_z zx_lb zx_ub _, 
        itv'_z_z zy_lb zy_ub _ => 
        andb (eqb zx_lb zy_lb) (eqb zx_ub zy_ub)
    end)
}.
Next Obligation.
  case:x; case:y; try (by constructor);
    try (by move=> *; apply/eqbP).
  move => xl xu xlu yl yu ylu.
  apply/equivP; first by apply andP.
  by constructor; move=>[Hl Hr]; split; apply/eqbP.
Qed.

Program Instance itvT'_posetDec : posetDec := { 
  leb_ := nosimpl (fun x y =>
    match x, y with
      | itv'_top, itv'_top => true
      | itv'_top, itv'_inf_z _ => false
      | itv'_top, itv'_z_inf _ => false
      | itv'_top, itv'_z_z _ _ _ => false

      | itv'_inf_z _, itv'_top => true
      | itv'_inf_z zx_ub, itv'_inf_z zy_ub => leb zx_ub zy_ub
      | itv'_inf_z _, itv'_z_inf _ => false
      | itv'_inf_z _, itv'_z_z _ _ _ => false

      | itv'_z_inf _, itv'_top => true
      | itv'_z_inf _, itv'_inf_z _ => false
      | itv'_z_inf zx_lb, itv'_z_inf zy_lb => leb zy_lb zx_lb
      | itv'_z_inf _, itv'_z_z _ _ _ => false

      | itv'_z_z _ _ _, itv'_top => true
      | itv'_z_z _ zx_ub _, itv'_inf_z zy_ub => leb zx_ub zy_ub
      | itv'_z_z zx_lb _ _, itv'_z_inf zy_lb => leb zy_lb zx_lb
      | itv'_z_z zx_lb zx_ub _, 
        itv'_z_z zy_lb zy_ub _ => 
        andb (leb zy_lb zx_lb) (leb zx_ub zy_ub)
    end)
}.
Next Obligation.
  case: x; case:y; try (by constructor); try (by move=> *; apply/lebP).
  move => xl xu xlu yl yu ylu.
  apply/equivP; first by apply andP.
  by constructor; move=>[Hl Hr]; split; apply/lebP.
Qed.

Program Definition op_add' x y :=
  match x, y with
    | itv'_top, itv'_top => itv'_top
    | itv'_top, itv'_inf_z _ => itv'_top
    | itv'_top, itv'_z_inf _ => itv'_top
    | itv'_top, itv'_z_z _ _ _ => itv'_top

    | itv'_inf_z _, itv'_top => itv'_top
    | itv'_inf_z zx_ub, itv'_inf_z zy_ub => 
      itv'_inf_z (zx_ub + zy_ub)
    | itv'_inf_z _, itv'_z_inf _ => itv'_top
    | itv'_inf_z zx_ub, itv'_z_z _ zy_ub _ => 
      itv'_inf_z (zx_ub + zy_ub)

    | itv'_z_inf _, itv'_top => itv'_top
    | itv'_z_inf _, itv'_inf_z _ => itv'_top
    | itv'_z_inf zx_lb, itv'_z_inf zy_lb => 
      itv'_z_inf (zx_lb + zy_lb)
    | itv'_z_inf zx_lb, itv'_z_z zy_lb _ _ => 
      itv'_z_inf (zx_lb + zy_lb)

    | itv'_z_z _ _ _, itv'_top => itv'_top
    | itv'_z_z _ zx_ub _, itv'_inf_z zy_ub => 
      itv'_inf_z (zx_ub + zy_ub)
    | itv'_z_z zx_lb _ _, itv'_z_inf zy_lb => 
      itv'_z_inf (zx_lb + zy_lb)
    | itv'_z_z zx_lb zx_ub _, 
      itv'_z_z zy_lb zy_ub _ => 
      itv'_z_z (zx_lb + zy_lb) (zx_ub + zy_ub) _
  end.
Next Obligation. by apply:Z.add_le_mono. Qed.

Lemma add'_preserve_top_l: ∀i, op_add' itv'_top i = itv'_top.
Proof. destruct i; auto. Qed.
Lemma add'_preserve_top_r: ∀i, op_add' i itv'_top = itv'_top.
Proof. destruct i; auto. Qed.

Add Parametric Morphism : op_add' with
    signature eq ==> eq ==> eq as op_add'_mor.
Proof.
  rewrite/eq; simpl; intros x y Hxy v w Hvw.
  destruct x,y,v,w; simpl; repeat inv_eq'; firstorder.
Qed.

Program Definition op_sub' x y :=
  match x, y with
    | itv'_top, itv'_top => itv'_top
    | itv'_top, itv'_inf_z _ => itv'_top
    | itv'_top, itv'_z_inf _ => itv'_top
    | itv'_top, itv'_z_z _ _ _ => itv'_top

    | itv'_inf_z _, itv'_top => itv'_top
    | itv'_inf_z _, itv'_inf_z _ => itv'_top
    | itv'_inf_z zx_ub, itv'_z_inf zy_lb => 
      itv'_inf_z (zx_ub - zy_lb)
    | itv'_inf_z zx_ub, itv'_z_z zy_lb _ _ => 
      itv'_inf_z (zx_ub - zy_lb)

    | itv'_z_inf _, itv'_top => itv'_top
    | itv'_z_inf zx_lb, itv'_inf_z zy_ub => 
      itv'_z_inf (zx_lb - zy_ub)
    | itv'_z_inf zx_lb, itv'_z_inf zy_lb => itv'_top
    | itv'_z_inf zx_lb, itv'_z_z _ zy_ub _ => 
      itv'_z_inf (zx_lb - zy_ub)

    | itv'_z_z _ _ _, itv'_top => itv'_top
    | itv'_z_z zx_lb _ _, itv'_inf_z zy_ub => 
      itv'_z_inf (zx_lb - zy_ub)
    | itv'_z_z _ zx_ub _, itv'_z_inf zy_lb => 
      itv'_inf_z (zx_ub - zy_lb)
    | itv'_z_z zx_lb zx_ub _, 
      itv'_z_z zy_lb zy_ub _ => 
      itv'_z_z (zx_lb - zy_ub) (zx_ub - zy_lb) _
  end.
Next Obligation. by apply:Z.sub_le_mono. Qed.

Lemma op_sub'_preserve_top_l: ∀i, op_sub' itv'_top i = itv'_top.
Proof. destruct i; auto. Qed.
Lemma op_sub'_preserve_top_r: ∀i, op_sub' i itv'_top = itv'_top.
Proof. destruct i; auto. Qed.

Add Parametric Morphism : op_sub' with
    signature eq ==> eq ==> eq as op_sub'_mor.
Proof.
  rewrite/le; simpl; intros x y Hxy v w Hvw.
  destruct x,y,v,w; simpl; repeat inv_eq'; firstorder.
  unfold eq; simpl; firstorder.
  unfold eq; simpl; firstorder.
Qed.

Program Definition op_mul' x y :=
  match x, y with
    | itv'_top, itv'_top => itv'_top
    | itv'_top, itv'_inf_z _ => itv'_top
    | itv'_top, itv'_z_inf _ => itv'_top
    | itv'_top, itv'_z_z zy_lb zy_ub _ => 
      if eqb zy_lb 0%Z && eqb zy_ub 0%Z
      then itv'_z_z 0%Z 0%Z _
      else itv'_top

    | itv'_inf_z _, itv'_top => itv'_top
    | itv'_inf_z zx_ub, itv'_inf_z zy_ub => 
      if leb zx_ub 0%Z && leb zy_ub 0%Z
      then itv'_z_inf (zx_ub * zy_ub)
      else itv'_top
    | itv'_inf_z zx_ub, itv'_z_inf zy_lb => 
      if leb zx_ub 0%Z && leb 0%Z zy_lb
      then itv'_inf_z (zx_ub * zy_lb)
      else itv'_top
    | itv'_inf_z zx_ub, itv'_z_z zy_lb zy_ub _ => 
      if eqb zy_lb 0%Z && eqb zy_ub 0%Z
      then itv'_z_z 0%Z 0%Z _
        else if leb 0%Z zy_lb
          then itv'_inf_z (Z.max (zx_ub * zy_lb) (zx_ub * zy_ub))
          else if leb zy_ub 0%Z
            then itv'_z_inf (Z.min (zx_ub * zy_lb) (zx_ub * zy_ub))
            else itv'_top

    | itv'_z_inf _, itv'_top => itv'_top
    | itv'_z_inf zx_lb, itv'_inf_z zy_ub => 
      if leb 0%Z zx_lb && leb zy_ub 0%Z
      then itv'_inf_z (zx_lb * zy_ub)
      else itv'_top
    | itv'_z_inf zx_lb, itv'_z_inf zy_lb => 
      if leb 0%Z zx_lb && leb 0%Z zy_lb
      then itv'_z_inf (zx_lb * zy_lb)
      else itv'_top
    | itv'_z_inf zx_lb, itv'_z_z zy_lb zy_ub _ => 
      if eqb zy_lb 0%Z && eqb zy_ub 0%Z
      then itv'_z_z 0%Z 0%Z _
        else if leb 0%Z zy_lb
          then itv'_z_inf (Z.min (zx_lb * zy_lb) (zx_lb * zy_ub))
          else if leb zy_ub 0%Z 
            then itv'_inf_z (Z.max (zx_lb * zy_lb) (zx_lb * zy_ub))
            else itv'_top

    | itv'_z_z zx_lb zx_ub _, itv'_top => 
      if eqb zx_lb 0%Z && eqb zx_ub 0%Z
      then itv'_z_z 0 0 _
      else itv'_top
    | itv'_z_z zx_lb zx_ub _, itv'_inf_z zy_ub => 
      if eqb zx_lb 0%Z && eqb zx_ub 0%Z
      then itv'_z_z 0%Z 0%Z _
        else if leb 0%Z zx_lb 
          then itv'_inf_z (Z.max (zx_lb * zy_ub) (zx_ub * zy_ub))
          else if leb zx_ub 0%Z
            then itv'_z_inf (Z.min (zx_lb * zy_ub) (zx_ub * zy_ub))
            else itv'_top
    | itv'_z_z zx_lb zx_ub _, itv'_z_inf zy_lb => 
      if eqb zx_lb 0%Z && eqb zx_ub 0%Z
      then itv'_z_z 0%Z 0%Z _
        else if leb 0%Z zx_lb
          then itv'_z_inf (Z.min (zx_lb * zy_lb) (zx_ub * zy_lb))
          else if leb zx_ub 0%Z 
            then itv'_inf_z (Z.max (zx_lb * zy_lb) (zx_ub * zy_lb))
            else itv'_top
    | itv'_z_z zx_lb zx_ub _, 
      itv'_z_z zy_lb zy_ub _ => 
      itv'_z_z
        (Z.min 
           (Z.min (zx_lb * zy_lb) (zx_lb * zy_ub))
           (Z.min (zx_ub * zy_lb) (zx_ub * zy_ub)))
        (Z.max
           (Z.max (zx_lb * zy_lb) (zx_lb * zy_ub))
           (Z.max (zx_ub * zy_lb) (zx_ub * zy_ub))) _
  end.
Next Obligation.
  unfold eq,le in *; simpl in *.
  eauto using Z.le_trans, Z.le_min_l, Z.le_max_l.
Qed.

Lemma Zmult_le_compat_r_neg: ∀ n m p : Z, 
  (n <= m)%Z → (p <= 0)%Z → (m * p <= n * p)%Z.
Proof.
  intros. apply Z.opp_le_mono.
  do 2 rewrite Zopp_mult_distr_r.
  apply Zmult_le_compat_r; omega.
Qed.

Lemma Zmult_le_compat_l_neg: ∀ n m p : Z, (n <= m)%Z → (p <= 0)%Z → (p * m <= p * n)%Z.
Proof.
  intros. apply Z.opp_le_mono.
  do 2 rewrite Zopp_mult_distr_l.
  apply Zmult_le_compat_l; omega.
Qed.

Lemma Z_min_glb_iff : forall x y z, z ⊑ x /\ z ⊑ y -> z ⊑ Z.min x y.
Proof. unfold le; simpl. apply Z.min_glb_iff. Qed.
Lemma Z_max_lub_iff : forall x y z, x ⊑ z /\ y ⊑ z -> Z.max x y ⊑ z.
Proof. unfold le; simpl. apply Z.max_lub_iff. Qed.

Lemma Z_mul_nonneg_nonneg: ∀ n m : Z, (0 ⊑ n)%Z → (0 ⊑ m)%Z → (0 ⊑ n * m)%Z.
Proof. unfold le; simpl; apply Z.mul_nonneg_nonneg. Qed.
Lemma Z_mul_nonpos_nonpos: ∀ n m : Z, (n ⊑ 0)%Z → (m ⊑ 0)%Z → (0 ⊑ n * m)%Z.
Proof. unfold le; simpl; apply Z.mul_nonpos_nonpos. Qed.
Lemma Z_mul_nonneg_nonpos: ∀ n m : Z, (0 ⊑ n)%Z → (m ⊑ 0)%Z → (n * m ⊑ 0)%Z.
Proof. unfold le; simpl; apply Z.mul_nonneg_nonpos. Qed.
Lemma Z_mul_nonpos_nonneg: ∀ n m : Z, (n ⊑ 0)%Z → (0 ⊑ m)%Z → (n * m ⊑ 0)%Z.
Proof. unfold le; simpl; apply Z.mul_nonpos_nonneg. Qed.

Lemma not_le_imply_le: forall (x y:Z), ¬x ⊑ y -> y ⊑ x.
Proof. unfold le; simpl; intros; omega. Qed.

Lemma Z_max_le_compat: forall n m p q : Z, 
  (n ⊑ m)%Z → (p ⊑ q)%Z → (Z.max n p ⊑ Z.max m q)%Z.
Proof. apply Z.max_le_compat. Qed.
Lemma Z_min_le_compat: forall n m p q : Z, 
  (n ⊑ m)%Z → (p ⊑ q)%Z → (Z.min n p ⊑ Z.min m q)%Z.
Proof. apply Z.min_le_compat. Qed.

Ltac mul_auto := repeat
match goal with
  | [H:itv'_top ⊑ itv'_inf_z _ |- _] => inversion H
  | [H:itv'_top ⊑ itv'_z_inf _ |- _] => inversion H
  | [H:itv'_top ⊑ itv'_z_z _ _ _ |- _] => inversion H
  | [H:itv'_z_inf _ ⊑ itv'_inf_z _ |- _] => inversion H
  | [H:itv'_z_inf _ ⊑ itv'_z_z _ _ _ |- _] => inversion H
  | [H:itv'_inf_z _ ⊑ itv'_z_inf _ |- _] => inversion H
  | [H:itv'_inf_z _ ⊑ itv'_z_z _ _ _ |- _] => inversion H

  | [|- ?x ⊑ ?x] => apply eq_le; reflexivity
  | [H:?x ≡ ?y |- ?x ⊑ ?y] => apply eq_le; exact H
  | [H:?x ≡ ?y |- ?y ⊑ ?x] => apply eq_le; symmetry; exact H
  | [H: ?P |- ?P] => exact H

  | [H:?x ≡ 0%Z |- (?x * _)%Z ⊑ 0%Z] => rewrite H; rewrite Z.mul_0_l
  | [H:?x ≡ 0%Z |- 0%Z ⊑ (?x * _)%Z] => rewrite H; rewrite Z.mul_0_l
  | [H:?x ≡ 0%Z |- (_ * ?x)%Z ⊑ 0%Z] => rewrite H; rewrite Z.mul_0_r
  | [H:?x ≡ 0%Z |- 0%Z ⊑ (_ * ?x)%Z] => rewrite H; rewrite Z.mul_0_r

  | [Hn:(0 ⊑ ?n)%Z,Hm:(0 ⊑ ?m)%Z |- (0 ⊑ ?n * ?m)%Z] => 
    apply Z_mul_nonneg_nonneg
  | [Hn:(?n ⊑ 0)%Z,Hm:(?m ⊑ 0)%Z |- (0 ⊑ ?n * ?m)%Z] => 
    apply Z_mul_nonpos_nonpos
  | [Hn:(0 ⊑ ?n)%Z,Hm:(?m ⊑ 0)%Z |- (?n * ?m ⊑ 0)%Z] => 
    apply Z_mul_nonneg_nonpos
  | [Hn:(?n ⊑ 0)%Z,Hm:(0 ⊑ ?m)%Z |- (?n * ?m ⊑ 0)%Z] => 
    apply Z_mul_nonpos_nonneg

  | [|- ?x ⊑ Z.max ?x _] => apply Z.le_max_l
  | [|- ?x ⊑ Z.max _ ?x] => apply Z.le_max_r
  | [|- Z.min ?x _ ⊑ ?x] => apply Z.le_min_l
  | [|- Z.min _ ?x ⊑ ?x] => apply Z.le_min_r

  | [|- _ -> _] => intros
  | [|- ?x ⊑ itv'_top] => apply (top_prop (H1:=itvT'_hasTop))
  | [H:_ ∧ _ |- _] => destruct H
  | [|-_ ∧ _ ] => split

  | [H:itv'_z_z _ _ _ ⊑ itv'_z_z _ _ _ |- _] => unfold le in H; simpl in H
  | [H:itv'_z_z _ _ _ ⊑ itv'_inf_z _ |- _] => unfold le in H; simpl in H
  | [H:itv'_z_z _ _ _ ⊑ itv'_z_inf _ |- _] => unfold le in H; simpl in H
  | [H:itv'_inf_z _ ⊑ itv'_inf_z _ |- _] => unfold le in H; simpl in H
  | [H:itv'_z_inf _ ⊑ itv'_z_inf _ |- _] => unfold le in H; simpl in H

  | [|- context [if ?x then _ else _]] =>
    let cond := fresh "cond" in
    let Hcond := fresh "Hcond" in 
    remember x as cond eqn:Hcond; symmetry in Hcond; destruct cond
  | [H: _ && _ = true |- _] => move/andP:H => H
  | [H: _ && _ = false |- _] => move/andP:H => H
  | [H : is_true (eqb _ _) |- _] => move/eqbP:H => H
  | [|- is_true (eqb _ _)] => apply/eqbP
  | [H : leb _ _ = true |- _] => move/lebP:H => H
  | [H : leb _ _ = false |- _] => move/lebP:H => H
  | [H : is_true (leb _ _) |- _] => move/lebP:H => H
  | [|- is_true (leb _ _)] => apply/lebP

  | [Hxy:?x ⊑ ?y, Hp:0%Z ⊑ ?p |- (?p * ?x)%Z ⊑ (?p * ?y)%Z] => 
    apply Zmult_le_compat_l; auto
  | [Hxy:?x ⊑ ?y, Hp:0%Z ⊑ ?p |- (?x * ?p)%Z ⊑ (?y * ?p)%Z] => 
    apply Zmult_le_compat_r; auto
  | [Hxy:?x ⊑ ?y, Hp:?p ⊑ 0%Z |- (?p * ?y)%Z ⊑ (?p * ?x)%Z] => 
    apply Zmult_le_compat_l_neg; auto
  | [Hxy:?x ⊑ ?y, Hp:?p ⊑ 0%Z |- (?y * ?p)%Z ⊑ (?x * ?p)%Z] => 
    apply Zmult_le_compat_r_neg; auto
end.

Add Parametric Morphism : op_mul' with
    signature eq ==> eq ==> eq as op_mul'_mor.
Proof.
  intros x y Hxy v w Hvw.
  destruct x,y,v,w; repeat inv_eq'; auto; subst; try reflexivity.
  simpl; split; reflexivity.
Qed.

Program Definition op_div' x y :=
  match x, y with
    | itv'_top, itv'_top => itv'_top
    | itv'_top, itv'_inf_z _ => itv'_top
    | itv'_top, itv'_z_inf _ => itv'_top
    | itv'_top, itv'_z_z _ _ _ => itv'_top

    | itv'_inf_z _, itv'_top => itv'_top
    | itv'_inf_z zx_ub, itv'_inf_z zy_ub => 
      if leb 0%Z zy_ub
      then itv'_top
      else itv'_z_inf (Z.min (Cdiv zx_ub zy_ub) 0%Z)
    | itv'_inf_z zx_ub, itv'_z_inf zy_lb => 
      if leb zy_lb 0%Z
      then itv'_top
      else itv'_inf_z (Z.max (Cdiv zx_ub zy_lb) 0%Z)
    | itv'_inf_z zx_ub, itv'_z_z zy_lb zy_ub _ => 
      if leb zy_lb 0%Z && leb 0%Z zy_ub
      then itv'_top
      else if ltb 0%Z zy_lb
           then itv'_inf_z (Z.max (Cdiv zx_ub zy_lb) (Cdiv zx_ub zy_ub))
           else itv'_z_inf (Z.min (Cdiv zx_ub zy_lb) (Cdiv zx_ub zy_ub))

    | itv'_z_inf _, itv'_top => itv'_top
    | itv'_z_inf zx_lb, itv'_inf_z zy_ub => 
      if leb 0%Z zy_ub
      then itv'_top
      else itv'_inf_z (Z.max (Cdiv zx_lb zy_ub) 0%Z)
    | itv'_z_inf zx_lb, itv'_z_inf zy_lb => 
      if leb zy_lb 0%Z
      then itv'_top
      else itv'_z_inf (Z.min (Cdiv zx_lb zy_lb) 0%Z)
    | itv'_z_inf zx_lb, itv'_z_z zy_lb zy_ub _ => 
      if leb zy_lb 0%Z && leb 0%Z zy_ub
      then itv'_top
      else if ltb 0%Z zy_lb
           then itv'_z_inf (Z.min (Cdiv zx_lb zy_lb) (Cdiv zx_lb zy_ub))
           else itv'_inf_z (Z.max (Cdiv zx_lb zy_lb) (Cdiv zx_lb zy_ub))

    | itv'_z_z _ _ _, itv'_top => itv'_top
    | itv'_z_z zx_lb zx_ub _, itv'_inf_z zy_ub => 
      if leb 0%Z zy_ub 
      then itv'_top
      else itv'_z_z 
             (Z.min (Cdiv zx_ub zy_ub) 0%Z)
             (Z.max (Cdiv zx_lb zy_ub) 0%Z) _
    | itv'_z_z zx_lb zx_ub _, itv'_z_inf zy_lb => 
      if leb zy_lb 0%Z
      then itv'_top
      else itv'_z_z 
             (Z.min (Cdiv zx_lb zy_lb) 0%Z)
             (Z.max (Cdiv zx_ub zy_lb) 0%Z) _
    | itv'_z_z zx_lb zx_ub _, 
      itv'_z_z zy_lb zy_ub _ => 
      if leb zy_lb 0%Z && leb 0%Z zy_ub
      then itv'_top
      else itv'_z_z
            (Z.min 
              (Z.min (Cdiv zx_lb zy_lb) (Cdiv zx_lb zy_ub))
              (Z.min (Cdiv zx_ub zy_lb) (Cdiv zx_ub zy_ub)))
            (Z.max
              (Z.max (Cdiv zx_lb zy_lb) (Cdiv zx_lb zy_ub))
              (Z.max (Cdiv zx_ub zy_lb) (Cdiv zx_ub zy_ub))) _
    end.
Next Obligation. 
  apply Zle_trans with 0%Z; auto using Z.le_min_r, Z.le_max_r.
Qed.
Next Obligation. 
  apply Zle_trans with 0%Z; auto using Z.le_min_r, Z.le_max_r.
Qed.
Next Obligation. 
  apply Zmin_le_Zmax_strong; apply Zmin_le_Zmax_strong; omega.
Qed.

Add Parametric Morphism : op_div' with
    signature eq ==> eq ==> eq as op_div'_mor.
Proof.
  intros x y Hxy v w Hvw.
  destruct x,y,v,w; repeat inv_eq'; subst; try reflexivity;
  repeat destif; auto; repeat constructor.
  unfold eq; simpl; repeat destif; auto; repeat constructor.
  unfold eq; simpl; repeat destif; auto; repeat constructor.
  unfold eq; simpl; repeat destif; auto; repeat constructor.
Qed.

  (* Ltac for cmod_definition *)
Ltac cmod_def :=
  repeat 
    match goal with
      | [H: _ |- (- ?x <= ?x)%Z] => apply pos_imply_opp_le
      | [H: _ |- (0 <= ?x - 1)%Z] => apply Z.le_0_sub
      | [H: _ |- (_ <= Z.max _ _)%Z] => apply Z.max_le_iff
      | [H: _ |- (1 <= Z.abs ?n)%Z] => 
        let Hn := fresh "Hn" in
          assert (Hn: (n ≠ 0)%Z); last (apply Z.abs_pos in Hn; omega)
      | [H: _ |- (- ?n <= 0)%Z] => apply Z.opp_nonpos_nonneg
      | [H: _ |- (0 <= Z.min ?n ?m)%Z] => apply Z.min_glb
      | [Cdlu : (0 < ?dl ∨ ?du < 0)%Z |- (1 <= Z.abs ?dl ∨ 1 <= Z.abs ?du)%Z] =>
        destruct Cdlu; [left|right]
    end; try omega.

Definition op_mod' (x y : itvT') : itvT'.
Proof with cmod_def.
  destruct x as [|u|l|l u Hlu],y as [|du|dl|dl du Hdlu]; simpl in *.

  apply itv'_top.
  apply itv'_top.
  apply itv'_top.
  destruct (l_z_u_dec dl du) as [Hdlzu|Hdlzu].
  { apply itv'_top. }      
  {
    assert (Hf: ((-((Z.max (Z.abs dl) (Z.abs du)) - 1)) <=
      ((Z.max (Z.abs dl) (Z.abs du)) - 1))%Z); first cmod_def.
    apply (itv'_z_z 
      (-((Z.max (Z.abs dl) (Z.abs du)) - 1))
      ((Z.max (Z.abs dl) (Z.abs du)) - 1) Hf). 
  }

  apply itv'_top.
  destruct (Z_le_dec 0 du) as [Hdu|Hdu].
  { apply itv'_top. }
  { 
    destruct (Z_le_dec u 0) as [Hu|Hu].
    apply (itv'_inf_z 0).
    apply (itv'_inf_z u).
  }
  destruct (Z_le_dec dl 0) as [Hdl|Hdl].
  { apply itv'_top. }
  { destruct (Z_le_dec u 0) as [Hu|Hu].
    apply (itv'_inf_z 0).
    apply (itv'_inf_z u).
  }
  destruct (l_z_u_dec dl du) as [Hdlzu|Hdlzu].
  { apply itv'_top. }
  {
    destruct (Z_le_dec u 0) as [Hu|Hu].
    {
      assert (Hf: ((-((Z.max (Z.abs dl) (Z.abs du)) - 1)) <=
        0)%Z); first cmod_def.
      apply (itv'_z_z 
        (-((Z.max (Z.abs dl) (Z.abs du)) - 1))
        0 Hf).
    }
    {
      assert (Hf: ((-((Z.max (Z.abs dl) (Z.abs du)) - 1)) <=
        (Z.min u ((Z.max (Z.abs dl) (Z.abs du)) - 1)))%Z); 
      first apply Zle_trans with 0%Z...
      apply (itv'_z_z 
        (-((Z.max (Z.abs dl) (Z.abs du)) - 1))
        (Z.min u ((Z.max (Z.abs dl) (Z.abs du)) - 1)) Hf).
    }
  }

  apply itv'_top.
  destruct (Z_le_dec 0 du) as [Hdu|Hdu].
  { apply itv'_top. }
  {
    destruct (Z_le_dec 0 l) as [Hl|Hl].
    apply (itv'_z_inf 0).
    apply (itv'_z_inf l).
  }
  destruct (Z_le_dec dl 0) as [Hdl|Hdl].
  { apply itv'_top. }
  { 
    destruct (Z_le_dec 0 l) as [Hl|Hl].
    apply (itv'_z_inf 0).
    apply (itv'_z_inf l).
  }
  destruct (l_z_u_dec dl du) as [Hdlzu|Hdlzu].
  { apply itv'_top. }
  {
    destruct (Z_le_dec 0 l) as [Hl|Hl].
    {
      assert (Hf: (0 <=
        ((Z.max (Z.abs dl) (Z.abs du)) - 1))%Z); first cmod_def.
      apply (itv'_z_z 
        0
        ((Z.max (Z.abs dl) (Z.abs du)) - 1) Hf).
    }
    {
      assert (Hf: ((-(Z.min (-l) ((Z.max (Z.abs dl) (Z.abs du)) - 1))) <=
        ((Z.max (Z.abs dl) (Z.abs du)) - 1))%Z);
      first apply Zle_trans with 0%Z...
      apply (itv'_z_z 
        (-(Z.min (-l) ((Z.max (Z.abs dl) (Z.abs du)) - 1)))
        ((Z.max (Z.abs dl) (Z.abs du)) - 1) Hf).
    }
  }

  apply itv'_top.
  inv_eq'; destruct (Z_le_dec 0 du) as [Hdu|Hdu].
  { apply itv'_top. }
  {
    destruct (Z_le_dec 0 l) as [Hl|Hl].
    {
      assert (Hf: (0 <= u)%Z); first cmod_def.
      apply (itv'_z_z 0 u Hf).
    }
    {
      destruct (Z_le_dec u 0) as [Hu|Hu].
      {
        assert (Hf: (l <= 0)%Z); first cmod_def.
        apply (itv'_z_z l 0 Hf).
      }
      { apply (itv'_z_z l u Hlu). }
    }
  }
  inv_eq'; destruct (Z_le_dec dl 0) as [Hdl|Hdl].
  { apply itv'_top. }
  { 
    destruct (Z_le_dec 0 l) as [Hl|Hl].
    {
      assert (Hf: (0 <= u)%Z); first cmod_def.
      apply (itv'_z_z 0 u Hf).
    }
    {
      destruct (Z_le_dec u 0) as [Hu|Hu].
      {
        assert (Hf: (l <= 0)%Z); first cmod_def.
        apply (itv'_z_z l 0 Hf).
      }
      apply (itv'_z_z l u Hlu).
    }
  }
  inv_eq'; destruct (l_z_u_dec dl du) as [Hdlzu|Hdlzu].
  { apply itv'_top. }
  {
    inv_eq'; destruct (Z_le_dec 0 l) as [Hl|Hl].
    {
      assert (Hf: (0 <=
        (Z.min u ((Z.max (Z.abs dl) (Z.abs du)) - 1)))%Z); first cmod_def.
      apply (itv'_z_z 
        0
        (Z.min u ((Z.max (Z.abs dl) (Z.abs du)) - 1)) Hf).
    }
    {
      destruct (Z_le_dec u 0) as [Hu|Hu].
      {
        assert (Hf: ((-(Z.min (-l) ((Z.max (Z.abs dl) (Z.abs du)) - 1))) <=
          0)%Z); first cmod_def.
        apply (itv'_z_z
          (-(Z.min (-l) ((Z.max (Z.abs dl) (Z.abs du)) - 1)))
          0 Hf).
      }
      { 
        assert (Hf: ((-(Z.min (-l) ((Z.max (Z.abs dl) (Z.abs du)) - 1))) <=
          (Z.min u ((Z.max (Z.abs dl) (Z.abs du)) - 1)))%Z);
        first apply Zle_trans with 0%Z...
        apply (itv'_z_z 
          (-(Z.min (-l) ((Z.max (Z.abs dl) (Z.abs du)) - 1)))
          (Z.min u ((Z.max (Z.abs dl) (Z.abs du)) - 1)) Hf).
      }
    }
  }
Defined.

Add Parametric Morphism : op_mod' with
    signature eq ==> eq ==> eq as op_mod'_mor.
Proof.
  intros x y Hxy v w Hvw.
  destruct x,y,v,w; repeat inv_eq'; subst; try reflexivity; unfold op_mod'.
  destruct (Z_le_dec 0 u2); try (constructor; reflexivity).
  destruct (Z_le_dec 0 l1); try (constructor; reflexivity).
  destruct (Z_le_dec u0 0); try (constructor; reflexivity).
  destruct (Z_le_dec l4 0); try (constructor; reflexivity).
  destruct (Z_le_dec 0 l1); try (constructor; reflexivity).
  destruct (Z_le_dec u0 0); try (constructor; reflexivity).
  destruct (l_z_u_dec l5 u2); try (constructor; reflexivity).
  destruct (Z_le_dec 0 l1); try (constructor; reflexivity).
  destruct (Z_le_dec u0 0); try (constructor; reflexivity).
Qed.

Program Definition op_lt' (x y:itvT') : itvT' := 
  match x, y with
    | itv'_top, itv'_top => itv'_btop
    | itv'_top, itv'_inf_z _ => itv'_btop
    | itv'_top, itv'_z_inf _ => itv'_btop
    | itv'_top, itv'_z_z zy_lb zy_ub _ => itv'_btop

    | itv'_inf_z _, itv'_top => itv'_btop
    | itv'_inf_z zx_ub, itv'_inf_z zy_ub => itv'_btop
    | itv'_inf_z zx_ub, itv'_z_inf zy_lb => 
      if ltb zx_ub zy_lb
        then itv'_z_z 1 1 _
        else itv'_btop
    | itv'_inf_z zx_ub, itv'_z_z zy_lb zy_ub _ => 
      if ltb zx_ub zy_lb
        then itv'_z_z 1 1 _
        else itv'_btop

    | itv'_z_inf _, itv'_top => itv'_btop
    | itv'_z_inf zx_lb, itv'_inf_z zy_ub => 
      if leb zy_ub zx_lb
        then itv'_z_z 0 0 _
        else itv'_btop
    | itv'_z_inf zx_lb, itv'_z_inf zy_lb => itv'_btop
    | itv'_z_inf zx_lb, itv'_z_z zy_lb zy_ub _ => 
      if leb zy_ub zx_lb
        then itv'_z_z 0 0 _
        else itv'_btop

    | itv'_z_z zx_lb zx_ub _, itv'_top => itv'_btop
    | itv'_z_z zx_lb zx_ub _, itv'_inf_z zy_ub => 
      if leb zy_ub zx_lb
        then itv'_z_z 0 0 _
        else itv'_btop
    | itv'_z_z zx_lb zx_ub _, itv'_z_inf zy_lb => 
      if ltb zx_ub zy_lb
        then itv'_z_z 1 1 _
        else itv'_btop
    | itv'_z_z zx_lb zx_ub _, itv'_z_z zy_lb zy_ub _ => 
      if ltb zx_ub zy_lb
        then itv'_z_z 1 1 _
        else if leb zy_ub zx_lb
        then itv'_z_z 0 0 _
        else itv'_btop
  end.

Add Parametric Morphism : op_lt' with
    signature eq ==> eq ==> eq as op_lt'_mor.
Proof.
  intros x y Hxy v w Hvw.
  destruct x,y,v,w; repeat inv_eq'; subst; try reflexivity;
  repeat destif; auto; repeat constructor.
Qed.

Program Definition op_le' (x y:itvT') : itvT':=
  match x, y with
    | itv'_top, itv'_top => itv'_btop
    | itv'_top, itv'_inf_z _ => itv'_btop
    | itv'_top, itv'_z_inf _ => itv'_btop
    | itv'_top, itv'_z_z zy_lb zy_ub _ => itv'_btop

    | itv'_inf_z _, itv'_top => itv'_btop
    | itv'_inf_z zx_ub, itv'_inf_z zy_ub => itv'_btop
    | itv'_inf_z zx_ub, itv'_z_inf zy_lb => 
      if leb zx_ub zy_lb
        then itv'_z_z 1 1 _
        else itv'_btop
    | itv'_inf_z zx_ub, itv'_z_z zy_lb zy_ub _ => 
      if leb zx_ub zy_lb
        then itv'_z_z 1 1 _
        else itv'_btop

    | itv'_z_inf _, itv'_top => itv'_btop
    | itv'_z_inf zx_lb, itv'_inf_z zy_ub => 
      if ltb zy_ub zx_lb
        then itv'_z_z 0 0 _
        else itv'_btop
    | itv'_z_inf zx_lb, itv'_z_inf zy_lb => itv'_btop
    | itv'_z_inf zx_lb, itv'_z_z zy_lb zy_ub _ => 
      if ltb zy_ub zx_lb
        then itv'_z_z 0 0 _
        else itv'_btop

    | itv'_z_z zx_lb zx_ub _, itv'_top => itv'_btop
    | itv'_z_z zx_lb zx_ub _, itv'_inf_z zy_ub => 
      if ltb zy_ub zx_lb
        then itv'_z_z 0 0 _
        else itv'_btop
    | itv'_z_z zx_lb zx_ub _, itv'_z_inf zy_lb => 
      if leb zx_ub zy_lb
        then itv'_z_z 1 1 _
        else itv'_btop
    | itv'_z_z zx_lb zx_ub _, itv'_z_z zy_lb zy_ub _ => 
      if leb zx_ub zy_lb
        then itv'_z_z 1 1 _
        else if ltb zy_ub zx_lb
        then itv'_z_z 0 0 _
        else itv'_btop
  end.

Add Parametric Morphism : op_le' with
    signature eq ==> eq ==> eq as op_le'_mor.
Proof.
  intros x y Hxy v w Hvw.
  destruct x,y,v,w; repeat inv_eq'; subst; try reflexivity;
  repeat destif; auto; repeat constructor.
Qed.

Definition op_gt' (x y:itvT') : itvT' := op_lt' y x.

Add Parametric Morphism : op_gt' with
    signature eq ==> eq ==> eq as op_gt'_mor.
Proof. intros. apply op_lt'_mor; auto. Qed.

Definition op_ge' (x y:itvT') : itvT' := op_le' y x.

Add Parametric Morphism : op_ge' with
    signature eq ==> eq ==> eq as op_ge'_mor.
Proof. intros. apply op_le'_mor; auto. Qed.

Program Definition op_eq' (x y:itvT') : itvT' := 
  match x, y with
    | itv'_top, itv'_top => itv'_btop
    | itv'_top, itv'_inf_z _ => itv'_btop
    | itv'_top, itv'_z_inf _ => itv'_btop
    | itv'_top, itv'_z_z zy_lb zy_ub _ => itv'_btop

    | itv'_inf_z _, itv'_top => itv'_btop
    | itv'_inf_z zx_ub, itv'_inf_z zy_ub => itv'_btop
    | itv'_inf_z zx_ub, itv'_z_inf zy_lb => 
      if ltb zx_ub zy_lb
        then itv'_z_z 0 0 _
        else itv'_btop
    | itv'_inf_z zx_ub, itv'_z_z zy_lb zy_ub _ => 
      if ltb zx_ub zy_lb
        then itv'_z_z 0 0 _
        else itv'_btop

    | itv'_z_inf _, itv'_top => itv'_btop
    | itv'_z_inf zx_lb, itv'_inf_z zy_ub => 
      if ltb zy_ub zx_lb
        then itv'_z_z 0 0 _
        else itv'_btop
    | itv'_z_inf zx_lb, itv'_z_inf zy_lb => itv'_btop
    | itv'_z_inf zx_lb, itv'_z_z zy_lb zy_ub _ => 
      if ltb zy_ub zx_lb
        then itv'_z_z 0 0 _
        else itv'_btop

    | itv'_z_z zx_lb zx_ub _, itv'_top => itv'_btop
    | itv'_z_z zx_lb zx_ub _, itv'_inf_z zy_ub => 
      if ltb zy_ub zx_lb
        then itv'_z_z 0 0 _
        else itv'_btop
    | itv'_z_z zx_lb zx_ub _, itv'_z_inf zy_lb => 
      if ltb zx_ub zy_lb
        then itv'_z_z 0 0 _
        else itv'_btop
    | itv'_z_z zx_lb zx_ub _, itv'_z_z zy_lb zy_ub _ => 
      if ltb zx_ub zy_lb
        then itv'_z_z 0 0 _
        else if ltb zy_ub zx_lb
        then itv'_z_z 0 0 _
        else if eqb zx_lb zx_ub && eqb zx_ub zy_lb && eqb zy_lb zy_ub
        then itv'_z_z 1 1 _
        else itv'_btop
  end.

Add Parametric Morphism : op_eq' with
    signature eq ==> eq ==> eq as op_eq'_mor.
Proof.
  intros x y Hxy v w Hvw.
  destruct x,y,v,w; repeat inv_eq'; subst; try reflexivity;
  repeat destif; auto; repeat constructor.
Qed.

Program Definition op_not' (x: itvT') : itvT' := 
  if eqb (itv'_z_z 0 0 _) x 
  then itv'_z_z 1 1 _
  else if leb (itv'_z_z 0 0 _) x
  then itv'_btop
  else itv'_z_z 0 0 _.

Add Parametric Morphism : op_not' with
    signature eq ==> eq as op_not'_mor.
Proof.
  intros x y Hxy. 
  destruct x,y; repeat inv_eq'; subst; try reflexivity;
  repeat destif; auto; repeat constructor.
Qed.

Definition op_neq' (x y:itvT') : itvT' := op_not' (op_eq' x y).

Add Parametric Morphism : op_neq' with
    signature eq ==> eq ==> eq as op_neq'_mor.
Proof.
  intros x y Hxy v w Hvw. apply op_not'_mor; apply op_eq'_mor; auto.
Qed.

Program Definition op_and' (x y:itvT') : itvT' := 
  if eqb (itv'_z_z 0 0 _) x 
  then itv'_z_z 0 0 _
  else if eqb (itv'_z_z 0 0 _) y
  then itv'_z_z 0 0 _
  else if negb (leb (itv'_z_z 0 0 _) x) && negb (leb (itv'_z_z 0 0 _) y)
  then itv'_z_z 1 1 _
  else itv'_btop.

Add Parametric Morphism : op_and' with
    signature eq ==> eq ==> eq as op_and'_mor.
Proof.
  intros x y Hxy v w Hvw.
  destruct x,y,v,w; repeat inv_eq'; subst; try reflexivity;
  repeat destif; auto; repeat constructor.
Qed.

Program Definition op_or' (x y:itvT') : itvT' :=
  if eqb (itv'_z_z 0 0 _) x && eqb (itv'_z_z 0 0 _) y
  then itv'_z_z 0 0 _
  else if negb (leb (itv'_z_z 0 0 _) x) || negb (leb (itv'_z_z 0 0 _) y)
  then itv'_z_z 1 1 _
  else itv'_btop.

Add Parametric Morphism : op_or' with
    signature eq ==> eq ==> eq as op_or'_mor.
Proof.
  intros x y Hxy v w Hvw.
  destruct x,y,v,w; repeat inv_eq'; subst; try reflexivity;
  repeat destif; auto; repeat constructor.
Qed.

Definition op_band' (x y:itvT') : itvT' := itv'_top.

Add Parametric Morphism : op_band' with
    signature eq ==> eq ==> eq as op_band'_mor.
Proof. intros. unfold op_band'. reflexivity. Qed.

Definition op_bor' (x y:itvT') : itvT' := itv'_top.

Add Parametric Morphism : op_bor' with
    signature eq ==> eq ==> eq as op_bor'_mor.
Proof. intros. unfold op_bor'. reflexivity. Qed.

Definition op_bxor' (x y:itvT') : itvT' := itv'_top.

Add Parametric Morphism : op_bxor' with
    signature eq ==> eq ==> eq as op_bxor'_mor.
Proof. intros. unfold op_bxor'. reflexivity. Qed.

Definition op_lshift' (x y:itvT') : itvT' := itv'_top.

Add Parametric Morphism : op_lshift' with
    signature eq ==> eq ==> eq as op_lshift'_mor.
Proof. intros. unfold op_lshift'. reflexivity. Qed.

Definition op_rshift' (x y:itvT') : itvT' := itv'_top.

Add Parametric Morphism : op_rshift' with
    signature eq ==> eq ==> eq as op_rshift'_mor.
Proof. intros. unfold op_rshift'. reflexivity. Qed.

Definition op_bnot' (x: itvT') : itvT' := itv'_top.

Add Parametric Morphism : op_bnot' with
    signature eq ==> eq as op_bnot'_mor.
Proof. intros. unfold op_bnot'. reflexivity. Qed.

Definition itvT := option itvT'.

Definition op_add : itvT -> itvT -> itvT := lift2 op_add'.
Definition op_sub : itvT -> itvT -> itvT := lift2 op_sub'.
Definition op_mul : itvT -> itvT -> itvT := lift2 op_mul'.
Definition op_div : itvT -> itvT -> itvT := lift2 op_div'.
Definition op_mod : itvT -> itvT -> itvT := lift2 op_mod'.
Definition op_lt : itvT -> itvT -> itvT := lift2 op_lt'.
Definition op_le : itvT -> itvT -> itvT := lift2 op_le'.
Definition op_gt : itvT -> itvT -> itvT := lift2 op_gt'.
Definition op_ge : itvT -> itvT -> itvT := lift2 op_ge'.
Definition op_eq : itvT -> itvT -> itvT := lift2 op_eq'.
Definition op_neq : itvT -> itvT -> itvT := lift2 op_neq'.
Definition op_and : itvT -> itvT -> itvT := lift2 op_and'.
Definition op_or : itvT -> itvT -> itvT := lift2 op_or'.
Definition op_band : itvT -> itvT -> itvT := lift2 op_band'.
Definition op_bor : itvT -> itvT -> itvT := lift2 op_bor'.
Definition op_bxor : itvT -> itvT -> itvT := lift2 op_bxor'.
Definition op_lshift : itvT -> itvT -> itvT := lift2 op_lshift'.
Definition op_rshift : itvT -> itvT -> itvT := lift2 op_rshift'.

Definition op_not : itvT -> itvT := lift1 op_not'.
Definition op_bnot : itvT -> itvT := lift1 op_bnot'.

Add Parametric Morphism : op_add with
    signature eq ==> eq ==> eq as op_add_mor.
Proof. exact/lift2_mor. Qed.

Add Parametric Morphism : op_sub with
    signature eq ==> eq ==> eq as op_sub_mor.
Proof. exact/lift2_mor. Qed.

Add Parametric Morphism : op_mul with
    signature eq ==> eq ==> eq as op_mul_mor.
Proof. exact/lift2_mor. Qed.

Add Parametric Morphism : op_div with
    signature eq ==> eq ==> eq as op_div_mor.
Proof. exact/lift2_mor. Qed.

Add Parametric Morphism : op_mod with
    signature eq ==> eq ==> eq as op_mod_mor.
Proof. exact/lift2_mor. Qed.

Add Parametric Morphism : op_lt with
    signature eq ==> eq ==> eq as op_lt_mor.
Proof. exact/lift2_mor. Qed.

Add Parametric Morphism : op_le with
    signature eq ==> eq ==> eq as op_le_mor.
Proof. exact/lift2_mor. Qed.

Add Parametric Morphism : op_gt with
    signature eq ==> eq ==> eq as op_gt_mor.
Proof. exact/lift2_mor. Qed.

Add Parametric Morphism : op_ge with
    signature eq ==> eq ==> eq as op_ge_mor.
Proof. exact/lift2_mor. Qed.

Add Parametric Morphism : op_eq with
    signature eq ==> eq ==> eq as op_eq_mor.
Proof. exact/lift2_mor. Qed.

Add Parametric Morphism : op_neq with
    signature eq ==> eq ==> eq as op_neq_mor.
Proof. exact/lift2_mor. Qed.

Add Parametric Morphism : op_and with
    signature eq ==> eq ==> eq as op_and_mor.
Proof. exact/lift2_mor. Qed.

Add Parametric Morphism : op_or with
    signature eq ==> eq ==> eq as op_or_mor.
Proof. exact/lift2_mor. Qed.

Add Parametric Morphism : op_band with
    signature eq ==> eq ==> eq as op_band_mor.
Proof. exact/lift2_mor. Qed.

Add Parametric Morphism : op_bor with
    signature eq ==> eq ==> eq as op_bor_mor.
Proof. exact/lift2_mor. Qed.

Add Parametric Morphism : op_bxor with
    signature eq ==> eq ==> eq as op_bxor_mor.
Proof. exact/lift2_mor. Qed.

Add Parametric Morphism : op_lshift with
    signature eq ==> eq ==> eq as op_lshift_mor.
Proof. exact/lift2_mor. Qed.

Add Parametric Morphism : op_rshift with
    signature eq ==> eq ==> eq as op_rshift_mor.
Proof. exact/lift2_mor. Qed.

Add Parametric Morphism : op_not with
    signature eq ==> eq as op_not_mor.
Proof. exact/lift1_mor. Qed.

Add Parametric Morphism : op_bnot with
    signature eq ==> eq as op_bnot_mor.
Proof. exact/lift1_mor. Qed.
