Require Import Main.
Require Import Arith ZArith.

Require Import BinInt BinNat BinPos BinPosDef.

Require Import extraction.ExtrOcamlBasic.
Require Import extraction.ExtrOcamlNatInt.
Require Import extraction.ExtrOcamlZInt.

Set Extraction AccessOpaque.

Extraction Blacklist List Main.

Extract Constant negb => "not".
Extract Constant length => "List.length".
Extract Constant fst => "fst".
Extract Constant snd => "snd".

(* BinInt *)
Extract Constant BinInt.Z.two => "2".
Extract Constant BinInt.Z.double => "(fun z -> 2*z)".
Extract Constant BinInt.Z.succ_double => "(fun z -> 2*z + 1)".
Extract Constant BinInt.Z.pred_double => "(fun z -> 2*z - 1)".
Extract Constant BinInt.Z.pos_sub => "(fun x y -> Pervasives.abs (x-y))".
Extract Constant BinInt.Z.pow_pos => "(fun z n -> int_of_float ((float_of_int z) ** (float_of_int n)))".
Extract Constant BinInt.Z.pow => "(fun z n -> pow_pos z n)".
Extract Constant BinInt.Z.square => "(fun z -> z*z)".
Extract Constant BinInt.Z.sgn => "(fun z -> if z=0 then 0 else if z>0 then 1 else (-1))".
Extract Constant BinInt.Z.leb => "(fun x y -> x<=y)".
Extract Constant BinInt.Z.ltb => "(fun x y -> x<y)".
Extract Constant BinInt.Z.geb => "(fun x y -> x>=y)".
Extract Constant BinInt.Z.gtb => "(fun x y -> x>y)".
Extract Constant BinInt.Z.eqb => "(fun x y -> x=y)".
Extract Constant BinInt.Z.abs_nat => "(fun z -> Pervasives.abs z)".
Extract Constant BinInt.Z.to_nat => "(fun z -> z)".
Extract Constant BinInt.Z.to_N => "(fun z -> z)".
Extract Constant BinInt.Z.of_nat => "(fun z -> z)".
Extract Constant BinInt.Z.of_N => "(fun z -> z)".
Extract Constant BinInt.Z.to_pos => "(fun z -> z)".
Extract Constant BinInt.Z.div => "(fun x y -> if y=0 then 0 else x/y)".
Extract Constant BinInt.Z.modulo => "(fun x y -> if y=0 then x else x mod y)".
Extract Constant BinInt.Z.even => "(fun z -> (z mod 2) = 0)".
Extract Constant BinInt.Z.odd => "(fun z -> (z mod 2) = 1)".
Extract Constant BinInt.Z.div2 => "(fun z -> z/2)".
Extract Constant BinInt.Z.sqrt => "(fun z -> int_of_float (sqrt (float_of_int z)))".
Extract Constant BinInt.Z.eq_dec => "(fun x y -> x=y)".

(* BinNat *)
Extract Constant BinNat.N.two => "2".
Extract Constant BinNat.N.double => "(fun z -> 2*z)".
Extract Constant BinNat.N.succ_double => "(fun z -> 2*z + 1)".
Extract Constant BinNat.N.succ_pos => "(fun z -> z+1)".
Extract Constant BinNat.N.eqb => "(fun x y -> x=y)".
Extract Constant BinNat.N.leb => "(fun x y -> x<=y)".
Extract Constant BinNat.N.ltb => "(fun x y -> x<y)".
Extract Constant BinNat.N.div2 => "(fun z -> z/2)".
Extract Constant BinNat.N.even => "(fun z -> (z mod 2) = 0)".
Extract Constant BinNat.N.odd => "(fun z -> (z mod 2) = 1)".
Extract Constant BinNat.N.pow => "(fun z n -> int_of_float ((float_of_int z) ** (float_of_int n)))".
Extract Constant BinNat.N.square => "(fun z -> z*z)".
Extract Constant BinNat.N.sqrt => "(fun z -> int_of_float (sqrt (float_of_int z)))".
Extract Constant BinNat.N.to_nat => "(fun z -> z)".
Extract Constant BinNat.N.of_nat => "(fun z -> z)".
Extract Constant BinNat.N.eq_dec => "(fun x y -> x=y)".

(* BinPos *)
Extract Constant BinPos.Pos.add_carry => "(fun x y -> x+y+1)".
Extract Constant BinPos.Pos.pred_double => "(fun z -> 2*z - 1)".
Extract Constant BinPos.Pos.pred_N => "pred".
Extract Constant BinPos.Pos.pow => "(fun z n -> int_of_float ((float_of_int z) ** (float_of_int n)))".
Extract Constant BinPos.Pos.square => "(fun z -> z*z)".
Extract Constant BinPos.Pos.div2 => "(fun z -> z/2)".
Extract Constant BinPos.Pos.eqb => "(fun x y -> x=y)".
Extract Constant BinPos.Pos.leb => "(fun x y -> x<=y)".
Extract Constant BinPos.Pos.ltb => "(fun x y -> x<y)".
Extract Constant BinPos.Pos.sqrt => "(fun z -> int_of_float (sqrt (float_of_int z)))".
Extract Constant BinPos.Pos.to_nat => "(fun z -> z)".
Extract Constant BinPos.Pos.of_nat => "(fun z -> if z<=0 then 1 else z)".
Extract Constant BinPos.Pos.eq_dec => "(fun x y -> x=y)".

(* BinPosDef *)
Extract Constant BinPosDef.Pos.succ => "(fun z -> z+1)".
Extract Constant BinPosDef.Pos.add => "(fun x y -> x+y)".
Extract Constant BinPosDef.Pos.add_carry => "(fun x y -> x+y+1)".
Extract Constant BinPosDef.Pos.pred_double => "(fun z -> 2*z - 1)".
Extract Constant BinPosDef.Pos.pred => "(fun z -> if z<=1 then z else z-1)".
Extract Constant BinPosDef.Pos.pred_N => "pred".
Extract Constant BinPosDef.Pos.sub => "(fun x y -> x-y)".
Extract Constant BinPosDef.Pos.mul => "(fun x y -> x*y)".
Extract Constant BinPosDef.Pos.pow => "(fun z n -> int_of_float ((float_of_int z) ** (float_of_int n)))".
Extract Constant BinPosDef.Pos.square => "(fun z -> z*z)".
Extract Constant BinPosDef.Pos.div2 => "(fun z -> z/2)".
Extract Constant BinPosDef.Pos.min => "Pervasives.min".
Extract Constant BinPosDef.Pos.max => "Pervasives.max".
Extract Constant BinPosDef.Pos.eqb => "(fun x y -> x=y)".
Extract Constant BinPosDef.Pos.leb => "(fun x y -> x<=y)".
Extract Constant BinPosDef.Pos.ltb => "(fun x y -> x<y)".
Extract Constant BinPosDef.Pos.sqrt => "(fun z -> int_of_float (sqrt (float_of_int z)))".
Extract Constant BinPosDef.Pos.to_nat => "(fun z -> z)".
Extract Constant BinPosDef.Pos.of_nat => "(fun z -> if z<=0 then 1 else z)".
Extract Constant BinPosDef.Pos.of_succ_nat => "(fun z -> z+1)".

Separate Extraction validate.
