Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.
Require Import Setoid.
Require Import Morphisms.

Require Import LibTactics.
Require Import DomType.
Require Import DomBasic.
Require Import DomBridge.

Require Import DomAbsValue.
Require Import DomSet.
Require Import DomMap.
Require Import DomAcc.
Require Import DomAbsMem.
Require Import DomAbsState.

Require Import Syntax.

Require Import SemAbsBlock.
Require Import SemAccInsen.
Require Import SemInsenLocal.

Set Implicit Arguments.

Definition validate_postfix (prog:progT) (acc:accT) (state:stateT) :=
  andb (acc_correct prog acc state)
       (postfix prog acc state).

Section Accs.
  Variable (prog:progT) (acc:accT).
  
  Definition acc_correct_acclocs_blk (fid:fidT) (bid:bidT) (cmd:cmdT) (state:func_stateT): bool :=
    let instrs := cinstrs cmd in
    let tmn := ctmn cmd in
    let instrs_mem: memT := state[(bid,false)]! in
    let tmn_mem': memT := SemAbsBlock.sem_instrs acc fid (cinstrs cmd) state[(bid,false)]! in
    let tmn_mem: memT := state[(bid,true)]! in
    acc_correct_acclocs_instrs acc fid instrs instrs_mem &&
    acc_correct_acclocs_tmn acc fid tmn tmn_mem' &&
    acc_correct_acclocs_tmn acc fid tmn tmn_mem.

  Definition acc_correct_acclocs_func (fid: fidT) (fbody:funcT) (state:func_stateT): bool :=
    let alocs := mlookup (Key:=fidT) (Elt:=Set! [locT]) fid (acclocs acc) in
    andb
    (map_forall
      (fun bid cmd => acc_correct_acclocs_blk fid bid cmd state)
      (fcmds fbody))
    (list_forall
      (fun i => (inr (fid, i), nil) !\in alocs)
      (fparams fbody)).

  Definition acc_correct_acclocs_star_func (fid: fidT): bool :=
    set_forall
      (fun callinfo =>
         andb (callinfo.1 !\in ((callfroms_star acc)[fid]!:Set![fidT]))
              (leb (A:=Set![fidT])
                   (callfroms_star acc)[callinfo.1]!
                   (callfroms_star acc)[fid]!))
      ((callfroms acc)[fid]! : Set![fidT * bidT]).

  Definition acc_correct_acclocs_callfroms_func (fid:fidT): bool :=
    set_forall
      (fun callinfo =>
         leb (A:=Set![locT]) (acclocs acc)[fid]! (acclocs acc)[callinfo.1]!)
      ((callfroms acc)[fid]! : Set![fidT * bidT]).

  Definition use_le_var
    (fid:fidT) (x:varT) (lhs rhs:memT): bool :=
    leb (A:=valueT) lhs[(var_loc fid x nil)]! rhs[(var_loc fid x nil)]!.

  Fixpoint use_le_expr (fid:fidT) (e: exprT) (lhs rhs: memT): bool :=
    match e with
      | expr_const _ => true
      | expr_op2 o e1 e2 =>
        use_le_expr fid e1 lhs rhs &&
        use_le_expr fid e2 lhs rhs        
      | expr_op1 o e1 =>
        use_le_expr fid e1 lhs rhs
      | expr_value_of lv =>
        use_le_lv fid lv lhs rhs &&
        match sem_expr_lv fid lv lhs with
        | Elt s =>
          SetInterface.fold
            (fun (loc:locT) (acc:bool) => acc && (leb (A:=valueT) lhs[loc]! rhs[loc]!))
            s
            true
        | Top =>
          false
        end
      | expr_addr_of lv => use_le_lv fid lv lhs rhs
      | expr_cast _ e1 => use_le_expr fid e1 lhs rhs
      | expr_size_of e1 => use_le_expr fid e1 lhs rhs
    end

  with use_le_lv (fid:fidT) (l:lv) (lhs rhs: memT): bool :=
         match l with
           | lv_id x => true
           | lv_deref e1 => use_le_expr fid e1 lhs rhs
           | lv_array e1 e2 =>
             use_le_expr fid e1 lhs rhs &&
             use_le_expr fid e2 lhs rhs
           | lv_struct l pos =>
             use_le_lv fid l lhs rhs
         end.

  Fixpoint use_le_exprs (fid:fidT) (es: list exprT) (lhs rhs: memT): bool :=
    match es with
      | nil => true
      | cons e es1 =>
        use_le_expr fid e lhs rhs &&
        use_le_exprs fid es1 lhs rhs
    end.

  Definition use_le_locs (fid:fidT) (e: exprT) (lhs rhs: memT): bool :=
    match locs_of (sem_expr fid e lhs) with
      | Elt s =>
        SetInterface.fold
          (fun (loc:locT) (acc:bool) => acc && (leb (A:=valueT) lhs[loc]! rhs[loc]!))
          s
          true
      | Top =>
        false
    end.
  
  Fixpoint use_le_locss
           (fid:fidT) (args:list exprT) (lhs rhs:memT): bool :=
    match args with
      | nil => true
      | cons arg tl =>
        use_le_locs fid arg lhs rhs &&
        (use_le_locss fid tl lhs rhs)
    end.

  Definition use_le_tmn (fid:fidT) (tmn:tmnT) (lhs rhs:memT): bool :=
    match tmn with
      | tmn_jump _ => true
      | tmn_call _ _ es ret =>
        use_le_exprs fid es lhs rhs &&
        (if ret is Some ret'
         then use_le_var fid ret' lhs rhs
         else true)
      | tmn_exit e =>
        use_le_expr fid e lhs rhs
    end.

  Definition use_le_instr
             (fid:fidT) (instr:instrT) (lhs rhs:memT): bool :=
    match instr with
      | instr_assign lv e => 
        use_le_lv fid lv lhs rhs &&
        use_le_expr fid e lhs rhs
      | instr_alloc lv (alloc_arr sz str) site =>
        use_le_lv fid lv lhs rhs &&
        use_le_expr fid sz lhs rhs &&
        use_le_expr fid str lhs rhs
      | instr_assume e => use_le_expr fid e lhs rhs
      | instr_libcall ol lc args site =>
        (* Hypothesis for return lvalue: except for scanf *)
        match lc with
          | scanf => true
          | strcpy =>
            let ldst := locs_of (sem_expr fid (hd (expr_const lit_bot) args) lhs) in
            let rdst := locs_of (sem_expr fid (hd (expr_const lit_bot) args) rhs) in

            let lsrc := locs_of (sem_expr fid (hd (expr_const lit_bot) (tl args)) lhs) in
            let rsrc := locs_of (sem_expr fid (hd (expr_const lit_bot) (tl args)) rhs) in

            let lsrcValue := memory_lookup lsrc lhs in
            let rsrcValue := memory_lookup rsrc rhs in
            
            let lnewMem := memory_update acc ldst lsrcValue lhs in
            let rnewMem := memory_update acc rdst rsrcValue rhs in

            if ol is Some lv
            then use_le_lv fid lv lnewMem rnewMem
            else true
          | _ =>
            if ol is Some lv
            then use_le_lv fid lv lhs rhs
            else true
        end
        &&

        (* Hypothesis for argument expressions:
          strcpy, scanf, builtin_alloca, and builtin_choose_expr *)
        match lc with
          | strcpy => use_le_exprs fid args lhs rhs
          | scanf => use_le_exprs fid args lhs rhs
          | builtin_alloca => use_le_exprs fid args lhs rhs
          | builtin_choose_expr => use_le_exprs fid args lhs rhs
          | _ => true
        end
        &&

        (* Hypothesis for argument lvalues:
          strcpy and scanf *)
        match lc with
          | strcpy => use_le_locss fid args lhs rhs
          | _ => true
        end
      | instr_forget aloc => true
      | instr_skip => true
    end.

  Fixpoint use_le_instrs (fid:fidT) (instrs: list instrT) (lhs rhs: memT) :=
    match instrs with
      | nil => true
      | cons instr next_instrs =>
        use_le_instr fid instr lhs rhs &&
        (let newlhs := sem_instr acc fid instr lhs in
         let newrhs := sem_instr acc fid instr rhs in
         use_le_instrs fid next_instrs newlhs newrhs)
    end.

  Definition use_le_blk (fid:fidT) (bid:bidT) (cmd:cmdT) (lin rin: memT): bool :=
    let instrs := cinstrs cmd in
    let tmn := ctmn cmd in
    use_le_instrs fid instrs lin rin &&
    use_le_tmn fid tmn
    (sem_instrs acc fid instrs lin)
    (sem_instrs acc fid instrs rin).

  Definition use_le_func (fid:fidT) (func:funcT) (lhs rhs: func_stateT): bool :=
    map_forall
      (fun bid cmd => use_le_blk
                        fid bid cmd
                        lhs[(bid,false)]!
                        rhs[(bid,false)]!)
      (fcmds func).
End Accs.


Section EqbInters.
  Variable (prog:progT) (acc:accT).

  Definition eqb_inter_tmn (fid:fidT) (tmn:tmnT) (lhs rhs:memT): bool :=
    match tmn with
      | tmn_jump _ => true
      | tmn_call nbs _ _ _ => eqb lhs rhs
      | tmn_exit e => eqb lhs rhs
    end.

  Definition eqb_inter_blk (fid:fidT) (bid:bidT) (cmd:cmdT) (lhs rhs: func_stateT): bool :=
    match ctmn cmd with
      | tmn_jump _ => true
      | tmn_call nbid _ _ _ =>
        (eqb (A:=memT) lhs[(nbid,false)]! rhs[(nbid,false)]!) &&
        (eqb (A:=memT) lhs[(bid,true)]! rhs[(bid,true)]!)
      | tmn_exit e => eqb (A:=memT) lhs[(bid,true)]! rhs[(bid,true)]!
    end.

  Definition eqb_inter_func (fid:fidT) (func:funcT) (lhs rhs: func_stateT): bool :=
    eqb (A:=memT) lhs[(fentry func,false)]! rhs[(fentry func,false)]! &&
    map_forall
      (fun bid cmd => eqb_inter_blk fid bid cmd lhs rhs)
      (fcmds func).
End EqbInters.


Section Postfixs.
  Variable (prog:progT) (acc:accT) (state:stateT').

  Definition postfix_jump (state:func_stateT) (fid:fidT) (func:funcT) (bid:bidT) (nbids:Set![bidT]) :=
    set_forall
      (fun nbid => leb (A:=memT) state[(bid,true)]! state[(nbid,false)]!)
      nbids.

  Definition postfix_call (state:func_stateT) (fid:fidT) (func:funcT) (bid:bidT) (nbid:bidT) (nfid:fidT) (args:list exprT) :=
    if (pfuncs prog)[nfid] is Some nfunc
    then let locs := (acclocs acc)[nfid]! in
         let locs' := (set_to_powerset locs) in
         let mem : memT := state[(bid,true)]! in
         let argvals := sem_exprs fid args mem in
         let mem_in : memT := mrestrict (E:=valueT) locs' mem in
         let mem_out : memT := mrestrict (E:=valueT) (powerset_neg locs') mem in
         let mem_bound := sem_bind_params acc nfid argvals (fparams nfunc) mem_in in
         andb (smem (S:=Set![fidT * bidT]) (fid,bid) (callfroms acc)[nfid]!)
              (leb mem_out state[(nbid,false)]!)
    else true.

  Definition postfix_block (state:func_stateT) (fid:fidT) (func:funcT) (bid:bidT) (cmd:cmdT) :=
    let: (instrs, tmn) := cmd in
    (leb (sem_instrs acc fid instrs state[(bid,false)]!) state[(bid,true)]!) &&
    (match tmn with
       | tmn_jump nbids => postfix_jump state fid func bid nbids
       | tmn_call nbid nfids args retto =>
         set_forall
           (fun nfid => postfix_call state fid func bid nbid nfid args)
           nfids
       | tmn_exit ret => true
     end).
  
  Definition postfix_func (state:func_stateT) (fid:fidT) (func:funcT) : bool :=
    map_forall
      (postfix_block state fid func)
      (fcmds func).

  Definition postfix_inter_call (func_state:func_stateT) (state:stateT') (fid:fidT) (func:funcT) (bid:bidT) (nbid:bidT) (nfid:fidT) (args:list exprT) :=
    if (pfuncs prog)[nfid] is Some nfunc
    then let locs := (acclocs acc)[nfid]! in
         let locs' := (set_to_powerset locs) in
         let mem : memT := func_state[(bid,true)]! in
         let argvals := sem_exprs fid args mem in
         let mem_in : memT := mrestrict (E:=valueT) locs' mem in
         let mem_bound := sem_bind_params acc nfid argvals (fparams nfunc) mem_in in
         (leb mem_bound (state[nfid]!)[(fentry nfunc, false)]!)
    else true.

  Definition postfix_inter_exit (func_state:func_stateT) (state:stateT') (fid:fidT) (func:funcT) (bid:bidT) (ret:exprT) :=
    let mem := func_state[(bid,true)]! in
    let result := sem_expr fid ret mem in
    set_forall
      (fun (callinfo : fidT * bidT) =>
         if (pfuncs prog)[callinfo.1] is Some nfunc
         then if (fcmds nfunc)[callinfo.2] is Some cmd
              then match ctmn cmd with
                     | tmn_call nbid _ _ retto =>
                       let npidx := (callinfo.1, nbid, false) in
                       let mem_removed :=
                           if not_recursive acc fid
                           then mrestrict (E:=valueT) (filter_fid fid) mem
                           else mem
                       in
                       let mem_bound := sem_bind_ret acc callinfo.1 result retto mem_removed in
                       leb mem_bound (state[callinfo.1]!)[(nbid, false)]!
                     | _ => true
                   end
              else true
         else true)
      (callfroms acc)[fid]!.
  
  Definition postfix_inter_block (func_state:func_stateT) (state:stateT') (fid:fidT) (func:funcT) (bid:bidT) (cmd:cmdT) :=
    (match cmd.2 with
       | tmn_jump nbids => true
       | tmn_call nbid nfids args retto =>
         set_forall
           (fun nfid => postfix_inter_call func_state state fid func bid nbid nfid args)
           nfids
       | tmn_exit ret => postfix_inter_exit func_state state fid func bid ret
     end).

  Definition postfix_inter_func (state:stateT') (fid:fidT) (func:funcT) : bool :=
    map_forall
      (postfix_inter_block state[fid]! state fid func)
      (fcmds func).
End Postfixs.


Section Validators.
  Variables (prog:progT) (acc:accT).

  Definition notop (state: func_stateT): bool :=
    map_forall
      (fun _ mem =>
         if mem is Elt _ then true else false)
      state.
  
  Definition validate_func (fid:fidT) (func:funcT) (func_state propagated_func_state:func_stateT) (propagated_inter_state:stateT') : bool :=
    [&& (acc_correct_callfroms_func acc fid func),
     (acc_correct_acclocs_star_func acc fid),
     (acc_correct_acclocs_func acc fid func propagated_func_state),
     (acc_correct_acclocs_callfroms_func acc fid),
     (postfix_func prog acc propagated_func_state fid func),
     (eqb_inter_func fid func propagated_func_state propagated_inter_state[fid]!) &
     (use_le_func acc fid func propagated_func_state func_state)].

  Definition validate (state propagated_state propagated_inter_state: stateT') : bool :=
    [&& map_forall
          (fun fid func =>
             (validate_func fid func state[fid]! propagated_state[fid]! propagated_inter_state) &&
             (postfix_inter_func prog acc propagated_inter_state fid func))
          (pfuncs prog),
     map_forall
       (fun fid _ => (pfuncs prog)[fid])
       (callfroms acc) &
     map_forall
       (fun _ func_state => notop func_state)
       propagated_state].
End Validators.
