Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.
Require Import Setoid.
Require Import Morphisms.

Require Import LibTactics.
Require Import DomType.
Require Import DomBasic.
Require Import DomUtil.
Require Import DomSet.
Require Import DomMap.

Require Import Syntax.

Require Import DomAbsValue.
Require Import DomAcc.
Require Import DomAbsMem.
Require Import DomPidx.
Require Import DomPidxSen.

Require Import SemAbsBlock.

Section Accs.
  Variable (prog:progT) (acc:accT).

  Definition stateT := pidxT ->> memT.
  Global Instance stateT_maplike : maplike (Key:=pidxT) (Elt:=memT) (M:=stateT) := _.
  
  Definition acc_correct_callfroms_blk
             (fid:fidT) (bid:bidT) (cmd:cmdT) : Prop :=
    match ctmn cmd with
      | tmn_call _ nfids _ _ =>
        set_forall
          (fun nfid =>
             (fid,bid) !\in ((callfroms acc)[nfid]!:Set![fidT*bidT]))
          nfids
      | _ => True
    end.

  Definition acc_correct_callfroms_func
             (fid: fidT) (fbody:funcT) : Prop := 
    forall bid cmd (Hcmd : MapsTo bid cmd (fcmds fbody)),
      acc_correct_callfroms_blk fid bid cmd.

  Definition acc_correct_callfroms_prog : Prop := 
    forall fid fbody (Hfbody : MapsTo fid fbody (pfuncs prog)),
      acc_correct_callfroms_func fid fbody.

  Definition acc_correct_callfroms_star : Prop :=
    forall (fid:fidT) (ci:fidT*bidT) (Hci : ci !\in ((callfroms acc)[fid]!:Set![fidT*bidT])),
      (ci.1 !\in ((callfroms_star acc)[fid]!:Set![fidT])) /\
      (le (A:=Set![fidT])
          (callfroms_star acc)[ci.1]!
          (callfroms_star acc)[fid]!).

  Definition acc_correct_acclocs_lv
             (fid:fidT) (l:lv) (mem:memT): Prop :=
    let alocs := set_to_powerset (acclocs acc)[fid]! in
    forall mem',
    eq
      (sem_expr_lv fid l (mrestrict_join (E:=valueT) alocs mem (Elt mem')))
      (sem_expr_lv fid l mem).
  
  Definition acc_correct_acclocs_lv_opt
    (fid:fidT) (opt_l:option lv) (mem:memT): Prop :=
    opt_l = None \/ 
    exists l, (opt_l = Some l /\ acc_correct_acclocs_lv fid l mem).

  Definition acc_correct_acclocs_expr
             (fid:fidT) (e:exprT) (mem:memT): Prop :=
    let alocs := set_to_powerset (acclocs acc)[fid]! in
    forall mem',
    eq
      (sem_expr fid e (mrestrict_join (E:=valueT) alocs mem (Elt mem')))
      (sem_expr fid e mem).

  Definition acc_correct_acclocs_exprs
             (fid:fidT) (es:list exprT) (mem:memT): Prop :=
    let alocs := set_to_powerset (acclocs acc)[fid]! in
    forall mem',
    eq
      (sem_exprs fid es (mrestrict_join (E:=valueT) alocs mem (Elt mem')))
      (sem_exprs fid es mem).

  Definition localized (f:memT -> memT) (locs: \powerset locT) (mem:memT) : Prop :=
    forall mem',
      le
        (f (mrestrict_join (E:=valueT) locs mem mem'))
        (mrestrict_join (E:=valueT) locs (f mem) mem').

  Definition acc_correct_acclocs_instr
             (fid:fidT) (instr:instrT) (mem:memT) : Prop :=
    let alocs := set_to_powerset (acclocs acc)[fid]! in
    localized (sem_instr acc fid instr) alocs mem.

  Program Definition acc_correct_acclocs_tmn
             (fid:fidT) (tmn:tmnT) (mem:memT): Prop :=
    match tmn with
      | tmn_jump _ => True
      | tmn_call _ _ es ret => 
        acc_correct_acclocs_exprs fid es mem /\
        (if ret is Some ret'
         then (var_loc fid ret' nil) !\in (acclocs acc)[fid]!
         else true)
      | tmn_exit e => acc_correct_acclocs_expr fid e mem
    end.

  Definition acc_correct_acclocs_blk
             (fid:fidT) (bid:bidT) (stack:list (fidT*bidT))
             (cmd:cmdT) (state:stateT) : Prop :=
    let instrs := cinstrs cmd in
    let tmn := ctmn cmd in
    let instrs_mem : memT := state[(fid,bid,false,stack)]! in
    let tmn_mem : memT := state[(fid,bid,true,stack)]! in
    acc_correct_acclocs_tmn fid tmn tmn_mem.

  Definition acc_correct_acclocs_func
             (fid: fidT) (fbody:funcT) (state:stateT): Prop :=
    let alocs := mlookup (Key:=fidT) (Elt:=Set! [locT])
      fid (acclocs acc) in
    (forall bid cmd stack (Hcmd: MapsTo bid cmd (fcmds fbody))
            (Hpidx: pidx_consistent prog fid stack),
       acc_correct_acclocs_blk fid bid stack cmd state) /\
    (forall i (Hi : List.In i (fparams fbody)), ((inr (fid, i)), nil) !\in alocs).

  Definition acc_correct_acclocs_prog (state:stateT): Prop :=
    forall fid fbody (Hfbody : MapsTo fid fbody (pfuncs prog)),
      acc_correct_acclocs_func fid fbody state.

  Definition acc_correct_acclocs_callfroms : Prop :=
    forall (fid:fidT) (ci:fidT*bidT) (Hci: ci !\in ((callfroms acc)[fid]! : Set![fidT*bidT])),
      le (A:=Set![locT]) (acclocs acc)[fid]! (acclocs acc)[ci.1]!.

  Definition acc_correct (state:stateT) : Prop :=
    [/\ acc_correct_callfroms_prog,
        acc_correct_callfroms_star,
        acc_correct_acclocs_prog state &
        acc_correct_acclocs_callfroms].

  Lemma acc_correct_consistent (fid:fidT) (stack:list (fidT*bidT))
        (Hcf : acc_correct_callfroms_prog)
        (Hcs : acc_correct_callfroms_star)
        (Hconsistent : pidx_consistent prog fid stack)
        (Hnr : not_recursive acc fid) :
    List.find (fun fb => eqb fid fb.1) stack = None.
  Proof.
    move: Hconsistent Hnr; rewrite/not_recursive.
    move: {2 4}fid=> ofid.
    elim: stack fid=> [|[pfid pbid] stack IH] cfid Hc Hofid; first done.
    move: Hc; rewrite simpl_pidx_consistent.
    case Hfunc: ((pfuncs prog)[pfid])=> [func|]; last done.
    apply MF.find_mapsto_iff in Hfunc.
    case Hcmd: ((fcmds func)[pbid])=> [[instrs tmn]|]; last done.
    apply MF.find_mapsto_iff in Hcmd.
    case: tmn Hcmd=> //= nbid nfid args retto Hcmd.
    move/andP=> [Hnfid Hc].
    move: {Hcf} (Hcf _ _ Hfunc _ _ Hcmd)=> Hcf.
    move/set_forallP in Hcf.
    exploit (Hcf _ cfid); eauto.
    move=> {Hcf} Hcf.
    rewrite/acc_correct_callfroms_blk/ctmn/snd in Hcf.
    case Heqb: (eqb ofid pfid).
    { move/eqbP in Heqb.
      usual_tac Heqb.
      move: {Hcs} (Hcs _ _ Hcf)=> [Hcs _].
      by move/negP in Hofid; contradict Hofid.
    }
    { eapply IH; first by eauto.
      move/negP in Hofid; apply/negP.
      move=> Hofid'; apply: Hofid.
      move: {Hcs} (Hcs _ _ Hcf)=> [_ Hcs].
      apply SetFacts.mem_iff.
      apply Hcs.
      by apply SetFacts.mem_iff.
    }    
  Qed.

  (* about accessibilities *)

  Lemma acc_imply_call_subset_acclocs
    func fid bid instrs (nfid:fidT) (nfids:Set![fidT]) nbid args retto state
    (Hacc : acc_correct state)
    (Hfunc:(pfuncs prog) [fid] = Some func)
    (Hcmd:(fcmds func) [bid] = Some (instrs, tmn_call nbid nfids args retto))
    (Hnfid: nfid !\in nfids) :
    subset (set_to_powerset (acclocs acc) [nfid ]!) (set_to_powerset (acclocs acc) [fid ]!).
  Proof.
    move=> loc.
    rewrite 2! set_to_powerset_spec.
    move=> Hin.
    case: Hacc=> [Hcf _ _ Hac].
    apply MF.find_mapsto_iff in Hfunc.
    apply MF.find_mapsto_iff in Hcmd.
    move: {Hcf} (Hcf _ _ Hfunc _ _ Hcmd)=> Hcf.
    rewrite/acc_correct_callfroms_blk/ctmn/snd in Hcf.
    rewrite/acc_correct_acclocs_callfroms in Hac.
    apply SetFacts.mem_iff.
    move/set_forallP in Hcf.
    exploit (Hcf _ nfid); eauto.
    move=> {Hcf} Hcf.
    apply (Hac nfid (fid,bid) Hcf).
    by apply SetFacts.mem_iff.
  Qed.

  Lemma acc_imply_params_in_callee
    (i:idT)
    nfid nfunc (Hnfunc: MapsTo nfid nfunc (pfuncs prog))
    (params := fparams nfunc : list idT)
    (Hinparams:List.In i params)
    (state:stateT) (Hacc : acc_correct state) :
    var_loc nfid (local i) nil !\in (acclocs acc) [nfid ]!.
  Proof.
    rewrite/var_loc/local/llocal
           /pifid/pidxT_pidx/fst.
    case: Hacc=> [_ _ Ha _].
    move: {Ha} (Ha _ _ Hnfunc)=> [_ Ha].
    apply Ha.
    apply Hinparams.
  Qed.

  Lemma acc_imply_params_in_caller : forall 
    func fid bid instrs
    nfid nfunc (Hnfunc: MapsTo nfid nfunc (pfuncs prog))
    (nfids:Set![fidT]) (Hnfid: nfid !\in nfids)
    nbid args retto
    (params := fparams nfunc : list idT)
    (Hfunc:(pfuncs prog) [fid] = Some func)
    (Hcmd:(fcmds func) [bid] = Some (instrs, tmn_call nbid nfids args retto))
    (i:idT) (Hinparams:List.In i params)
    (state:stateT) (Hacc : acc_correct state),
    var_loc nfid (local i) nil
    !\in (acclocs acc) [fid ]!.
  Proof. 
    intros.
    assert (Hacclocs:subset 
                       (set_to_powerset (acclocs acc) [nfid ]!)
                       (set_to_powerset (acclocs acc) [fid ]!)).
    eapply acc_imply_call_subset_acclocs with 
      (func:=func) (bid:=bid) (instrs:=instrs)
      (nbid:=nbid) (args:=args) (retto:=retto) (state:=state); eauto.
    apply Hacclocs.
    eapply acc_imply_params_in_callee with (state:=state); eauto.
  Qed.
End Accs.
