Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.

Require Import LibTactics.
Require Import DomType.
Require Import DomBasic.
Require Import DomUtil.
Require Import DomItv.
Require Import DomSet.
Require Import DomMap.

Require Import Syntax.

Require Import DomAbsValue.
Require Import DomAcc.
Require Import DomAbsMem.

Require Import SemAbsBlock.
Require Import SemAccInsen.

Set Implicit Arguments.

Section Postfixs.
  Variable (prog:progT) (acc:accT).

  Definition postfix_jump (state:stateT) (fid:fidT) (func:funcT) (bid:bidT) (nbids:Set![bidT]) :=
    set_forall
      (fun nbid => leb (A:=memT) state[(fid,bid,true)]! state[(fid,nbid,false)]!)
      nbids.

  Definition postfix_call (state:stateT) (fid:fidT) (func:funcT) (bid:bidT) (nbid:bidT) (nfid:fidT) (args:list exprT) (retto:option varT) :=
    if (pfuncs prog)[nfid] is Some nfunc
    then let locs := (acclocs acc)[nfid]! in
         let locs' := (set_to_powerset locs) in
         let pidx := (fid,bid,true) in
         let npidx := (nfid,fentry nfunc,false) in
         let mem : memT := state[pidx]! in
         let argvals := sem_exprs fid args mem in
         let mem_in : memT := mrestrict (E:=valueT) locs' mem in
         let mem_out : memT := mrestrict (E:=valueT) (powerset_neg locs') mem in
         let mem_bound := sem_bind_params acc nfid argvals (fparams nfunc) mem_in in
         [&& (smem (S:=Set![fidT * bidT]) (fid,bid) (callfroms acc)[nfid]!),
             (leb mem_bound state[npidx]!) &
             (leb mem_out state[(fid,nbid,false)]!)]
    else true.

  Definition postfix_exit (state:stateT) (fid:fidT) (func:funcT) (bid:bidT) (ret:exprT) :=
    let pidx := (fid,bid,true) in
    let mem := state[pidx]! in
    let result := sem_expr fid ret mem in
    set_forall
      (fun (callinfo : fidT * bidT) =>
         if (pfuncs prog)[callinfo.1] is Some nfunc
         then if (fcmds nfunc)[callinfo.2] is Some cmd
              then match ctmn cmd with
                     | tmn_call nbid _ _ retto =>
                       let npidx := (callinfo.1, nbid, false) in
                       let mem_removed :=
                           if not_recursive acc fid
                           then mrestrict (E:=valueT) (filter_fid fid) mem
                           else mem
                       in
                       let mem_bound := sem_bind_ret acc callinfo.1 result retto mem_removed in
                       leb mem_bound state[npidx]!
                     | _ => true
                   end
              else true
         else true)
      (callfroms acc)[fid]!.

  Definition postfix_block (state:stateT) (fid:fidT) (func:funcT) (bid:bidT) (cmd:cmdT) :=
    let: (instrs, tmn) := cmd in
    (leb (sem_instrs acc fid instrs state[(fid,bid,false)]!) state[(fid,bid,true)]!) &&
    (match tmn with
       | tmn_jump nbids => postfix_jump state fid func bid nbids
       | tmn_call nbid nfids args retto =>
         set_forall
           (fun nfid => postfix_call state fid func bid nbid nfid args retto)
           nfids
       | tmn_exit ret => postfix_exit state fid func bid ret
     end).

  Definition postfix_func (state:stateT) (fid:fidT) (func:funcT) : bool :=
    map_forall
      (postfix_block state fid func)
      (fcmds func).
  
  Definition postfix (state:stateT) : bool :=
    map_forall
      (postfix_func state)
      (pfuncs prog).
End Postfixs.
