Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.
Require Import Setoid.
Require Import Morphisms.

Require Import DomType.
Require Import DomBasic.
Require Import DomSet.
Require Import DomMap.
Require Import DomPidx.
Require Import DomPidxSen.
Require Import DomAbsValue.
Require Import DomAcc.
Require Import DomAbsMem.

Require Import Syntax.

Require Import SemAbsBlock.
Require Import SemAccSen.

Require Import List.

Set Implicit Arguments.

Section Postfixs.
  Variable (prog:progT) (acc:accT).

  Inductive postfix_instrs (fid:fidT) : list instrT -> memT -> memT -> Prop :=
  | postfix_instrs_nil :
      forall mem1 mem2 (Hmem : le mem1 mem2),
        postfix_instrs fid nil mem1 mem2
  | postfix_instrs_cons :
      forall istr istrs mem1 mem2 mem3
             (Hsem : le (sem_instr acc fid istr mem1) mem2)
             (Hpostfix : postfix_instrs fid istrs mem2 mem3),
        postfix_instrs fid (istr::istrs) mem1 mem3.

  Definition postfix_jump
             (state:stateT) (fid:fidT) (func:funcT)
             (bid:bidT) (stack:list (fidT*bidT)) (nbids:Set![bidT]) :=
    forall nbid (Hnbid : smem (Elt:=bidT) nbid nbids),
      le (A:=memT)
         state[(fid,bid,true,stack)]!
         state[(fid,nbid,false,stack)]!.

  Definition postfix_call
             (state:stateT) (fid:fidT) (func:funcT)
             (bid:bidT) (stack:list (fidT*bidT)) (nbid:bidT) (nfid:fidT) (args:list exprT) :=
    forall nfunc (Hnfunc : MapsTo nfid nfunc (pfuncs prog)),
      let pidx := (fid,bid,true,stack) in
      let npidx := (nfid,fentry nfunc,false,(fid,bid)::stack) in
      let mem := state[pidx]! in
      let argvals := sem_exprs fid args mem in
      let mem_bound := sem_bind_params acc nfid argvals (fparams nfunc) mem in
      (le mem_bound state[npidx]!).

  Definition postfix_exit
             (state:stateT) (fid:fidT) (func:funcT)
             (bid:bidT) (stack:list (fidT*bidT)) (ret:exprT) :=
    let pidx := (fid,bid,true,stack) in
    let mem := state[pidx]! in
    let result := sem_expr fid ret mem in
    if stack is (ofid,obid)::stack'
    then forall nfunc (Hnfunc : MapsTo ofid nfunc (pfuncs prog)),
         forall cmd (Hcmd : MapsTo obid cmd (fcmds nfunc)),
           if ctmn cmd is tmn_call nbid _ _ retto
           then let opidx := (ofid, obid, false, stack') in
                let npidx := (ofid, nbid, false, stack') in
                let mem_removed :=
                    if not_recursive acc fid
                    then mrestrict (E:=valueT) (filter_fid fid) mem
                    else mem in
                let mem_bound := sem_bind_ret acc ofid result retto mem_removed in
                le mem_bound state[npidx]!
           else True
    else True.

  Definition postfix_block
             (state:stateT) (fid:fidT) (func:funcT)
             (bid:bidT) (stack:list (fidT*bidT)) (cmd:cmdT) :=
    let: (instrs, tmn) := cmd in
    postfix_instrs fid instrs state[(fid,bid,false,stack)]! state[(fid,bid,true,stack)]! /\
    (match tmn with
       | tmn_jump nbids => postfix_jump state fid func bid stack nbids
       | tmn_call nbid nfids args retto =>
         forall nfid (Hnfid: nfid !\in nfids),
           postfix_call state fid func bid stack nbid nfid args
       | tmn_exit ret => postfix_exit state fid func bid stack ret
     end).

  Definition postfix_func (state:stateT) (fid:fidT) (func:funcT) :=
    forall bid stack cmd (Hcmd : MapsTo bid cmd (fcmds func)),
      if pidx_consistent prog fid stack
      then postfix_block state fid func bid stack cmd
      else
        (eq (A:=memT) state[(fid,bid,false,stack)]! bot) /\
        (eq (A:=memT) state[(fid,bid,true,stack)]! bot).

  Definition postfix (state:stateT) :=
    forall fid func (Hfunc : MapsTo fid func (pfuncs prog)),
      postfix_func state fid func.
End Postfixs.
