Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.
Require Import Setoid.
Require Import Morphisms.

Require Import LibTactics.
Require Import DomType.
Require Import DomBasic.
Require Import DomBridge.

Require Import DomAbsValue.
Require Import DomSet.
Require Import DomMap.
Require Import DomAcc.
Require Import DomAbsMem.
Require Import DomAbsState.

Require Import Syntax.

Require Import SemAbsBlock.
Require Import SemAccInsen.
Require Import SemInsenLocal.

Require Import SemCon.
Require Import CorAbs.
Require Import CorCon.
Require Import CorConExpr.
Require Import CorAcc.
Require Import CorAccExpr.
Require Import CorBasic.
Require Import DomPidxSen.
Require Import DomPidx.

Set Implicit Arguments.


Inductive gamma_use_expr
          (pidx:DomPidxSen.pidxT) (lhs:DomConMem.memT) (rhs:DomAbsMem.memT): forall (e: exprT), Prop :=
| gamma_expr_const:
    forall c, gamma_use_expr pidx lhs rhs (expr_const c)
| gamma_expr_op2:
    forall o e1 e2
           (He1: gamma_use_expr pidx lhs rhs e1)
           (He2: gamma_use_expr pidx lhs rhs e2),
      gamma_use_expr pidx lhs rhs (expr_op2 o e1 e2)
| gamma_expr_op1:
    forall o e (He: gamma_use_expr pidx lhs rhs e),
      gamma_use_expr pidx lhs rhs (expr_op1 o e)
| gamma_expr_value_of:
    forall (lv:lv) (Hlv: gamma_use_lv pidx lhs rhs lv)
           (Hval: forall (cloc:DomConValue.locT) (Hcloc: SemCon.sem_lv pidx lv lhs cloc),
                    (lhs[cloc]!: option DomConValue.valueT) \abs rhs[t_loc cloc]!),
      gamma_use_expr pidx lhs rhs (expr_value_of lv)
| gamma_expr_addr_of:
    forall lv (Hlv: gamma_use_lv pidx lhs rhs lv),
      gamma_use_expr pidx lhs rhs (expr_addr_of lv)
| gamma_expr_cast:
    forall c e (He: gamma_use_expr pidx lhs rhs e),
      gamma_use_expr pidx lhs rhs (expr_cast c e)
| gamma_expr_size_of:
    forall e (He: gamma_use_expr pidx lhs rhs e),
      gamma_use_expr pidx lhs rhs (expr_size_of e)

with gamma_use_lv (pidx:DomPidxSen.pidxT) (lhs:DomConMem.memT) (rhs:DomAbsMem.memT): forall (l:lv), Prop :=
     | gamma_use_lv_id:
         forall x,
           gamma_use_lv pidx lhs rhs (lv_id x)
     | gamma_use_lv_deref:
         forall e (He: gamma_use_expr pidx lhs rhs e),
           gamma_use_lv pidx lhs rhs (lv_deref e)
     | gamma_use_lv_array:
         forall e1 e2
                (He1: gamma_use_expr pidx lhs rhs e1)
                (He2: gamma_use_expr pidx lhs rhs e2),
           gamma_use_lv pidx lhs rhs (Syntax.lv_array e1 e2)
     | gamma_use_lv_struct:
         forall l pos (He: gamma_use_lv pidx lhs rhs l),
           gamma_use_lv pidx lhs rhs (lv_struct l pos).

Inductive gamma_use_exprs (pidx:pidxT) (lhs:DomConMem.memT) (rhs:DomAbsMem.memT): forall (es: list exprT), Prop :=
| gamma_use_exprs_nil:
    gamma_use_exprs pidx lhs rhs nil
| gamma_use_exprs_cons:
    forall e (He: gamma_use_expr pidx lhs rhs e)
           es (Hes: gamma_use_exprs pidx lhs rhs es),
      gamma_use_exprs pidx lhs rhs (e::es).

Definition gamma_use_locs (pidx:pidxT) (lhs:DomConMem.memT) (rhs:DomAbsMem.memT) (e: exprT): Prop :=
  forall v (Hv: SemCon.sem_expr pidx e lhs v),
    if v is inl (inr loc)
    then loc \abs (locs_of (SemAbsBlock.sem_expr (pifid pidx) e rhs))
    else True.

Inductive gamma_use_locss (pidx:pidxT) (lhs:DomConMem.memT) (rhs:DomAbsMem.memT): forall (args:list exprT), Prop :=
| gamma_use_locss_nil:
    gamma_use_locss pidx lhs rhs nil
| gamma_use_locss_cons:
    forall e (He: gamma_use_locs pidx lhs rhs e)
           es (Hes: gamma_use_locss pidx lhs rhs es),
      gamma_use_locss pidx lhs rhs (e::es).

Inductive gamma_use_tmn (pidx:pidxT) (lhs:DomConMem.memT) (rhs:DomAbsMem.memT): forall (tmn:tmnT), Prop :=
| gamma_use_tmn_jump:
    forall bs,
      gamma_use_tmn pidx lhs rhs (tmn_jump bs)
| gamma_use_tmn_call:
    forall nf nbs args (ret:option varT)
           (Hargs: gamma_use_exprs pidx lhs rhs args)
           (Hret:
              if ret is Some ret'
              then (lhs[SemCon.var_loc ret' pidx]!: option DomConValue.valueT) \abs rhs[(DomAbsValue.var_loc (pifid pidx) ret' nil)]!
              else True),
      gamma_use_tmn pidx lhs rhs (tmn_call nf nbs args ret)
| gamma_use_tmn_exit:
    forall e (He: gamma_use_expr pidx lhs rhs e),
      gamma_use_tmn pidx lhs rhs (tmn_exit e).

Inductive gamma_use_instr
          (acc:accT) (pidx:pidxT) (lhs:DomConMem.memT) (rhs:DomAbsMem.memT): forall (instr:instrT), Prop :=
| gamma_use_instr_assign:
    forall lv (Hlv: gamma_use_lv pidx lhs rhs lv)
           e (He: gamma_use_expr pidx lhs rhs e),
      gamma_use_instr acc pidx lhs rhs (instr_assign lv e)
| gamma_use_instr_alloc:
    forall lv sz str site
           (Hlv: gamma_use_lv pidx lhs rhs lv)
           (Hsz: gamma_use_expr pidx lhs rhs sz)
           (Hstr: gamma_use_expr pidx lhs rhs str),
      gamma_use_instr acc pidx lhs rhs (instr_alloc lv (alloc_arr sz str) site)
| gamma_use_instr_assume:
    forall e (He: gamma_use_expr pidx lhs rhs e),
      gamma_use_instr acc pidx lhs rhs (instr_assume e)
| gamma_use_instr_forget:
    forall aloc,
      gamma_use_instr acc pidx lhs rhs (instr_forget aloc)
| gamma_use_instr_skip:
    gamma_use_instr acc pidx lhs rhs instr_skip
| gamma_use_instr_libcall:
    forall ol lc args site
           (Hlvalue:
              match lc with
                | scanf => True
                | strcpy =>
                  (forall ldst (Hldst: SemCon.sem_expr pidx (hd (expr_const lit_bot) args) lhs (inl (inr ldst)))
                          lsrc (Hlsrc: SemCon.sem_expr pidx (hd (expr_const lit_bot) (tl args)) lhs (inl (inr lsrc)))
                          lsrcValue (Hlval: eq (Some lsrcValue) (lhs[lsrc]!:option DomConValue.valueT))
                          lnewMem (Hlmem: eq lnewMem (lhs[ldst <- Some lsrcValue]!:DomConMem.memT))
                          (Hwf: DomConMem.mem_wf pidx lnewMem),
                     
                     let rdst := locs_of (SemAbsBlock.sem_expr (pifid pidx) (hd (expr_const lit_bot) args) rhs) in
                     let rsrc := locs_of (SemAbsBlock.sem_expr (pifid pidx) (hd (expr_const lit_bot) (tl args)) rhs) in
                     let rsrcValue := memory_lookup rsrc rhs in
                     let rnewMem := memory_update acc rdst rsrcValue rhs in
                     if ol is Some lv
                     then gamma_use_lv pidx lnewMem rnewMem lv
                     else True)
                | _ =>
                  if ol is Some lv
                  then gamma_use_lv pidx lhs rhs lv
                  else True
              end)
           (Hargs:
              match lc with
                | strcpy => gamma_use_exprs pidx lhs rhs args
                | scanf => gamma_use_exprs pidx lhs rhs args
                | builtin_alloca => gamma_use_exprs pidx lhs rhs args
                | builtin_choose_expr => gamma_use_exprs pidx lhs rhs args
                | _ => True
              end)
           (Hargs':
              match lc with
                | strcpy => gamma_use_locss pidx lhs rhs args
                | _ => True
              end),
      gamma_use_instr acc pidx lhs rhs (instr_libcall ol lc args site).

Inductive gamma_use_instrs (acc:accT) (pidx:pidxT) (lhs:DomConMem.memT) (rhs: memT): forall (instrs: list instrT), Prop :=
| gamma_use_instrs_nil:
    gamma_use_instrs acc pidx lhs rhs nil
| gamma_use_instrs_cons:
    forall i istrs
           (Hi: gamma_use_instr acc pidx lhs rhs i)
           (Histrs: forall cmem (Hcmem: SemCon.sem_instr pidx i lhs cmem),
                      gamma_use_instrs acc pidx cmem (SemAbsBlock.sem_instr acc (pifid pidx) i rhs) istrs),
      gamma_use_instrs acc pidx lhs rhs (i::istrs).

Definition gamma_use_blk (acc:accT) (pidx:pidxT) (cmd:cmdT)
           (lin: DomConMem.memT) (rin: memT): Prop :=
  let instrs := cinstrs cmd in
  let tmn := ctmn cmd in
  gamma_use_instrs acc pidx lin rin instrs /\
  forall lout (Hlout: SemCon.sem_instrs pidx instrs lin lout),
    gamma_use_tmn pidx lout (SemAbsBlock.sem_instrs acc (pifid pidx) instrs rin) tmn.


Section GammaUseSpecs.
  Variable (prog:progT) (acc:accT).

  Require Import DomZ.
  Require Import DomSet.
  Require Import DomConValue.
  Require Import DomConMem.

  Require Import SemCon.
  Require Import DomConValue.
  Require Import DomAbsValue.
  Require Import DomAcc.
  Require Import DomAbsMem.
  Require Import DomPidx.
  Require Import DomPidxSen.

  Require Import SemAbsBlock.
  Require Import SemAccSen.
  Require Import SemSen.

  Require Import CorBasic.
  Require Import CorAbs.
  Require Import CorItv.

  Lemma gamma_use_expr_proper2_ : forall (pidx : pidxT) (e : exprT)
    (cm : DomConMem.memT) 
    (v : DomConValue.valueT) (Hv : SemCon.sem_expr pidx e cm v)
    (am : memT) (Hm: gamma_use_expr pidx cm am e),
    v \abs SemAbsBlock.sem_expr (pifid pidx) e am
  with gamma_use_lv_proper2_ : forall (pidx : pidxT) (l : lv)
    (m : DomConMem.memT) (am : memT) (Hm: gamma_use_lv pidx m am l)
    (con_loc : DomConValue.locT) (Hsem_lv : sem_lv pidx l m con_loc),
    con_loc \abs sem_expr_lv (pifid pidx) l am.
  Proof.
    (* sem_expr *)
    intros pidx e cm v Hv.
    induction Hv; intros; rewrite/vint/vheap/vloc.

    { inv Hm.
      inv Hlit; rewrite/sem_expr/=.
      { by constructor; omega. }
      { by constructor. }
      move: v=> [[v|v]|[[a u] o]]; [exact: abs_itvT'_top|done|].
      by do ? split; apply: abs_itvT'_top.
    }      
    { inv Hm.
      exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_add_proper2.
    }
    { inv Hm.
      exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      rewrite/sem_expr-/sem_expr.
      move: (sem_expr _ _ _) (sem_expr _ _ _) IHHv1 IHHv2.
      move=> [[i1 a1] s1] [[i2 a2] s2].
      rewrite/vop2.
      rewrite/abs_relation/value_abs/itv_of/arrv_of/fst/snd.
      move=> Habs Habs'.
      rewrite/vint/vheap in Habs Habs'.
      eapply abs_relation_proper; last by eapply arrv_op_add_spec; eauto.
      reflexivity.
      move: (arrv_op_add _ _) (arrv_op_add _ _)=> [l|] [r|];
      try exact: top_prop.
      rewrite/= ? map_mlookup_spec map_join_spec.
      apply join_rhs.
    }
    { inv Hm.
      exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      rewrite/sem_expr-/sem_expr.
      move: (sem_expr _ _ _) (sem_expr _ _ _) IHHv1 IHHv2.
      move=> [[i1 a1] s1] [[i2 a2] s2].
      rewrite/vop2.
      rewrite/abs_relation/value_abs/itv_of/arrv_of/fst/snd.
      move=> Habs Habs'.
      rewrite/vint/vheap in Habs Habs'.
      eapply abs_relation_proper; last by eapply arrv_op_add_spec; eauto.
      reflexivity.
      move: (arrv_op_add _ _) (arrv_op_add _ _)=> [l|] [r|]; try exact: top_prop.
      rewrite/= ? map_mlookup_spec map_join_spec.
      apply join_lhs.
    }
    { inv Hm.
      exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_sub_proper2.
    }
    { inv Hm.
      exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      rewrite/sem_expr-/sem_expr.
      move: (sem_expr _ _ _) (sem_expr _ _ _) IHHv1 IHHv2.
      move=> [[i1 a1] s1] [[i2 a2] s2].
      rewrite/vop2.
      rewrite/abs_relation/value_abs/itv_of/arrv_of/fst/snd.
      move=> Habs Habs'.
      by eapply arrv_op_sub_spec; eauto.
    }
    { inv Hm.
      exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_mul_proper2.
    }
    { inv Hm.
      exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_div_proper2.
    }
    { inv Hm.
      exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_mod_proper2.
    }
    { inv Hm.
      exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_lt_proper2.
    }
    { inv Hm.
      exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_le_proper2.
    }
    { inv Hm.
      exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_gt_proper2.
    }
    { inv Hm.
      exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_ge_proper2.
    }
    { inv Hm.
      exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_eq_proper2.
    }
    { inv Hm.
      exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_neq_proper2.
    }
    { inv Hm.
      exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_and_proper2.
    }
    { inv Hm.
      exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_or_proper2.
    }
    { inv Hm.
      exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_band_proper2.
    }
    { inv Hm.
      exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_bor_proper2.
    }
    { inv Hm.
      exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_bxor_proper2.
    }
    { inv Hm.
      exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_lshift_proper2.
    }
    { inv Hm.
      exploit IHHv1; eauto=> {IHHv1} IHHv1.
      exploit IHHv2; eauto=> {IHHv2} IHHv2.
      by apply op_rshift_proper2.
    }
    { inv Hm.
      exploit IHHv; eauto=> {IHHv} IHHv.
      by apply op_not_proper2.
    }
    { inv Hm.
      exploit IHHv; eauto=> {IHHv} IHHv.
      by apply op_bnot_proper2.
    }
    { inv Hm.
      replace (sem_expr (pifid pidx) (expr_value_of l) am) with
        (memory_lookup (sem_expr_lv (pifid pidx) l am) am);
        last reflexivity.
      rewrite/memory_lookup.
      assert (lval \abs sem_expr_lv (pifid pidx) l am).
      apply gamma_use_lv_proper2_ with m; assumption.
      remember (sem_expr_lv (pifid pidx) l am). destruct l0;
      last apply abs_value_top.
      assert (SetInterface.mem (t_loc lval) s); first apply H.
      exploit Hval; eauto.
      rewrite Hlookup.
      move: (t_loc lval) H0=> aloc Hmem Hv.
      rewrite SetInterface.fold_1.
      rewrite <- fold_left_rev_right.
      rewrite SetFacts.elements_b in Hmem.
      rewrite existsb_rev in Hmem.
      elim: (rev (SetInterface.elements s)) Hmem=> [|ll llocs IH] Hmem; first done.
      rewrite/= in Hmem.
      unfold fold_right; fold fold_right.
      case Hll: (aloc == ll) in Hmem.
      { rewrite <- join_lhs.
        rewrite is_compare_eqb in Hll.
        move/eqbP in Hll.
        usual_tac Hll.
        by apply Hv.
      }
      { rewrite <- join_rhs.
        by apply IH.
      }
    }
    { inv Hm.
      replace (sem_expr (pifid pidx) (expr_addr_of l) am) with
        ((bot, sem_expr_lv (pifid pidx) l am, bot):valueT);
        last reflexivity.
      apply gamma_use_lv_proper2_ with m; auto. }
    { inv Hm.
      replace (sem_expr (pifid pidx) (expr_cast t e) am) with
      (cast t (sem_expr (pifid pidx) e am));
      last reflexivity.
      apply cast_sound. assumption.
      apply IHHv. assumption. }
    { inv Hm0.
      replace (sem_expr (pifid pidx) (expr_cast t e) am) with
      (cast t (sem_expr (pifid pidx) e am));
      last reflexivity.
      apply cast_sound_not with z. assumption.
      assumption.
      apply IHHv. assumption. }
    { inv Hm.
      replace (sem_expr (pifid pidx) (expr_cast (typ_pointer t) e) am) with 
      (cast (typ_pointer t) (sem_expr (pifid pidx) e am));
      last reflexivity.
      apply cast_sound_array. assumption.
      apply IHHv. assumption. }
    { exact: abs_itvT'_top. }

  (* lv *)
    intros pidx l m am Hv con_loc Hsem_lv.
    induction Hsem_lv; intros; inv Hv.
    { simpl. unfold lglobal.
      rewrite SetFacts.singleton_b. 
      rewrite DomBridge.is_compare_eqb.
      apply/eqbP. reflexivity. }
    { simpl. unfold llocal.
      rewrite SetFacts.singleton_b. 
      rewrite DomBridge.is_compare_eqb.
      apply/eqbP. reflexivity. }
    { replace (sem_expr_lv (pifid pidx) (lv_deref e) am) with 
      (let v := sem_expr (pifid pidx) e am in 
        locs_of v ⊔ locs_from_arrvT (arrv_of v)); 
      last reflexivity.
      assert ((vloc l) \abs sem_expr (pifid pidx) e am).
      apply gamma_use_expr_proper2_ with m; assumption.
      rewrite/abs_relation/value_abs in H.
      move:H. apply abs_relation_proper. reflexivity.
      apply join_lhs_.
    }
    { replace (sem_expr_lv (pifid pidx) (lv_deref e) am) with
      (let v := sem_expr (pifid pidx) e am in 
        locs_of v ⊔ locs_from_arrvT (arrv_of v)); 
      last reflexivity.
      assert ((vheap al uq (ofs, sz, st)) \abs sem_expr (pifid pidx) e am).
      apply gamma_use_expr_proper2_ with m; assumption.
      rewrite/abs_relation/value_abs in H.
      rewrite/ locs_from_arrvT.
      destruct (arrv_of (sem_expr (pifid pidx) e am)).
      unfold lheap.

      assert((DomConValue.lheap al uq ofs) \abs
        (Elt (fold
          (λ (k : idT) (v : ossT) (s : SetInterface.set locT),
            if ~~ eqb v bot then
              SetInterface.add (lheap k nil) s else s) a
          SetInterface.empty):locsT)).
      { apply collect_key_abs with sz st; assumption. }

      move:H0. apply abs_relation_proper. reflexivity.
      apply join_rhs_.
      
      rewrite join_top_rhs. constructor.
    }
    { replace (sem_expr_lv (pifid pidx) (Syntax.lv_array arr idx) am) with
      (let v := sem_expr (pifid pidx) arr am in locs_from_arrvT (arrv_of v));
      last reflexivity.
      rewrite/locs_from_arrvT.
      assert ((DomConValue.vheap al uq (ofs, sz, st))
                \abs sem_expr (pifid pidx) arr am).
      apply gamma_use_expr_proper2_ with m; assumption.
      remember (sem_expr (pifid pidx) arr am).
      move: v Heqv H=> [[i a] s] Heqv H.
      rewrite/arrv_of/fst/snd. destruct s.

      assert ((DomConValue.lheap al uq ofs) \abs
        (Elt (fold
          (λ (k : idT) (v : ossT) (s : Set! [locT]),
            if ~~ eqb v bot then SetInterface.add (lheap k nil) s else s) a0
          SetInterface.empty) : locsT)).
      apply collect_key_abs with sz st; auto. apply H0.

      move:H. rewrite/abs_relation/loc_abs/t_loc. done.
    }
    { unfold sem_expr_lv; fold sem_expr_lv.
      move: (sem_expr_lv _ _ _) (IHHsem_lv He)=> [alocs|] Hlocs; last done.
      clear Hsem_lv IHHsem_lv.
      move: loc Hlocs=> [loc poses] Hlocs.
      rewrite/abs_relation/loc_abs/t_loc/fst/snd in Hlocs *.
      replace
        (match loc with
           | inl (inl (a, _, _)) => lheap a (pos :: poses)
           | inl (inr gvar) => lglobal gvar (pos :: poses)
           | inr (fid, bid, _) => llocal fid bid (pos :: poses)
         end)
        with
        (proj1_sig (append_field pos)
                   match loc with
                     | inl (inl (a, _, _)) => lheap a (poses)
                     | inl (inr gvar) => lglobal gvar (poses)
                     | inr (fid, bid, _) => llocal fid bid (poses)
                   end); last first.
      { rewrite/append_field/=.
        by move: loc Hlocs=> [[[[a u] o]|loc]|[[fvar vvar] dvar]] Hlocs.
      }
      by apply DomSet.mapP.
    }
  Qed.
End GammaUseSpecs.
