Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.
Require Import List.
Require Import Setoid.
Require Import Morphisms.

Require Import LibTactics.
Require Import DomType.
Require Import DomBasic.
Require Import DomSet.
Require Import DomMap.
Require Import DomUtil.
Require Import DomAbsValue.
Require Import DomAcc.
Require Import DomAbsMem.
Require Import DomPidx.
Require Import DomPidxSen.
Require Import DomPidxInsen.

Require Import Syntax.

Require Import SemAbsBlock.
Require Import SemAccSen.
Require Import SemAccInsen.

Require Import SemSenLocal.
Require Import SemInsenLocal.

Require Import CorBasic.

Set Implicit Arguments.

Section Tstates.
  Variable (prog:progT) (acc:accT).

  Program Definition tstate (state:stateT) : SemAccSen.stateT :=
    fun pidx =>
      if pidx_consistent prog (pifid pidx) pidx.2
      then state[pidx.1]!
      else bot.
  Next Obligation.
    move => x y Hxy; usual_tac Hxy.
    move: y=> [[[fid bid] epilogue] stack] /=.
    exact: eq_refl.
  Qed.

  Add Parametric Morphism : tstate with
      signature eq ==> eq as tstate_mor.
  Proof.
    move=> x y Hxy [[[fid bid] epilogue] stack].
    rewrite/tstate/=/fst/snd; dom_tac.
    case: (pidx_consistent _ _); last reflexivity.
    rewrite Hxy.
    reflexivity.
  Qed.

  Lemma tstate_lookup (state:stateT) (pidx:DomPidxSen.pidxT) :
    (@mlookup _ _ _ _ _ _ _ _ _ SemAccSen.stateT_maplike pidx (tstate state)) =
    if pidx_consistent prog (pifid pidx) pidx.2
    then state[pidx.1]!
    else bot.
  Proof. done. Qed.
End Tstates.
