exception TODO
exception TrError of string
exception ValiError of string

(************)
(* validate *)
(************)

let eval_top_to_top (inputof:Dom.Table.t) (g:Dom.global)
  (opts:Options.t) (pi:Pre.preinfo) : Dom.global =
  let ograph = g.Dom.graph in
  let ocfgs = ograph.M5graph.cfgs in
  
  let exp_eval_top_to_top (e:Syn.exp) (mem:Dom.Mem.t) : Syn.exp =
    if (Eval.eval e mem = Dom.Value.top) 
    then Syn.Econst Syn.Ctop
    else e
  in

  let rec fact_eval_top_to_top (f:Syn.fact) (mem:Dom.Mem.t) : Syn.fact =
    match f with
    | Syn.Pbin (r,e1,e2) -> 
      Syn.Pbin (r,exp_eval_top_to_top e1 mem,exp_eval_top_to_top e2 mem)
    | Syn.Pnot f1 -> 
      Syn.Pnot (fact_eval_top_to_top f1 mem)
  in

  let block_eval_top_to_top (st:Dom.State.t) (blk:(Cfg.bcmd * Syn.tag) list) (ctx:M5graph.ctx) 
    : (Cfg.bcmd * Syn.tag) list =
    let (_, res_blk) =
      List.fold_left
        (fun (acc_mem, acc_cmds) (cmd, tag) ->
          let new_mem = Semantics.react opts pi ctx tag cmd g acc_mem in
            match cmd with
              | Cfg.SET (lv, e) -> 
                (new_mem,
                 acc_cmds @
                   [(Cfg.SET (lv,exp_eval_top_to_top e acc_mem), tag)])
              | Cfg.ASSUME p ->
                (new_mem,
                 acc_cmds @ 
                   [(Cfg.ASSUME (fact_eval_top_to_top p acc_mem), tag)])
              | Cfg.ASSERT p ->
                (new_mem,
                 acc_cmds @ 
                   [(Cfg.ASSERT (fact_eval_top_to_top p acc_mem), tag)])
              | _ -> (new_mem, acc_cmds @ [(cmd, tag)])
        )
        (Dom.State.join_all_mems st, [])
        blk
    in
    res_blk
  in

  let blocks_eval_top_to_top (inputof:Dom.Table.t) (ctx:Syn.pid) (blks:(Cfg.node, Cfg.block) PMap.t)
    : (Cfg.node, Cfg.block) PMap.t =
    PMap.mapi
      (fun nd blk ->
        match blk with
          | Cfg.CALL _ -> blk
          | Cfg.BLOCK blk' -> try (
            let st = PMap.find (ctx, nd) inputof.Dom.Table.table in
              Cfg.BLOCK (block_eval_top_to_top st blk' ctx)
            )
	    with Not_found -> blk
      )
      blks
  in

  let mcfgs =
    PMap.mapi
      (fun ctx cfg -> { cfg with Cfg.blocks =
                    (blocks_eval_top_to_top inputof ctx cfg.Cfg.blocks) } )
      ocfgs in
  let mgraph = { ograph with M5graph.cfgs = mcfgs } in
  { g with Dom.graph = mgraph }

let toploc_assign_to_nop (inputof:Dom.Table.t) (g:Dom.global)
  (opts:Options.t) (pi:Pre.preinfo) : Dom.global =
  let ograph = g.Dom.graph in
  let ocfgs = ograph.M5graph.cfgs in
  
  let block_toploc_assign_to_nop (st:Dom.State.t) (blk:(Cfg.bcmd * Syn.tag) list) (ctx:M5graph.ctx) 
    : (Cfg.bcmd * Syn.tag) list =
    let (_, res_blk) =
      List.fold_left
        (fun (acc_mem, acc_cmds) (cmd, tag) ->
          let new_mem = Semantics.react opts pi ctx tag cmd g acc_mem in
            match cmd with
              | Cfg.SET (lv, _) -> (
                if (Eval.evalLocs lv acc_mem = Dom.Address.Top) then
                  (new_mem, acc_cmds @ [(Cfg.NOP, tag)])
                else
                  (new_mem, acc_cmds @ [(cmd, tag)])
              )
              | _ -> (new_mem, acc_cmds @ [(cmd, tag)])
        )
        (Dom.State.join_all_mems st, [])
        blk
    in
      res_blk
  in

  let blocks_toploc_assign_to_nop (inputof:Dom.Table.t) (ctx:Syn.pid) (blks:(Cfg.node, Cfg.block) PMap.t)
    : (Cfg.node, Cfg.block) PMap.t =
    PMap.mapi
      (fun nd blk ->
        match blk with
          | Cfg.CALL _ -> blk
          | Cfg.BLOCK blk' -> try (
            let st = PMap.find (ctx, nd) inputof.Dom.Table.table in
              Cfg.BLOCK (block_toploc_assign_to_nop st blk' ctx)
            )
	    with Not_found -> blk
      )
      blks
  in

  let mcfgs =
    PMap.mapi
      (fun ctx cfg -> { cfg with Cfg.blocks =
                    (blocks_toploc_assign_to_nop inputof ctx cfg.Cfg.blocks) } )
      ocfgs in
  let mgraph = { ograph with M5graph.cfgs = mcfgs } in
  { g with Dom.graph = mgraph }

let validate_intra_edges
    (pos2str: (int, string) PMap.t) 
    (prog: Syntax.progT) (acc: DomAcc.accT) (fid: DomBasic.fidT) (func: Syntax.funcT)
    (propagated_func_state: DomAbsState.func_stateT)
    (orig_state: DomAbsState.func_stateT)
    : bool =
  (&&) (if SemAccInsen.acc_correct_callfroms_func acc fid func
    then true
    else (prerr_endline "validate_intra failed: acc_correct_callfroms_func"; false))
    ((&&) (if Main0.acc_correct_acclocs_star_func acc fid
      then true
      else (prerr_endline "validate_intra failed: acc_correct_acclocs_star_func"; false))
        ((&&) (if Main0.acc_correct_acclocs_func pos2str acc fid func propagated_func_state
          then true
          else (prerr_endline "validate_intra failed: acc_correct_acclocs_func"; false))
            ((&&) (if Main0.acc_correct_acclocs_callfroms_func acc fid
              then true
              else (prerr_endline "validate_intra failed: acc_correct_acclocs_callfroms_func"; false))
                ((&&) (if Main0.postfix_func pos2str prog acc propagated_func_state fid func
                  then true
                  else (prerr_endline "validate_intra failed: postfix_func"; false))
                    (if Main0.use_le_func acc fid func propagated_func_state orig_state
                     then true
                     else (prerr_endline "validate_intra faile: use_le_func"; false))))))

(* validate_inter_edges validates inter-edges.  The state of its
   inputs is a translated and propagated states, but it only includes
   some memories those are directly related to inter-procedural edges,
   for example,

   input memory of entry node / output memory of exit node / output
   memory of call node / input memory of return node.
*)
let validate_inter_edges
    (pos2str: (int, string) PMap.t) 
    (prog: Syntax.progT) (acc: DomAcc.accT)
    (propagated_state: DomAbsState.stateT')
    : bool =
  (DomMap.map_forall 
     DomBasic.positive_setoid DomBasic.positive_poset
     DomBasic.positive_setoidDec DomBasic.positive_posetDec
     (fun fid func ->
       (Main0.postfix_inter_func pos2str prog acc propagated_state fid func))
     (Syntax.pfuncs prog))
  &&
    (DomMap.map_forall 
       DomBasic.positive_setoid DomBasic.positive_poset
       DomBasic.positive_setoidDec DomBasic.positive_posetDec 
       (fun fid x -> Ssrbool.isSome
         (MapInterface.find
            (DomBridge.order_OrderedType 
               DomBasic.positive_setoid DomBasic.positive_poset
               DomBasic.positive_setoidDec DomBasic.positive_posetDec)
            (MapAVLInstance.coq_MapAVL_FMap
               (DomBridge.order_OrderedType
                  DomBasic.positive_setoid DomBasic.positive_poset
                  DomBasic.positive_setoidDec DomBasic.positive_posetDec))
            fid (Syntax.pfuncs prog)))
       acc.DomAcc.callfroms)
  && 
    (DomMap.map_forall
       DomBasic.positive_setoid DomBasic.positive_poset
       DomBasic.positive_setoidDec DomBasic.positive_posetDec 
       (fun x func_state -> Main0.notop func_state)
       propagated_state)

let validate (opts:Options.t) ((inputof:Dom.Table.t), (outputof:Dom.Table.t), (g:Dom.global),
  (du:Ssarua.t), (pre:Pre.preinfo), (pid2joinpoints:(Syn.pid,Ssarua.joinpoints) PMap.t))
    : unit =
  let _ = Profiler.reset () in

  (* deadcode elimination *)
  let _ = Profiler.start_event "validation-translation-dce" in
  let (inputof,outputof,g,du,pre) =
    Dce.deadcode_elimination inputof outputof g du opts pre in
  let _ = Profiler.finish_event "validation-translation-dce" () in

  (* temporary locations-related modification
     
     This generate a map of (node -> temporary_locs).  The meaning of
     the mapping is that some "temporary locations" are defined in a
     "node".  This map will be used to add "forget (loc)" statements
     after each program transitions.
  *)
  let _ = Profiler.start_event "validation-translation-node2temps" in
  let node2temps = TempLoc.get_node2temps opts pre g inputof in
  let _ = Profiler.finish_event "validation-translation-node2temps" () in

  (* substitute top-related statement to nop and top *)
  let _ = Profiler.start_event "validation-translation-top2nop" in
  let g = toploc_assign_to_nop inputof g opts pre in
  (* let g = eval_top_to_top inputof g opts pre in *)
  let _ = Profiler.finish_event "validation-translation-top2nop" () in

  (* generate str2pos and pos2str *)
  let _ = Profiler.start_event "validation-util-strpos" in
  let strs = StrCollector.unions 
    [ StrCollector.collect_strs_table inputof 
    ; StrCollector.collect_strs_table outputof 
    ; StrCollector.collect_strs_global g 
    ; StrCollector.collect_strs_ssarua du 
    ; StrCollector.collect_strs_pre pre ] in
  let strs = PSet.fold 
    (fun s acc -> 
      if Str0.exists s "->"
      then 
        let subs = Str0.splits s "->" in
        List.fold_left (fun acc x -> PSet.add x acc) acc subs
      else acc)
    strs strs in
  let (str2pos,pos2str) = StrCollector.gen_mapping_btwn_str_pos strs in
  let _ = Profiler.finish_event "validation-util-strpos" () in

  (* locals *)
  let locals = pre.Pre.locals in

  (* translate accessibility *)
  let _ = Profiler.start_event "validation-translation-acc" in
  let translated_acc = 
    StepManager.stepf false "translate accessibility" Translator.tr_access
      (locals, str2pos, g, pre)
  in
  let _ = Profiler.finish_event "validation-translation-acc" () in

  (* translate program *)
  let _ = Profiler.start_event "validation-translation-pgm" in
  let prog = StepManager.stepf false "translate program" Translator.tr_global
      (locals, str2pos, inputof, g) in
  let _ = Profiler.finish_event "validation-translation-pgm" () in

  (* all functions in the program *)
  let cfgs = g.Dom.graph.M5graph.cfgs in

  (* validate intra-procedural edges: it returns a map from pid to
     filtered-states that only include memory states of entry, exit,
     call, and return nodes.  The return value is used for the
     validation of inter-procedural edges. *)
  let filtered_states = PMap.mapi
    (fun p pcfg ->
      let fid_coq = PMap.find p str2pos in

      (* get translated cfg *)
      let translated_cfg =
        match Orders.intMapFind fid_coq (Syntax.pfuncs prog) with
        | Some cfg -> cfg
        | None -> raise (TrError ("translated_cfg: "^p^" is not translated."))
      in

      (* add "forget (temp_loc)" using node2temps *)
      let _ = Profiler.start_event "validation-translation-addforget" in
      let translated_cfg = TempLoc.add_forget 
        locals str2pos pos2str p translated_cfg node2temps 
      in
      let _ = Profiler.finish_event "validation-translation-addforget" () in

      (* get translated func_state *)
      let translated_func_state =
        let _ = Profiler.start_event "validation-translation-states" in
        let nodes = Cfg.nodesof pcfg in
        let translated_inputof = Translator.tr_table_fun locals str2pos p
          nodes false inputof in
        let result = Translator.tr_table_fun_w_init locals str2pos p
          nodes true outputof translated_inputof in
        let _ = Profiler.finish_event "validation-translation-states" () in
          result
      in

      (* propagation and related validation: inputof and outputof *)
      (* profilers are equipped inside of propagate function *)
      let (propagated_state_p, p_validate_success) =
        Propagator.propagate locals str2pos pos2str 
          p (PMap.find p cfgs) g pre
          prog translated_cfg translated_func_state translated_acc
          du.Ssarua.dugraph pid2joinpoints inputof outputof
      in

      let _ =
        if (not p_validate_success)
        then (prerr_endline ("propagation validation failed: "^p))
        else ()
      in

      let _ = Profiler.start_event "validation-validation-intra" in
      let validation_result = 
        if not (p = "_G_")
        then validate_intra_edges pos2str prog translated_acc fid_coq 
          translated_cfg propagated_state_p translated_func_state
        else false
      in
      let _ = 
        if p = "_G_"
        then prerr_endline 
          "validation aborted: _G_ is not on the validation process"
        else if validation_result
        then prerr_endline ("validation succeeded(intra): "^p)
        else 
          prerr_endline ("validation failed(intra): "^p)
      in
      let _ = Profiler.finish_event "validation-validation-intra" () in

      (* This filters memories of entry,exit,call,return nodes from
         the propagated state.  The process is for the validation of
         inter-edges. *)
      let _ = Profiler.start_event "validation-validation-filter" in
      let nodes = Cfg.nodesof pcfg in
      let filtered_state =
        List.fold_left
          (fun filtered_state node ->
            if Cfg.is_entry node || Cfg.is_returnnode pcfg node
            then 
              (* add memory to filtered_state using propagated_state_p *)
              let bb_input = 
                Translator.tr_m5graph_node_bidBool str2pos false 
                  (M5graph.make_node p node)
              in
              let mem_input =  
                match Orders.bidBoolMapFind bb_input propagated_state_p with 
                | Some m -> m
                | None -> DomBasic.Elt Orders.locTMapEmpty
              in
              let filtered_state =
                Orders.bidBoolMapAdd bb_input mem_input filtered_state
              in
              filtered_state
            else if Cfg.is_callnode pcfg node || Cfg.is_next_exit pcfg node
            then 
              (* add memory to filtered_state using propagated_state_p *)
              let bb_output = 
                Translator.tr_m5graph_node_bidBool str2pos true 
                  (M5graph.make_node p node)
              in
              let mem_output =  
                match Orders.bidBoolMapFind bb_output propagated_state_p with 
                | Some m -> m
                | None -> DomBasic.Elt Orders.locTMapEmpty
              in
              let filtered_state =
                Orders.bidBoolMapAdd bb_output mem_output filtered_state
              in
              filtered_state
            else 
              filtered_state
          )
          Orders.bidBoolMapEmpty
          nodes
      in
      let _ = Profiler.finish_event "validation-validation-filter" () in
        filtered_state
    )
    cfgs
  in

  (* validate inter-procedural edges *)
  let _ = Profiler.start_event "validation-validation-inter" in
  let state_coq = 
    PMap.foldi
      (fun p func_state state_coq ->
        let fid = PMap.find p str2pos in
        Orders.intMapAdd fid func_state state_coq
      )
      filtered_states
      Orders.intMapEmpty
  in
  let validation_result = validate_inter_edges pos2str prog translated_acc
    state_coq 
  in
  let _ = 
    if validation_result
    then
      prerr_endline "validation succeeded(inter)"
    else 
      prerr_endline "validation failed(inter)"
  in
  let _ = Profiler.finish_event "validation-validation-inter" () in
    ()
