exception PrError of string

(*************)
(* propagate *)
(*************)

let valueTop : DomAbsValue.valueT = 
  DomType.top 
    DomAbsValue.valueT_setoid
    DomAbsValue.valueT_poset
    DomAbsValue.valueT_hasTop

let valueBot : DomAbsValue.valueT = 
  DomType.bot
    DomAbsValue.valueT_setoid
    DomAbsValue.valueT_poset
    DomAbsValue.valueT_hasBot

let project 
    (pos2str: (int, string) PMap.t)
    (mem:DomAbsMem.memT) (locs:DomAbsValue.locT PSet.t) 
    : DomAbsMem.memT =

  match mem with 
  | DomBasic.Top -> 
    DomBasic.Elt
      (PSet.fold
         (fun l acc ->
           Orders.locTMapAdd l valueTop acc
         )
         locs
         Orders.locTMapEmpty
      )
  | DomBasic.Elt mem' ->
    DomBasic.Elt
      (Orders.locTMapFold
         (fun l v acc ->
           if PSet.mem l locs
           then 
             Orders.locTMapAdd l v acc
           else 
             acc
         )
         mem'
         Orders.locTMapEmpty
      )

(* The cut function cuts memory locations of locs from mem. *)
let cut
    (mem:DomAbsMem.memT) (locs:DomAbsValue.locT PSet.t) 
    : DomAbsMem.memT =
  match mem with 
  | DomBasic.Top -> DomBasic.Top
  | DomBasic.Elt mem' ->
    DomBasic.Elt
      (PSet.fold (fun l acc -> Orders.locTMapRemove l acc) locs mem')

let locs_coq2ocaml (locs_coq:DomAbsValue.locT SetInterface.set)
    : DomAbsValue.locT PSet.t = 
  Orders.locTSetFold 
    (fun loc_coq acc -> PSet.add loc_coq acc)
    locs_coq
    PSet.empty

let cut_coq
    (mem:DomAbsMem.memT) (locs_coq:DomAbsValue.locT SetInterface.set) 
    : DomAbsMem.memT =
  cut mem (locs_coq2ocaml locs_coq)

let valueJoin (v1:DomAbsValue.valueT) (v2:DomAbsValue.valueT) 
    : DomAbsValue.valueT = 
  DomType.join
    DomAbsValue.valueT_setoid
    DomAbsValue.valueT_poset
    DomAbsValue.valueT_hasJoin
    v1 v2

let memJoin' 
    (mem1:(DomAbsValue.locT, DomAbsValue.valueT) MapInterface.dict)
    (mem2:(DomAbsValue.locT, DomAbsValue.valueT) MapInterface.dict)
    : (DomAbsValue.locT, DomAbsValue.valueT) MapInterface.dict =
  Orders.locTMapFold
    (fun l v2 acc ->
      let v = 
        match Orders.locTMapFind l acc with 
        | Some v1 -> valueJoin v1 v2
        | None -> v2
      in
      Orders.locTMapAdd l v acc 
    )
    mem2
    mem1

let memJoin (mem1:DomAbsMem.memT) (mem2:DomAbsMem.memT) 
    : DomAbsMem.memT =
  match mem1 with 
  | DomBasic.Top -> mem1
  | DomBasic.Elt mem1' ->
    (match mem2 with 
    | DomBasic.Top -> mem2
    | DomBasic.Elt mem2' -> 
      if (Orders.locTMapCardinal mem1' >= Orders.locTMapCardinal mem2')
      then DomBasic.Elt (memJoin' mem1' mem2')
      else DomBasic.Elt (memJoin' mem2' mem1')
    )

(* memJoin_restrict: joined memories have a value of mem1 if an entry
   is in dom(mem1) and mem2 if an entry is not in dom(mem1).

   mem1 join|_{dom(mem1)} mem2
*)
let memJoin_restrict (mem1:DomAbsMem.memT) (mem2:DomAbsMem.memT) 
    : DomAbsMem.memT =
  match mem1 with 
  | DomBasic.Top -> mem1
  | DomBasic.Elt mem1' ->
    (match mem2 with 
    | DomBasic.Top -> raise (PrError "memJoin_restrict: there is top memory")
    | DomBasic.Elt mem2' -> 
      DomBasic.Elt 
        (Orders.locTMapFold
           (fun l v2 acc ->
             if Orders.locTMapMem l mem1'
             then acc
             else Orders.locTMapAdd l v2 acc
           )
           mem2'
           mem1'
        )
    )
(* memOverride: override all memory mappings in mem2 to mem1.  Results
   are same to mem2 join|_{dom(mem2)} mem1 *)
let memOverride (mem1:DomAbsMem.memT) (mem2:DomAbsMem.memT) 
    : DomAbsMem.memT =
  match mem1 with 
  | DomBasic.Top -> raise (PrError "memOverride1: there is top memory")
  | DomBasic.Elt mem1' ->
    (match mem2 with 
    | DomBasic.Top -> raise (PrError "memOverride2: there is top memory")
    | DomBasic.Elt mem2' -> 
      DomBasic.Elt 
        (Orders.locTMapFold
           (fun l v2 acc ->
             Orders.locTMapAdd l v2 acc
           )
           mem2'
           mem1'
        )
    )

(* memOverride_restrict: joined memories have a value of mem2 if an entry
   is in dom(mem1).

   mem2|_{dom(mem1)} join_{dom(mem1)} mem1
*)
let memOverride_restrict (mem1:DomAbsMem.memT) (mem2:DomAbsMem.memT) 
    : DomAbsMem.memT =
  match mem1 with 
  | DomBasic.Top -> mem1
  | DomBasic.Elt mem1' ->
    (match mem2 with 
    | DomBasic.Top -> raise (PrError "memJoin_restrict: there is top memory")
    | DomBasic.Elt mem2' -> 
      DomBasic.Elt 
        (Orders.locTMapFold
           (fun l v2 acc ->
             if Orders.locTMapMem l mem1'
             then Orders.locTMapAdd l v2 acc
             else acc
           )
           mem2' mem1'
        )
    )

let lub (mems:DomAbsMem.memT list) : DomAbsMem.memT =
  List.fold_left
    (fun acc m -> memJoin acc m)
    (DomBasic.Elt Orders.locTMapEmpty)
    mems

let getinput 
    (locals: (Syn.vid,Syn.fid) PMap.t) 
    (str2pos: (string, int) PMap.t)
    (pos2str: (int, string) PMap.t)
    (defuse:GraphLib.DUGraph.t) (node:M5graph.node) (func_state:DomAbsState.func_stateT)
    (locs_joinpoints:Dom.Addr1.t PSet.t)
    (outputof:Dom.Table.t)
    : DomAbsMem.memT = 
  let preds = GraphLib.DUGraph.pred defuse node in
  let mems = 
    (List.map 
       (fun pnode -> 
         let propagated_locs = GraphLib.DUGraph.addrsetof defuse pnode node in
         let propagated_locs = 
           PSet.filter (fun l -> not (l = Dom.Addr1.Null)) propagated_locs 
         in

         let propagated_locs_coq = 
           PSet.map 
             (fun addr1 -> Translator.tr_addr1 locals str2pos addr1)
             propagated_locs
         in

         let premem = 
           let outputmem_pnode = 
             Dom.Table.find outputof (Dom.Work.make0 pnode)
           in
           Translator.tr_mem locals str2pos outputmem_pnode
         in
         let income = project pos2str premem propagated_locs_coq in
         income

       ) 
       preds
    )
  in
  lub mems

let get_args_from_fid_set 
    (prog:Syntax.progT) (fid_set:DomBasic.fidT SetInterface.set)
    : (DomBasic.fidT * DomBasic.idT list) list = 
  let funcs = Syntax.pfuncs prog in
  Orders.intSetFold
    (fun fid acc ->
(*      let func = *)
        match Orders.intMapFind fid funcs with
        | Some f -> (fid, Syntax.fparams f)::acc
        | None -> acc
(*          raise (PrError "get_args_from_fid_set: no syntax for a function")
      in
      (fid, Syntax.fparams func) :: acc*)
    )
    fid_set
    []

let remove_entries
    (args:(DomBasic.fidT * DomBasic.idT list) list) (mem:DomAbsMem.memT)
    : DomAbsMem.memT = 
  List.fold_left
    (fun acc (fid,args) ->
      List.fold_left
        (fun acc arg ->
          let loc = DomAbsValue.llocal fid arg [] in
          match acc with
          | DomBasic.Top -> acc
          | DomBasic.Elt acc' -> 
            (match Orders.locTMapFind loc acc' with
            | Some _ -> DomBasic.Elt (Orders.locTMapRemove loc acc')
            | None -> DomBasic.Elt acc'
            )
        )
        acc
        args
    )
    mem 
    args

let remove_arg_binding
    (prog:Syntax.progT) (fid:DomBasic.fidT) 
    (translated_cfg:Syntax.funcT) 
    (translated_state:DomAbsState.stateT') 
    : DomAbsState.stateT' =
  let cmds = Syntax.fcmds translated_cfg in
  let func_state = 
    match Orders.intMapFind fid translated_state with
    | Some fs -> fs
    | None -> Orders.bidBoolMapEmpty
  in
  let func_state = 
    Orders.intMapFold
      (fun bid cmd acc ->
        match Syntax.ctmn cmd with
        | Syntax.Coq_tmn_call (_, fid_set, _, _) -> 
          let args = get_args_from_fid_set prog fid_set in
          let mem = 
            match Orders.bidBoolMapFind (bid,true) acc with
            | Some m -> m
            | None -> DomBasic.Elt Orders.intMapEmpty
          in
          let mem = remove_entries args mem in
          let acc = Orders.bidBoolMapAdd (bid,true) mem acc in
          acc
        | _ -> acc
      )
      cmds 
      func_state
  in
  let translated_state = Orders.intMapAdd fid func_state translated_state in
  translated_state

let propagate_validate
    (orig:DomAbsState.func_stateT)
    (propagated:DomAbsState.func_stateT)
    : bool =
  Orders.bidBoolMapFold
    (fun k m b -> 
       if (not b) then b else (
	 match (Orders.bidBoolMapFind k propagated) with
	   | None -> false
	   | Some pm -> (
	       match (m, pm) with
		 | (DomBasic.Top, DomBasic.Top) -> true
		 | (DomBasic.Elt m, DomBasic.Elt pm) -> (
		     Orders.locTMapFold
		       (fun k v b ->
			  if (not b) then b else (
			    match (Orders.locTMapFind k pm) with
			      | None -> false
			      | Some pv -> (v = pv)
			  )
		       )
		       pm
		       true
		   )
		 | _ -> false
	     )
       )
    )
    orig
    true

let rec get_pre_order (root:Cfg.Node.t) (children_of:Cfg.Node.t -> Cfg.Node.t PSet.t) 
    : Cfg.Node.t list =
  let childs = children_of root in
  root :: 
    (PSet.fold
       (fun node acc -> (get_pre_order node children_of)@acc)
       childs []
    )

let propagate 
    (locals: (Syn.vid,Syn.fid) PMap.t) 
    (str2pos: (string, int) PMap.t)
    (pos2str: (int, string) PMap.t)
    (p:Syn.pid) (cfg:Cfg.t) (g:Dom.global) (pre:Pre.preinfo) 
    (prog:Syntax.progT)
    (translated_cfg:Syntax.funcT) 
    (func_state:DomAbsState.func_stateT) 
    (translated_acc:DomAcc.accT)
    (defuse:GraphLib.DUGraph.t)
    (pid2joinpoints:(Syn.pid,Ssarua.joinpoints) PMap.t)
    (inputof: Dom.Table.t)
    (outputof: Dom.Table.t)
    : DomAbsState.func_stateT * bool (* checks whether the propagation validation succeeds. *) =

  let fid_coq = PMap.find p str2pos in

  let _ = Profiler.start_event "validation-propagation-states" in

  (* prepare dominance tree and related function *)
  let domtree = cfg.Cfg.domtree in
  let children_of node = PSet.remove node
    (Vocabularies.list2set (Cfg.CFGraph.succ domtree node)) 
  in
  let parent_of node = 
    try 
      PSet.choose 
        (PSet.remove node 
           (Vocabularies.list2set (Cfg.CFGraph.pred domtree node)))
    with Not_found -> raise (PrError "propagate: noparent")
  in

  (* make worklist based on dominance tree: it is the pre-order of the
     dominance tree. *)
  let init_node = Cfg.Node.ENTRY in
  let worklist_domtree = get_pre_order init_node children_of in

  (* propagate on each nodes 
     
     1) If a location is not in joinpoints(locations that should be
     assigned in phi-node), the value is propagated from immediate
     dominance node.

     2) If a location is in joinpoints, the value is propagated by
     defuse edges.  *)

  let joinpoints = 
    match PMap.find_option p pid2joinpoints with
    | Some jp -> jp
    | None -> PMap.empty
  in
  let propagated_state = 
    List.fold_left
      (fun acc node ->
        let bb_coq = Translator.tr_m5graph_node_bidBool str2pos false 
          (M5graph.make_node p node)
        in

        let locs_joinpoints = 
          match PMap.find_option (M5graph.make_node p node) joinpoints with
          | Some locs_jp -> locs_jp
          | None -> PSet.empty
        in
        let locs_joinpoints = 
          PSet.filter (fun l -> not (l = Dom.Addr1.Null)) locs_joinpoints
        in
        let locs_joinpoints_coq =
          PSet.map
            (fun a -> Translator.tr_addr1 locals str2pos a) 
            locs_joinpoints
        in

        (* step1: copy memory from immediate dominant node *)
        let bb_coq_idom = 
          if Cfg.is_entry node
          then Translator.tr_m5graph_node_bidBool str2pos false
            (M5graph.make_node p node)
          else Translator.tr_m5graph_node_bidBool str2pos true
            (M5graph.make_node p (parent_of node))
        in
        let input_mem_from_idom = 
          match Orders.bidBoolMapFind bb_coq_idom acc with
          | Some m -> m
          | None -> DomBasic.Elt Orders.locTMapEmpty
        in
        let input_mem_from_idom = 
          if Cfg.is_entry node
          then input_mem_from_idom
          else cut input_mem_from_idom locs_joinpoints_coq 
        in

        (* If the node is a return node, the input memory is localized
           by access information of callees. *)
        let input_mem_from_idom = 
          if Cfg.is_returnnode cfg node
          then 
            let callnode = Cfg.callof cfg node in
            let callnode_coq = Translator.tr_cfg_node callnode in
            let callees = 
              match Orders.intMapFind callnode_coq (Syntax.fcmds translated_cfg)
              with 
              | Some (_,Syntax.Coq_tmn_call (_,fset,_,_)) -> fset
              | _ -> raise 
                (PrError "Callnode's terminator does not include callees")
            in
            let locs_access_callees = 
              Orders.intSetFold
                (fun callee acc_locs ->
                  match Orders.intMapFind callee 
                    (DomAcc.acclocs translated_acc)
                  with
                  | Some locs -> Orders.intSetUnion acc_locs locs
                  | None -> acc_locs
                )
                callees
                Orders.intSetEmpty
            in
            cut_coq input_mem_from_idom locs_access_callees
          else input_mem_from_idom
        in

        (* It is a strange propagation to deal with the return nodes
           specially. *)
        (* If the node is a return node, add the original memories on
           the node. *)
        (*
        let input_mem_from_idom = 
          if Cfg.is_returnnode cfg node 
          then 
            let orig_mem = 
              match Orders.bidBoolMapFind bb_coq acc with
              | Some m -> m
              | None -> DomBasic.Elt Orders.locTMapEmpty
            in
            memOverride input_mem_from_idom orig_mem
          else input_mem_from_idom
        in
        *)

        (* step2: propagate locations in joinpoints by defuse edges

           This should not be necessary to entry and return node. *)
        let input_mem_from_def = getinput 
          locals str2pos pos2str defuse (M5graph.make_node p node)
          acc locs_joinpoints outputof
        in

        (* add: input memory *)
        let input_mem = 
          memOverride input_mem_from_idom input_mem_from_def 
        in

        (* We use original input memory because the analysis result is
           not a post fixpoint.  Which means the input memories of the
           analysis result can be bigger than the memories we get from
           the validation.  This prevents some non-false assumptions
           on the main analysis from being false assumptions after
           dead code elimination. *)
        let original_input_mem = 
          Translator.tr_mem locals str2pos 
            (Dom.Table.find inputof
               (Dom.Work.make0 (M5graph.make_node p node)))
        in
        let input_mem = memOverride input_mem original_input_mem in

        let acc = Orders.bidBoolMapAdd bb_coq input_mem acc in

        (* add: output memory *)
        let bid_coq = (fun (x,_) -> x) bb_coq in
        let instrs = 
          match Orders.intMapFind bid_coq (Syntax.fcmds translated_cfg) with 
          | Some (intr_list,_) -> intr_list
          | None -> []
        in
        let output_mem = 
          SemAbsBlock.sem_instrs translated_acc fid_coq instrs input_mem 
        in
        let original_output_mem = 
          Translator.tr_mem locals str2pos 
            (Dom.Table.find outputof
               (Dom.Work.make0 (M5graph.make_node p node)))
        in
        let output_mem = 
          if Cfg.is_callnode cfg node
          then output_mem
          else memOverride_restrict output_mem original_output_mem 
        in

        let bb_coq_output = Translator.tr_m5graph_node_bidBool str2pos true 
          (M5graph.make_node p node)
        in
        let acc = Orders.bidBoolMapAdd bb_coq_output output_mem acc in
        
        acc
      )
      func_state
      worklist_domtree
  in

  let _ = Profiler.finish_event "validation-propagation-states" () in

  let validated = propagate_validate func_state propagated_state in
  (propagated_state, validated)

