open Datatypes
open DomBasic
open DomBridge
open DomItv
open DomMap
open DomSet
open DomType
open MapAVLInstance
open MapInterface
open SetAVLInstance
open SetInterface

type locT = ((allocT, idT) sum, fidT * idT) sum * idT list

(** val locT_setoid : locT setoid **)

let locT_setoid =
  pair_setoid
    (sum_setoid varT_setoid (pair_setoid positive_setoid positive_setoid))
    (list_setoid positive_setoid)

(** val locT_setoidDec : locT setoidDec **)

let locT_setoidDec =
  pair_setoidDec
    (sum_setoid varT_setoid (pair_setoid positive_setoid positive_setoid))
    (list_setoid positive_setoid)
    (sum_setoidDec varT_setoid (pair_setoid positive_setoid positive_setoid)
      varT_setoidDec
      (pair_setoidDec positive_setoid positive_setoid positive_setoidDec
        positive_setoidDec))
    (list_setoidDec positive_setoid positive_setoidDec)

(** val locT_poset : locT poset **)

let locT_poset =
  pair_poset
    (sum_setoid varT_setoid (pair_setoid positive_setoid positive_setoid))
    (list_setoid positive_setoid)
    (sum_poset varT_setoid (pair_setoid positive_setoid positive_setoid)
      varT_poset
      (pair_poset positive_setoid positive_setoid positive_poset
        positive_poset)) (list_poset positive_setoid positive_poset)

(** val locT_posetDec : locT posetDec **)

let locT_posetDec =
  pair_posetDec
    (sum_setoid varT_setoid (pair_setoid positive_setoid positive_setoid))
    (list_setoid positive_setoid)
    (sum_poset varT_setoid (pair_setoid positive_setoid positive_setoid)
      varT_poset
      (pair_poset positive_setoid positive_setoid positive_poset
        positive_poset)) (list_poset positive_setoid positive_poset)
    (sum_setoidDec varT_setoid (pair_setoid positive_setoid positive_setoid)
      varT_setoidDec
      (pair_setoidDec positive_setoid positive_setoid positive_setoidDec
        positive_setoidDec))
    (list_setoidDec positive_setoid positive_setoidDec)
    (sum_posetDec varT_setoid (pair_setoid positive_setoid positive_setoid)
      varT_poset
      (pair_poset positive_setoid positive_setoid positive_poset
        positive_poset) varT_setoidDec
      (pair_setoidDec positive_setoid positive_setoid positive_setoidDec
        positive_setoidDec) varT_posetDec
      (pair_posetDec positive_setoid positive_setoid positive_poset
        positive_poset positive_setoidDec positive_setoidDec
        positive_posetDec positive_posetDec))
    (list_posetDec positive_setoid positive_poset positive_setoidDec
      positive_posetDec)

(** val lheap : allocT -> idT list -> locT **)

let lheap a fs =
  ((Coq_inl (Coq_inl a)), fs)

(** val lglobal : idT -> idT list -> locT **)

let lglobal id fs =
  ((Coq_inl (Coq_inr id)), fs)

(** val llocal : fidT -> idT -> idT list -> locT **)

let llocal fid id fs =
  ((Coq_inr (fid, id)), fs)

(** val var_loc : fidT -> varT -> idT list -> locT **)

let var_loc fid var fs =
  match var with
  | Coq_inl gvar -> lglobal gvar fs
  | Coq_inr lvar -> llocal fid lvar fs

type locsT = locT set withtop

(** val filter_fid : fidT -> locT powerset **)

let filter_fid fid = function
| (s, wildcard'0) ->
  (match s with
   | Coq_inl s0 -> true
   | Coq_inr p ->
     let (fid', wildcard') = p in
     negb (eqb positive_setoid positive_setoidDec fid fid'))

type osT = itvT * itvT

(** val osT_poset : osT poset **)

let osT_poset =
  pair_poset_pointwise (option_setoid itvT'_setoid)
    (option_setoid itvT'_setoid) (option_poset itvT'_setoid itvT'_poset)
    (option_poset itvT'_setoid itvT'_poset)

(** val osT_hasBot : osT hasBot **)

let osT_hasBot =
  pair_hasBot_pointwise (option_setoid itvT'_setoid)
    (option_setoid itvT'_setoid) (option_poset itvT'_setoid itvT'_poset)
    (option_poset itvT'_setoid itvT'_poset)
    (option_hasBot itvT'_setoid itvT'_poset)
    (option_hasBot itvT'_setoid itvT'_poset)

(** val osT_hasJoin : osT hasJoin **)

let osT_hasJoin =
  pair_hasJoin_pointwise (option_setoid itvT'_setoid)
    (option_setoid itvT'_setoid) (option_poset itvT'_setoid itvT'_poset)
    (option_poset itvT'_setoid itvT'_poset)
    (option_hasJoin itvT'_setoid itvT'_poset itvT'_hasJoin)
    (option_hasJoin itvT'_setoid itvT'_poset itvT'_hasJoin)

type ossT = osT * itvT

(** val ossT_poset : ossT poset **)

let ossT_poset =
  pair_poset_pointwise
    (pair_setoid (option_setoid itvT'_setoid) (option_setoid itvT'_setoid))
    (option_setoid itvT'_setoid) osT_poset
    (option_poset itvT'_setoid itvT'_poset)

(** val ossT_hasBot : ossT hasBot **)

let ossT_hasBot =
  pair_hasBot_pointwise
    (pair_setoid (option_setoid itvT'_setoid) (option_setoid itvT'_setoid))
    (option_setoid itvT'_setoid) osT_poset
    (option_poset itvT'_setoid itvT'_poset) osT_hasBot
    (option_hasBot itvT'_setoid itvT'_poset)

(** val ossT_hasJoin : ossT hasJoin **)

let ossT_hasJoin =
  pair_hasJoin_pointwise
    (pair_setoid (option_setoid itvT'_setoid) (option_setoid itvT'_setoid))
    (option_setoid itvT'_setoid) osT_poset
    (option_poset itvT'_setoid itvT'_poset) osT_hasJoin
    (option_hasJoin itvT'_setoid itvT'_poset itvT'_hasJoin)

type arrvT' = (allocT, ossT) dict

(** val arrvT'_poset : arrvT' poset **)

let arrvT'_poset =
  map_poset positive_setoid positive_poset positive_setoidDec
    positive_posetDec
    (pair_setoid
      (pair_setoid (option_setoid itvT'_setoid) (option_setoid itvT'_setoid))
      (option_setoid itvT'_setoid)) ossT_poset ossT_hasBot

(** val arrvT'_hasBot : arrvT' hasBot **)

let arrvT'_hasBot =
  map_hasBot positive_setoid positive_poset positive_setoidDec
    positive_posetDec
    (pair_setoid
      (pair_setoid (option_setoid itvT'_setoid) (option_setoid itvT'_setoid))
      (option_setoid itvT'_setoid)) ossT_poset ossT_hasBot

(** val arrvT'_hasJoin : arrvT' hasJoin **)

let arrvT'_hasJoin =
  map_hasJoin positive_setoid positive_poset positive_setoidDec
    positive_posetDec
    (pair_setoid
      (pair_setoid (option_setoid itvT'_setoid) (option_setoid itvT'_setoid))
      (option_setoid itvT'_setoid)) ossT_poset ossT_hasBot ossT_hasJoin

type arrvT = arrvT' withtop

(** val arrvT_poset : arrvT poset **)

let arrvT_poset =
  withtop_poset
    (map_setoid positive_setoid positive_poset positive_setoidDec
      positive_posetDec
      (pair_setoid
        (pair_setoid (option_setoid itvT'_setoid)
          (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
      ossT_poset ossT_hasBot)
    (map_poset positive_setoid positive_poset positive_setoidDec
      positive_posetDec
      (pair_setoid
        (pair_setoid (option_setoid itvT'_setoid)
          (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
      ossT_poset ossT_hasBot)

(** val arrvT_hasBot : arrvT hasBot **)

let arrvT_hasBot =
  withtop_hasBot
    (map_setoid positive_setoid positive_poset positive_setoidDec
      positive_posetDec
      (pair_setoid
        (pair_setoid (option_setoid itvT'_setoid)
          (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
      ossT_poset ossT_hasBot)
    (map_poset positive_setoid positive_poset positive_setoidDec
      positive_posetDec
      (pair_setoid
        (pair_setoid (option_setoid itvT'_setoid)
          (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
      ossT_poset ossT_hasBot) arrvT'_hasBot

(** val arrvT_hasJoin : arrvT hasJoin **)

let arrvT_hasJoin =
  withtop_hasJoin
    (map_setoid positive_setoid positive_poset positive_setoidDec
      positive_posetDec
      (pair_setoid
        (pair_setoid (option_setoid itvT'_setoid)
          (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
      ossT_poset ossT_hasBot)
    (map_poset positive_setoid positive_poset positive_setoidDec
      positive_posetDec
      (pair_setoid
        (pair_setoid (option_setoid itvT'_setoid)
          (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
      ossT_poset ossT_hasBot) arrvT'_hasJoin

(** val locs_from_arrvT : arrvT -> locsT **)

let locs_from_arrvT = function
| Elt av ->
  Elt
    (MapInterface.fold
      (order_OrderedType positive_setoid positive_poset positive_setoidDec
        positive_posetDec)
      (coq_MapAVL_FMap
        (order_OrderedType positive_setoid positive_poset positive_setoidDec
          positive_posetDec)) (fun k v s ->
      if negb
           (eqb
             (pair_setoid
               (pair_setoid (option_setoid itvT'_setoid)
                 (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
             (pair_setoidDec
               (pair_setoid (option_setoid itvT'_setoid)
                 (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid)
               (pair_setoidDec (option_setoid itvT'_setoid)
                 (option_setoid itvT'_setoid)
                 (option_setoidDec itvT'_setoid itvT'_setoidDec)
                 (option_setoidDec itvT'_setoid itvT'_setoidDec))
               (option_setoidDec itvT'_setoid itvT'_setoidDec)) v
             (bot
               (pair_setoid
                 (pair_setoid (option_setoid itvT'_setoid)
                   (option_setoid itvT'_setoid))
                 (option_setoid itvT'_setoid)) ossT_poset ossT_hasBot))
      then (coq_SetAVL_FSet
             (order_OrderedType locT_setoid locT_poset locT_setoidDec
               locT_posetDec)).add (lheap k []) s
      else s) av
      (coq_SetAVL_FSet
        (order_OrderedType locT_setoid locT_poset locT_setoidDec
          locT_posetDec)).empty)
| Top ->
  top
    (withtop_setoid
      (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
    (withtop_poset
      (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
      (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec))
    (withtop_hasTop
      (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
      (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec))

type valueT = (itvT * locsT) * arrvT

(** val itv_of : valueT -> itvT **)

let itv_of v =
  fst (fst v)

(** val locs_of : valueT -> locsT **)

let locs_of v =
  snd (fst v)

(** val arrv_of : valueT -> arrvT **)

let arrv_of v =
  snd v

(** val itv_to_valueT : itvT -> valueT **)

let itv_to_valueT itv =
  ((itv,
    (bot
      (withtop_setoid
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
      (withtop_poset
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
        (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec))
      (withtop_hasBot
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
        (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
        (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)))),
    (bot
      (withtop_setoid
        (map_setoid positive_setoid positive_poset positive_setoidDec
          positive_posetDec
          (pair_setoid
            (pair_setoid (option_setoid itvT'_setoid)
              (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
          ossT_poset ossT_hasBot)) arrvT_poset arrvT_hasBot))

(** val locs_to_valueT : locsT -> valueT **)

let locs_to_valueT locs =
  (((bot (option_setoid itvT'_setoid) (option_poset itvT'_setoid itvT'_poset)
      (option_hasBot itvT'_setoid itvT'_poset)), locs),
    (bot
      (withtop_setoid
        (map_setoid positive_setoid positive_poset positive_setoidDec
          positive_posetDec
          (pair_setoid
            (pair_setoid (option_setoid itvT'_setoid)
              (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
          ossT_poset ossT_hasBot)) arrvT_poset arrvT_hasBot))

(** val arrv_to_valueT : arrvT -> valueT **)

let arrv_to_valueT arrv =
  (((bot (option_setoid itvT'_setoid) (option_poset itvT'_setoid itvT'_poset)
      (option_hasBot itvT'_setoid itvT'_poset)),
    (bot
      (withtop_setoid
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
      (withtop_poset
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
        (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec))
      (withtop_hasBot
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
        (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
        (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)))),
    arrv)

(** val valueT_setoid : valueT setoid **)

let valueT_setoid =
  pair_setoid
    (pair_setoid (option_setoid itvT'_setoid)
      (withtop_setoid
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)))
    (withtop_setoid
      (map_setoid positive_setoid positive_poset positive_setoidDec
        positive_posetDec
        (pair_setoid
          (pair_setoid (option_setoid itvT'_setoid)
            (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
        ossT_poset ossT_hasBot))

(** val valueT_poset : valueT poset **)

let valueT_poset =
  pair_poset_pointwise
    (pair_setoid (option_setoid itvT'_setoid)
      (withtop_setoid
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)))
    (withtop_setoid
      (map_setoid positive_setoid positive_poset positive_setoidDec
        positive_posetDec
        (pair_setoid
          (pair_setoid (option_setoid itvT'_setoid)
            (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
        ossT_poset ossT_hasBot))
    (pair_poset_pointwise (option_setoid itvT'_setoid)
      (withtop_setoid
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
      (option_poset itvT'_setoid itvT'_poset)
      (withtop_poset
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
        (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)))
    arrvT_poset

(** val valueT_hasBot : valueT hasBot **)

let valueT_hasBot =
  pair_hasBot_pointwise
    (pair_setoid (option_setoid itvT'_setoid)
      (withtop_setoid
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)))
    (withtop_setoid
      (map_setoid positive_setoid positive_poset positive_setoidDec
        positive_posetDec
        (pair_setoid
          (pair_setoid (option_setoid itvT'_setoid)
            (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
        ossT_poset ossT_hasBot))
    (pair_poset_pointwise (option_setoid itvT'_setoid)
      (withtop_setoid
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
      (option_poset itvT'_setoid itvT'_poset)
      (withtop_poset
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
        (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)))
    arrvT_poset
    (pair_hasBot_pointwise (option_setoid itvT'_setoid)
      (withtop_setoid
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
      (option_poset itvT'_setoid itvT'_poset)
      (withtop_poset
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
        (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec))
      (option_hasBot itvT'_setoid itvT'_poset)
      (withtop_hasBot
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
        (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
        (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)))
    arrvT_hasBot

(** val valueT_hasTop : valueT hasTop **)

let valueT_hasTop =
  pair_hasTop_pointwise
    (pair_setoid (option_setoid itvT'_setoid)
      (withtop_setoid
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)))
    (withtop_setoid
      (map_setoid positive_setoid positive_poset positive_setoidDec
        positive_posetDec
        (pair_setoid
          (pair_setoid (option_setoid itvT'_setoid)
            (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
        ossT_poset ossT_hasBot))
    (pair_poset_pointwise (option_setoid itvT'_setoid)
      (withtop_setoid
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
      (option_poset itvT'_setoid itvT'_poset)
      (withtop_poset
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
        (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)))
    arrvT_poset
    (pair_hasTop_pointwise (option_setoid itvT'_setoid)
      (withtop_setoid
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
      (option_poset itvT'_setoid itvT'_poset)
      (withtop_poset
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
        (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec))
      (option_hasTop itvT'_setoid itvT'_poset itvT'_hasTop)
      (withtop_hasTop
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
        (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)))
    (withtop_hasTop
      (map_setoid positive_setoid positive_poset positive_setoidDec
        positive_posetDec
        (pair_setoid
          (pair_setoid (option_setoid itvT'_setoid)
            (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
        ossT_poset ossT_hasBot)
      (map_poset positive_setoid positive_poset positive_setoidDec
        positive_posetDec
        (pair_setoid
          (pair_setoid (option_setoid itvT'_setoid)
            (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
        ossT_poset ossT_hasBot))

(** val valueT_hasJoin : valueT hasJoin **)

let valueT_hasJoin =
  pair_hasJoin_pointwise
    (pair_setoid (option_setoid itvT'_setoid)
      (withtop_setoid
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)))
    (withtop_setoid
      (map_setoid positive_setoid positive_poset positive_setoidDec
        positive_posetDec
        (pair_setoid
          (pair_setoid (option_setoid itvT'_setoid)
            (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
        ossT_poset ossT_hasBot))
    (pair_poset_pointwise (option_setoid itvT'_setoid)
      (withtop_setoid
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
      (option_poset itvT'_setoid itvT'_poset)
      (withtop_poset
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
        (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)))
    arrvT_poset
    (pair_hasJoin_pointwise (option_setoid itvT'_setoid)
      (withtop_setoid
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
      (option_poset itvT'_setoid itvT'_poset)
      (withtop_poset
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
        (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec))
      (option_hasJoin itvT'_setoid itvT'_poset itvT'_hasJoin)
      (withtop_hasJoin
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
        (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
        (set_hasJoin locT_setoid locT_poset locT_setoidDec locT_posetDec)))
    arrvT_hasJoin

(** val arrv_op_add : arrvT -> itvT -> arrvT **)

let arrv_op_add x y = match y with
| Some i ->
  (match x with
   | Elt xe ->
     Elt
       (map
         (order_OrderedType positive_setoid positive_poset positive_setoidDec
           positive_posetDec)
         (coq_MapAVL_FMap
           (order_OrderedType positive_setoid positive_poset
             positive_setoidDec positive_posetDec)) (fun e ->
         let (y0, st) = e in let (o, s) = y0 in (((op_add o y), s), st)) xe)
   | Top -> Top)
| None ->
  bot
    (withtop_setoid
      (map_setoid positive_setoid positive_poset positive_setoidDec
        positive_posetDec
        (pair_setoid
          (pair_setoid (option_setoid itvT'_setoid)
            (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
        ossT_poset ossT_hasBot)) arrvT_poset arrvT_hasBot

(** val arrv_op_sub : arrvT -> itvT -> arrvT **)

let arrv_op_sub x y = match y with
| Some i ->
  (match x with
   | Elt xe ->
     Elt
       (map
         (order_OrderedType positive_setoid positive_poset positive_setoidDec
           positive_posetDec)
         (coq_MapAVL_FMap
           (order_OrderedType positive_setoid positive_poset
             positive_setoidDec positive_posetDec)) (fun e ->
         let (y0, st) = e in let (o, s) = y0 in (((op_sub o y), s), st)) xe)
   | Top -> Top)
| None ->
  bot
    (withtop_setoid
      (map_setoid positive_setoid positive_poset positive_setoidDec
        positive_posetDec
        (pair_setoid
          (pair_setoid (option_setoid itvT'_setoid)
            (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
        ossT_poset ossT_hasBot)) arrvT_poset arrvT_hasBot

(** val vconst : int -> valueT **)

let vconst z =
  itv_to_valueT (Some (Coq_itv'_z_z (z, z)))

(** val vop1 : op1 -> valueT -> valueT **)

let vop1 op val0 =
  match op with
  | Coq_op_not -> itv_to_valueT (op_not (itv_of val0))
  | Coq_op_bnot -> itv_to_valueT (op_bnot (itv_of val0))

(** val vop2 : op2 -> valueT -> valueT -> valueT **)

let vop2 op x y =
  match op with
  | Coq_op_add ->
    (((op_add (itv_of x) (itv_of y)),
      (bot
        (withtop_setoid
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_poset
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_hasBot
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)))),
      (join
        (withtop_setoid
          (map_setoid positive_setoid positive_poset positive_setoidDec
            positive_posetDec
            (pair_setoid
              (pair_setoid (option_setoid itvT'_setoid)
                (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
            ossT_poset ossT_hasBot)) arrvT_poset arrvT_hasJoin
        (arrv_op_add (arrv_of x) (itv_of y))
        (arrv_op_add (arrv_of y) (itv_of x))))
  | Coq_op_sub ->
    (((op_sub (itv_of x) (itv_of y)),
      (bot
        (withtop_setoid
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_poset
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_hasBot
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)))),
      (arrv_op_sub (arrv_of x) (itv_of y)))
  | Coq_op_mul ->
    (((op_mul (itv_of x) (itv_of y)),
      (bot
        (withtop_setoid
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_poset
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_hasBot
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)))),
      (bot
        (withtop_setoid
          (map_setoid positive_setoid positive_poset positive_setoidDec
            positive_posetDec
            (pair_setoid
              (pair_setoid (option_setoid itvT'_setoid)
                (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
            ossT_poset ossT_hasBot)) arrvT_poset arrvT_hasBot))
  | Coq_op_div ->
    (((op_div (itv_of x) (itv_of y)),
      (bot
        (withtop_setoid
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_poset
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_hasBot
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)))),
      (bot
        (withtop_setoid
          (map_setoid positive_setoid positive_poset positive_setoidDec
            positive_posetDec
            (pair_setoid
              (pair_setoid (option_setoid itvT'_setoid)
                (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
            ossT_poset ossT_hasBot)) arrvT_poset arrvT_hasBot))
  | Coq_op_mod ->
    (((op_mod (itv_of x) (itv_of y)),
      (bot
        (withtop_setoid
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_poset
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_hasBot
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)))),
      (bot
        (withtop_setoid
          (map_setoid positive_setoid positive_poset positive_setoidDec
            positive_posetDec
            (pair_setoid
              (pair_setoid (option_setoid itvT'_setoid)
                (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
            ossT_poset ossT_hasBot)) arrvT_poset arrvT_hasBot))
  | Coq_op_lt ->
    (((op_lt (itv_of x) (itv_of y)),
      (bot
        (withtop_setoid
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_poset
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_hasBot
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)))),
      (bot
        (withtop_setoid
          (map_setoid positive_setoid positive_poset positive_setoidDec
            positive_posetDec
            (pair_setoid
              (pair_setoid (option_setoid itvT'_setoid)
                (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
            ossT_poset ossT_hasBot)) arrvT_poset arrvT_hasBot))
  | Coq_op_le ->
    (((op_le (itv_of x) (itv_of y)),
      (bot
        (withtop_setoid
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_poset
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_hasBot
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)))),
      (bot
        (withtop_setoid
          (map_setoid positive_setoid positive_poset positive_setoidDec
            positive_posetDec
            (pair_setoid
              (pair_setoid (option_setoid itvT'_setoid)
                (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
            ossT_poset ossT_hasBot)) arrvT_poset arrvT_hasBot))
  | Coq_op_gt ->
    (((op_gt (itv_of x) (itv_of y)),
      (bot
        (withtop_setoid
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_poset
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_hasBot
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)))),
      (bot
        (withtop_setoid
          (map_setoid positive_setoid positive_poset positive_setoidDec
            positive_posetDec
            (pair_setoid
              (pair_setoid (option_setoid itvT'_setoid)
                (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
            ossT_poset ossT_hasBot)) arrvT_poset arrvT_hasBot))
  | Coq_op_ge ->
    (((op_ge (itv_of x) (itv_of y)),
      (bot
        (withtop_setoid
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_poset
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_hasBot
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)))),
      (bot
        (withtop_setoid
          (map_setoid positive_setoid positive_poset positive_setoidDec
            positive_posetDec
            (pair_setoid
              (pair_setoid (option_setoid itvT'_setoid)
                (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
            ossT_poset ossT_hasBot)) arrvT_poset arrvT_hasBot))
  | Coq_op_eq ->
    (((op_eq (itv_of x) (itv_of y)),
      (bot
        (withtop_setoid
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_poset
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_hasBot
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)))),
      (bot
        (withtop_setoid
          (map_setoid positive_setoid positive_poset positive_setoidDec
            positive_posetDec
            (pair_setoid
              (pair_setoid (option_setoid itvT'_setoid)
                (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
            ossT_poset ossT_hasBot)) arrvT_poset arrvT_hasBot))
  | Coq_op_neq ->
    (((op_neq (itv_of x) (itv_of y)),
      (bot
        (withtop_setoid
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_poset
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_hasBot
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)))),
      (bot
        (withtop_setoid
          (map_setoid positive_setoid positive_poset positive_setoidDec
            positive_posetDec
            (pair_setoid
              (pair_setoid (option_setoid itvT'_setoid)
                (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
            ossT_poset ossT_hasBot)) arrvT_poset arrvT_hasBot))
  | Coq_op_and ->
    (((op_and (itv_of x) (itv_of y)),
      (bot
        (withtop_setoid
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_poset
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_hasBot
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)))),
      (bot
        (withtop_setoid
          (map_setoid positive_setoid positive_poset positive_setoidDec
            positive_posetDec
            (pair_setoid
              (pair_setoid (option_setoid itvT'_setoid)
                (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
            ossT_poset ossT_hasBot)) arrvT_poset arrvT_hasBot))
  | Coq_op_or ->
    (((op_or (itv_of x) (itv_of y)),
      (bot
        (withtop_setoid
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_poset
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_hasBot
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)))),
      (bot
        (withtop_setoid
          (map_setoid positive_setoid positive_poset positive_setoidDec
            positive_posetDec
            (pair_setoid
              (pair_setoid (option_setoid itvT'_setoid)
                (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
            ossT_poset ossT_hasBot)) arrvT_poset arrvT_hasBot))
  | Coq_op_band ->
    (((op_band (itv_of x) (itv_of y)),
      (bot
        (withtop_setoid
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_poset
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_hasBot
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)))),
      (bot
        (withtop_setoid
          (map_setoid positive_setoid positive_poset positive_setoidDec
            positive_posetDec
            (pair_setoid
              (pair_setoid (option_setoid itvT'_setoid)
                (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
            ossT_poset ossT_hasBot)) arrvT_poset arrvT_hasBot))
  | Coq_op_bor ->
    (((op_bor (itv_of x) (itv_of y)),
      (bot
        (withtop_setoid
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_poset
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_hasBot
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)))),
      (bot
        (withtop_setoid
          (map_setoid positive_setoid positive_poset positive_setoidDec
            positive_posetDec
            (pair_setoid
              (pair_setoid (option_setoid itvT'_setoid)
                (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
            ossT_poset ossT_hasBot)) arrvT_poset arrvT_hasBot))
  | Coq_op_bxor ->
    (((op_bxor (itv_of x) (itv_of y)),
      (bot
        (withtop_setoid
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_poset
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_hasBot
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)))),
      (bot
        (withtop_setoid
          (map_setoid positive_setoid positive_poset positive_setoidDec
            positive_posetDec
            (pair_setoid
              (pair_setoid (option_setoid itvT'_setoid)
                (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
            ossT_poset ossT_hasBot)) arrvT_poset arrvT_hasBot))
  | Coq_op_lshift ->
    (((op_lshift (itv_of x) (itv_of y)),
      (bot
        (withtop_setoid
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_poset
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_hasBot
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)))),
      (bot
        (withtop_setoid
          (map_setoid positive_setoid positive_poset positive_setoidDec
            positive_posetDec
            (pair_setoid
              (pair_setoid (option_setoid itvT'_setoid)
                (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
            ossT_poset ossT_hasBot)) arrvT_poset arrvT_hasBot))
  | Coq_op_rshift ->
    (((op_rshift (itv_of x) (itv_of y)),
      (bot
        (withtop_setoid
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_poset
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (withtop_hasBot
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)))),
      (bot
        (withtop_setoid
          (map_setoid positive_setoid positive_poset positive_setoidDec
            positive_posetDec
            (pair_setoid
              (pair_setoid (option_setoid itvT'_setoid)
                (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
            ossT_poset ossT_hasBot)) arrvT_poset arrvT_hasBot))

