open Datatypes
open DomAbsMem
open DomAbsState
open DomAbsValue
open DomAcc
open DomBasic
open DomBridge
open DomItv
open DomMap
open DomSet
open DomType
open LibTactics
open List0
open MapAVLInstance
open MapInterface
open SemAbsBlock
open SemAccInsen
open SetAVLInstance
open SetInterface
open Syntax
open Ssrbool

(** val acc_correct_acclocs_blk :
    accT -> fidT -> bidT -> cmdT -> func_stateT -> bool **)

let acc_correct_acclocs_blk pos2str acc fid bid cmd state =
  let instrs = cinstrs cmd in
  let tmn = ctmn cmd in
  let instrs_mem = func_stateT_maplike.mlookup (bid, false) state in
  let tmn_mem' =
    sem_instrs acc fid (cinstrs cmd)
      (func_stateT_maplike.mlookup (bid, false) state)
  in
  let tmn_mem = func_stateT_maplike.mlookup (bid, true) state in
  (&&)
    ((&&) (acc_correct_acclocs_instrs pos2str acc fid instrs instrs_mem)
      (acc_correct_acclocs_tmn pos2str acc fid tmn tmn_mem'))
    (acc_correct_acclocs_tmn pos2str acc fid tmn tmn_mem)

(** val acc_correct_acclocs_func :
    accT -> fidT -> funcT -> func_stateT -> bool **)

let acc_correct_acclocs_func pos2str acc fid fbody state =
  let alocs =
    (map_maplike positive_setoid positive_poset positive_setoidDec
      positive_posetDec
      (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
      (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
      (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)).mlookup
      fid acc.acclocs
  in
  (&&)
    (if map_forall positive_setoid positive_poset positive_setoidDec
       positive_posetDec (fun bid cmd ->
         acc_correct_acclocs_blk pos2str acc fid bid cmd state) (fcmds fbody)
     then true
     else (prerr_endline "acc_correct_acclocs failed: first"; false))
    (if list_forall (fun i ->
      (set_setlike locT_setoid locT_poset locT_setoidDec locT_posetDec).smem
        ((Coq_inr (fid,i)),[]) alocs) (fparams fbody)
     then true
     else (prerr_endline "acc_correct_acclocs failed: second"; false))

(** val acc_correct_acclocs_star_func : accT -> fidT -> bool **)

let acc_correct_acclocs_star_func acc fid =
  set_forall (pair_setoid positive_setoid positive_setoid)
    (pair_poset positive_setoid positive_setoid positive_poset
      positive_poset)
    (pair_setoidDec positive_setoid positive_setoid positive_setoidDec
      positive_setoidDec)
    (pair_posetDec positive_setoid positive_setoid positive_poset
      positive_poset positive_setoidDec positive_setoidDec positive_posetDec
      positive_posetDec) (fun callinfo ->
    (&&)
      ((set_setlike positive_setoid positive_poset positive_setoidDec
         positive_posetDec).smem (fst callinfo)
        ((map_maplike positive_setoid positive_poset positive_setoidDec
           positive_posetDec
           (set_setoid positive_setoid positive_poset positive_setoidDec
             positive_posetDec)
           (set_poset positive_setoid positive_poset positive_setoidDec
             positive_posetDec)
           (set_hasBot positive_setoid positive_poset positive_setoidDec
             positive_posetDec)).mlookup fid acc.callfroms_star))
      (leb
        (set_setoid positive_setoid positive_poset positive_setoidDec
          positive_posetDec)
        (set_poset positive_setoid positive_poset positive_setoidDec
          positive_posetDec)
        (set_setoidDec positive_setoid positive_poset positive_setoidDec
          positive_posetDec)
        (set_posetDec positive_setoid positive_poset positive_setoidDec
          positive_posetDec)
        ((map_maplike positive_setoid positive_poset positive_setoidDec
           positive_posetDec
           (set_setoid positive_setoid positive_poset positive_setoidDec
             positive_posetDec)
           (set_poset positive_setoid positive_poset positive_setoidDec
             positive_posetDec)
           (set_hasBot positive_setoid positive_poset positive_setoidDec
             positive_posetDec)).mlookup (fst callinfo) acc.callfroms_star)
        ((map_maplike positive_setoid positive_poset positive_setoidDec
           positive_posetDec
           (set_setoid positive_setoid positive_poset positive_setoidDec
             positive_posetDec)
           (set_poset positive_setoid positive_poset positive_setoidDec
             positive_posetDec)
           (set_hasBot positive_setoid positive_poset positive_setoidDec
             positive_posetDec)).mlookup fid acc.callfroms_star)))
    ((map_maplike positive_setoid positive_poset positive_setoidDec
       positive_posetDec
       (set_setoid (pair_setoid positive_setoid positive_setoid)
         (pair_poset positive_setoid positive_setoid positive_poset
           positive_poset)
         (pair_setoidDec positive_setoid positive_setoid positive_setoidDec
           positive_setoidDec)
         (pair_posetDec positive_setoid positive_setoid positive_poset
           positive_poset positive_setoidDec positive_setoidDec
           positive_posetDec positive_posetDec))
       (set_poset (pair_setoid positive_setoid positive_setoid)
         (pair_poset positive_setoid positive_setoid positive_poset
           positive_poset)
         (pair_setoidDec positive_setoid positive_setoid positive_setoidDec
           positive_setoidDec)
         (pair_posetDec positive_setoid positive_setoid positive_poset
           positive_poset positive_setoidDec positive_setoidDec
           positive_posetDec positive_posetDec))
       (set_hasBot (pair_setoid positive_setoid positive_setoid)
         (pair_poset positive_setoid positive_setoid positive_poset
           positive_poset)
         (pair_setoidDec positive_setoid positive_setoid positive_setoidDec
           positive_setoidDec)
         (pair_posetDec positive_setoid positive_setoid positive_poset
           positive_poset positive_setoidDec positive_setoidDec
           positive_posetDec positive_posetDec))).mlookup fid acc.callfroms)

(** val acc_correct_acclocs_callfroms_func : accT -> fidT -> bool **)

let acc_correct_acclocs_callfroms_func acc fid =
  set_forall (pair_setoid positive_setoid positive_setoid)
    (pair_poset positive_setoid positive_setoid positive_poset
      positive_poset)
    (pair_setoidDec positive_setoid positive_setoid positive_setoidDec
      positive_setoidDec)
    (pair_posetDec positive_setoid positive_setoid positive_poset
      positive_poset positive_setoidDec positive_setoidDec positive_posetDec
      positive_posetDec) (fun callinfo ->
    leb (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
      (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
      (set_setoidDec locT_setoid locT_poset locT_setoidDec locT_posetDec)
      (set_posetDec locT_setoid locT_poset locT_setoidDec locT_posetDec)
      ((map_maplike positive_setoid positive_poset positive_setoidDec
         positive_posetDec
         (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
         (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
         (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)).mlookup
        fid acc.acclocs)
      ((map_maplike positive_setoid positive_poset positive_setoidDec
         positive_posetDec
         (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
         (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
         (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)).mlookup
        (fst callinfo) acc.acclocs))
    ((map_maplike positive_setoid positive_poset positive_setoidDec
       positive_posetDec
       (set_setoid (pair_setoid positive_setoid positive_setoid)
         (pair_poset positive_setoid positive_setoid positive_poset
           positive_poset)
         (pair_setoidDec positive_setoid positive_setoid positive_setoidDec
           positive_setoidDec)
         (pair_posetDec positive_setoid positive_setoid positive_poset
           positive_poset positive_setoidDec positive_setoidDec
           positive_posetDec positive_posetDec))
       (set_poset (pair_setoid positive_setoid positive_setoid)
         (pair_poset positive_setoid positive_setoid positive_poset
           positive_poset)
         (pair_setoidDec positive_setoid positive_setoid positive_setoidDec
           positive_setoidDec)
         (pair_posetDec positive_setoid positive_setoid positive_poset
           positive_poset positive_setoidDec positive_setoidDec
           positive_posetDec positive_posetDec))
       (set_hasBot (pair_setoid positive_setoid positive_setoid)
         (pair_poset positive_setoid positive_setoid positive_poset
           positive_poset)
         (pair_setoidDec positive_setoid positive_setoid positive_setoidDec
           positive_setoidDec)
         (pair_posetDec positive_setoid positive_setoid positive_poset
           positive_poset positive_setoidDec positive_setoidDec
           positive_posetDec positive_posetDec))).mlookup fid acc.callfroms)

(** val use_le_var : fidT -> varT -> memT -> memT -> bool **)

let use_le_var fid x lhs rhs =
  leb valueT_setoid valueT_poset
    (pair_setoidDec
      (pair_setoid (option_setoid itvT'_setoid)
        (withtop_setoid
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)))
      (withtop_setoid
        (map_setoid positive_setoid positive_poset positive_setoidDec
          positive_posetDec
          (pair_setoid
            (pair_setoid (option_setoid itvT'_setoid)
              (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
          ossT_poset ossT_hasBot))
      (pair_setoidDec (option_setoid itvT'_setoid)
        (withtop_setoid
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (option_setoidDec itvT'_setoid itvT'_setoidDec)
        (withtop_setoidDec
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_setoidDec locT_setoid locT_poset locT_setoidDec locT_posetDec)))
      (withtop_setoidDec
        (map_setoid positive_setoid positive_poset positive_setoidDec
          positive_posetDec
          (pair_setoid
            (pair_setoid (option_setoid itvT'_setoid)
              (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
          ossT_poset ossT_hasBot)
        (map_setoidDec positive_setoid positive_poset positive_setoidDec
          positive_posetDec
          (pair_setoid
            (pair_setoid (option_setoid itvT'_setoid)
              (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
          ossT_poset ossT_hasBot
          (pair_setoidDec
            (pair_setoid (option_setoid itvT'_setoid)
              (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid)
            (pair_setoidDec (option_setoid itvT'_setoid)
              (option_setoid itvT'_setoid)
              (option_setoidDec itvT'_setoid itvT'_setoidDec)
              (option_setoidDec itvT'_setoid itvT'_setoidDec))
            (option_setoidDec itvT'_setoid itvT'_setoidDec)))))
    (pair_posetDec_pointwise
      (pair_setoid (option_setoid itvT'_setoid)
        (withtop_setoid
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)))
      (withtop_setoid
        (map_setoid positive_setoid positive_poset positive_setoidDec
          positive_posetDec
          (pair_setoid
            (pair_setoid (option_setoid itvT'_setoid)
              (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
          ossT_poset ossT_hasBot))
      (pair_poset_pointwise (option_setoid itvT'_setoid)
        (withtop_setoid
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (option_poset itvT'_setoid itvT'_poset)
        (withtop_poset
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)))
      arrvT_poset
      (pair_setoidDec (option_setoid itvT'_setoid)
        (withtop_setoid
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (option_setoidDec itvT'_setoid itvT'_setoidDec)
        (withtop_setoidDec
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_setoidDec locT_setoid locT_poset locT_setoidDec locT_posetDec)))
      (withtop_setoidDec
        (map_setoid positive_setoid positive_poset positive_setoidDec
          positive_posetDec
          (pair_setoid
            (pair_setoid (option_setoid itvT'_setoid)
              (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
          ossT_poset ossT_hasBot)
        (map_setoidDec positive_setoid positive_poset positive_setoidDec
          positive_posetDec
          (pair_setoid
            (pair_setoid (option_setoid itvT'_setoid)
              (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
          ossT_poset ossT_hasBot
          (pair_setoidDec
            (pair_setoid (option_setoid itvT'_setoid)
              (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid)
            (pair_setoidDec (option_setoid itvT'_setoid)
              (option_setoid itvT'_setoid)
              (option_setoidDec itvT'_setoid itvT'_setoidDec)
              (option_setoidDec itvT'_setoid itvT'_setoidDec))
            (option_setoidDec itvT'_setoid itvT'_setoidDec))))
      (pair_posetDec_pointwise (option_setoid itvT'_setoid)
        (withtop_setoid
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (option_poset itvT'_setoid itvT'_poset)
        (withtop_poset
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (option_setoidDec itvT'_setoid itvT'_setoidDec)
        (withtop_setoidDec
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_setoidDec locT_setoid locT_poset locT_setoidDec locT_posetDec))
        (option_posetDec itvT'_setoid itvT'_poset itvT'_setoidDec
          itvT'_posetDec)
        (withtop_posetDec
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_setoidDec locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_posetDec locT_setoid locT_poset locT_setoidDec locT_posetDec)))
      (withtop_posetDec
        (map_setoid positive_setoid positive_poset positive_setoidDec
          positive_posetDec
          (pair_setoid
            (pair_setoid (option_setoid itvT'_setoid)
              (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
          ossT_poset ossT_hasBot)
        (map_poset positive_setoid positive_poset positive_setoidDec
          positive_posetDec
          (pair_setoid
            (pair_setoid (option_setoid itvT'_setoid)
              (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
          ossT_poset ossT_hasBot)
        (map_setoidDec positive_setoid positive_poset positive_setoidDec
          positive_posetDec
          (pair_setoid
            (pair_setoid (option_setoid itvT'_setoid)
              (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
          ossT_poset ossT_hasBot
          (pair_setoidDec
            (pair_setoid (option_setoid itvT'_setoid)
              (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid)
            (pair_setoidDec (option_setoid itvT'_setoid)
              (option_setoid itvT'_setoid)
              (option_setoidDec itvT'_setoid itvT'_setoidDec)
              (option_setoidDec itvT'_setoid itvT'_setoidDec))
            (option_setoidDec itvT'_setoid itvT'_setoidDec)))
        (map_posetDec positive_setoid positive_poset positive_setoidDec
          positive_posetDec
          (pair_setoid
            (pair_setoid (option_setoid itvT'_setoid)
              (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid))
          ossT_poset ossT_hasBot
          (pair_setoidDec
            (pair_setoid (option_setoid itvT'_setoid)
              (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid)
            (pair_setoidDec (option_setoid itvT'_setoid)
              (option_setoid itvT'_setoid)
              (option_setoidDec itvT'_setoid itvT'_setoidDec)
              (option_setoidDec itvT'_setoid itvT'_setoidDec))
            (option_setoidDec itvT'_setoid itvT'_setoidDec))
          (pair_posetDec_pointwise
            (pair_setoid (option_setoid itvT'_setoid)
              (option_setoid itvT'_setoid)) (option_setoid itvT'_setoid)
            osT_poset (option_poset itvT'_setoid itvT'_poset)
            (pair_setoidDec (option_setoid itvT'_setoid)
              (option_setoid itvT'_setoid)
              (option_setoidDec itvT'_setoid itvT'_setoidDec)
              (option_setoidDec itvT'_setoid itvT'_setoidDec))
            (option_setoidDec itvT'_setoid itvT'_setoidDec)
            (pair_posetDec_pointwise (option_setoid itvT'_setoid)
              (option_setoid itvT'_setoid)
              (option_poset itvT'_setoid itvT'_poset)
              (option_poset itvT'_setoid itvT'_poset)
              (option_setoidDec itvT'_setoid itvT'_setoidDec)
              (option_setoidDec itvT'_setoid itvT'_setoidDec)
              (option_posetDec itvT'_setoid itvT'_poset itvT'_setoidDec
                itvT'_posetDec)
              (option_posetDec itvT'_setoid itvT'_poset itvT'_setoidDec
                itvT'_posetDec))
            (option_posetDec itvT'_setoid itvT'_poset itvT'_setoidDec
              itvT'_posetDec)))))
    (memT_maplike.mlookup (var_loc fid x []) lhs)
    (memT_maplike.mlookup (var_loc fid x []) rhs)

(** val use_le_expr : fidT -> exprT -> memT -> memT -> bool **)

let rec use_le_expr fid e lhs rhs =
  match e with
  | Coq_expr_const l -> true
  | Coq_expr_op2 (o, e1, e2) ->
    (&&) (use_le_expr fid e1 lhs rhs) (use_le_expr fid e2 lhs rhs)
  | Coq_expr_op1 (o, e1) -> use_le_expr fid e1 lhs rhs
  | Coq_expr_value_of lv0 ->
    (&&) (use_le_lv fid lv0 lhs rhs)
      (match sem_expr_lv fid lv0 lhs with
       | Elt s ->
         fold
           (order_OrderedType locT_setoid locT_poset locT_setoidDec
             locT_posetDec)
           (coq_SetAVL_FSet
             (order_OrderedType locT_setoid locT_poset locT_setoidDec
               locT_posetDec)) (fun loc acc ->
           (&&) acc
             (leb valueT_setoid valueT_poset
               (pair_setoidDec
                 (pair_setoid (option_setoid itvT'_setoid)
                   (withtop_setoid
                     (set_setoid locT_setoid locT_poset locT_setoidDec
                       locT_posetDec)))
                 (withtop_setoid
                   (map_setoid positive_setoid positive_poset
                     positive_setoidDec positive_posetDec
                     (pair_setoid
                       (pair_setoid (option_setoid itvT'_setoid)
                         (option_setoid itvT'_setoid))
                       (option_setoid itvT'_setoid)) ossT_poset ossT_hasBot))
                 (pair_setoidDec (option_setoid itvT'_setoid)
                   (withtop_setoid
                     (set_setoid locT_setoid locT_poset locT_setoidDec
                       locT_posetDec))
                   (option_setoidDec itvT'_setoid itvT'_setoidDec)
                   (withtop_setoidDec
                     (set_setoid locT_setoid locT_poset locT_setoidDec
                       locT_posetDec)
                     (set_setoidDec locT_setoid locT_poset locT_setoidDec
                       locT_posetDec)))
                 (withtop_setoidDec
                   (map_setoid positive_setoid positive_poset
                     positive_setoidDec positive_posetDec
                     (pair_setoid
                       (pair_setoid (option_setoid itvT'_setoid)
                         (option_setoid itvT'_setoid))
                       (option_setoid itvT'_setoid)) ossT_poset ossT_hasBot)
                   (map_setoidDec positive_setoid positive_poset
                     positive_setoidDec positive_posetDec
                     (pair_setoid
                       (pair_setoid (option_setoid itvT'_setoid)
                         (option_setoid itvT'_setoid))
                       (option_setoid itvT'_setoid)) ossT_poset ossT_hasBot
                     (pair_setoidDec
                       (pair_setoid (option_setoid itvT'_setoid)
                         (option_setoid itvT'_setoid))
                       (option_setoid itvT'_setoid)
                       (pair_setoidDec (option_setoid itvT'_setoid)
                         (option_setoid itvT'_setoid)
                         (option_setoidDec itvT'_setoid itvT'_setoidDec)
                         (option_setoidDec itvT'_setoid itvT'_setoidDec))
                       (option_setoidDec itvT'_setoid itvT'_setoidDec)))))
               (pair_posetDec_pointwise
                 (pair_setoid (option_setoid itvT'_setoid)
                   (withtop_setoid
                     (set_setoid locT_setoid locT_poset locT_setoidDec
                       locT_posetDec)))
                 (withtop_setoid
                   (map_setoid positive_setoid positive_poset
                     positive_setoidDec positive_posetDec
                     (pair_setoid
                       (pair_setoid (option_setoid itvT'_setoid)
                         (option_setoid itvT'_setoid))
                       (option_setoid itvT'_setoid)) ossT_poset ossT_hasBot))
                 (pair_poset_pointwise (option_setoid itvT'_setoid)
                   (withtop_setoid
                     (set_setoid locT_setoid locT_poset locT_setoidDec
                       locT_posetDec))
                   (option_poset itvT'_setoid itvT'_poset)
                   (withtop_poset
                     (set_setoid locT_setoid locT_poset locT_setoidDec
                       locT_posetDec)
                     (set_poset locT_setoid locT_poset locT_setoidDec
                       locT_posetDec))) arrvT_poset
                 (pair_setoidDec (option_setoid itvT'_setoid)
                   (withtop_setoid
                     (set_setoid locT_setoid locT_poset locT_setoidDec
                       locT_posetDec))
                   (option_setoidDec itvT'_setoid itvT'_setoidDec)
                   (withtop_setoidDec
                     (set_setoid locT_setoid locT_poset locT_setoidDec
                       locT_posetDec)
                     (set_setoidDec locT_setoid locT_poset locT_setoidDec
                       locT_posetDec)))
                 (withtop_setoidDec
                   (map_setoid positive_setoid positive_poset
                     positive_setoidDec positive_posetDec
                     (pair_setoid
                       (pair_setoid (option_setoid itvT'_setoid)
                         (option_setoid itvT'_setoid))
                       (option_setoid itvT'_setoid)) ossT_poset ossT_hasBot)
                   (map_setoidDec positive_setoid positive_poset
                     positive_setoidDec positive_posetDec
                     (pair_setoid
                       (pair_setoid (option_setoid itvT'_setoid)
                         (option_setoid itvT'_setoid))
                       (option_setoid itvT'_setoid)) ossT_poset ossT_hasBot
                     (pair_setoidDec
                       (pair_setoid (option_setoid itvT'_setoid)
                         (option_setoid itvT'_setoid))
                       (option_setoid itvT'_setoid)
                       (pair_setoidDec (option_setoid itvT'_setoid)
                         (option_setoid itvT'_setoid)
                         (option_setoidDec itvT'_setoid itvT'_setoidDec)
                         (option_setoidDec itvT'_setoid itvT'_setoidDec))
                       (option_setoidDec itvT'_setoid itvT'_setoidDec))))
                 (pair_posetDec_pointwise (option_setoid itvT'_setoid)
                   (withtop_setoid
                     (set_setoid locT_setoid locT_poset locT_setoidDec
                       locT_posetDec))
                   (option_poset itvT'_setoid itvT'_poset)
                   (withtop_poset
                     (set_setoid locT_setoid locT_poset locT_setoidDec
                       locT_posetDec)
                     (set_poset locT_setoid locT_poset locT_setoidDec
                       locT_posetDec))
                   (option_setoidDec itvT'_setoid itvT'_setoidDec)
                   (withtop_setoidDec
                     (set_setoid locT_setoid locT_poset locT_setoidDec
                       locT_posetDec)
                     (set_setoidDec locT_setoid locT_poset locT_setoidDec
                       locT_posetDec))
                   (option_posetDec itvT'_setoid itvT'_poset itvT'_setoidDec
                     itvT'_posetDec)
                   (withtop_posetDec
                     (set_setoid locT_setoid locT_poset locT_setoidDec
                       locT_posetDec)
                     (set_poset locT_setoid locT_poset locT_setoidDec
                       locT_posetDec)
                     (set_setoidDec locT_setoid locT_poset locT_setoidDec
                       locT_posetDec)
                     (set_posetDec locT_setoid locT_poset locT_setoidDec
                       locT_posetDec)))
                 (withtop_posetDec
                   (map_setoid positive_setoid positive_poset
                     positive_setoidDec positive_posetDec
                     (pair_setoid
                       (pair_setoid (option_setoid itvT'_setoid)
                         (option_setoid itvT'_setoid))
                       (option_setoid itvT'_setoid)) ossT_poset ossT_hasBot)
                   (map_poset positive_setoid positive_poset
                     positive_setoidDec positive_posetDec
                     (pair_setoid
                       (pair_setoid (option_setoid itvT'_setoid)
                         (option_setoid itvT'_setoid))
                       (option_setoid itvT'_setoid)) ossT_poset ossT_hasBot)
                   (map_setoidDec positive_setoid positive_poset
                     positive_setoidDec positive_posetDec
                     (pair_setoid
                       (pair_setoid (option_setoid itvT'_setoid)
                         (option_setoid itvT'_setoid))
                       (option_setoid itvT'_setoid)) ossT_poset ossT_hasBot
                     (pair_setoidDec
                       (pair_setoid (option_setoid itvT'_setoid)
                         (option_setoid itvT'_setoid))
                       (option_setoid itvT'_setoid)
                       (pair_setoidDec (option_setoid itvT'_setoid)
                         (option_setoid itvT'_setoid)
                         (option_setoidDec itvT'_setoid itvT'_setoidDec)
                         (option_setoidDec itvT'_setoid itvT'_setoidDec))
                       (option_setoidDec itvT'_setoid itvT'_setoidDec)))
                   (map_posetDec positive_setoid positive_poset
                     positive_setoidDec positive_posetDec
                     (pair_setoid
                       (pair_setoid (option_setoid itvT'_setoid)
                         (option_setoid itvT'_setoid))
                       (option_setoid itvT'_setoid)) ossT_poset ossT_hasBot
                     (pair_setoidDec
                       (pair_setoid (option_setoid itvT'_setoid)
                         (option_setoid itvT'_setoid))
                       (option_setoid itvT'_setoid)
                       (pair_setoidDec (option_setoid itvT'_setoid)
                         (option_setoid itvT'_setoid)
                         (option_setoidDec itvT'_setoid itvT'_setoidDec)
                         (option_setoidDec itvT'_setoid itvT'_setoidDec))
                       (option_setoidDec itvT'_setoid itvT'_setoidDec))
                     (pair_posetDec_pointwise
                       (pair_setoid (option_setoid itvT'_setoid)
                         (option_setoid itvT'_setoid))
                       (option_setoid itvT'_setoid) osT_poset
                       (option_poset itvT'_setoid itvT'_poset)
                       (pair_setoidDec (option_setoid itvT'_setoid)
                         (option_setoid itvT'_setoid)
                         (option_setoidDec itvT'_setoid itvT'_setoidDec)
                         (option_setoidDec itvT'_setoid itvT'_setoidDec))
                       (option_setoidDec itvT'_setoid itvT'_setoidDec)
                       (pair_posetDec_pointwise (option_setoid itvT'_setoid)
                         (option_setoid itvT'_setoid)
                         (option_poset itvT'_setoid itvT'_poset)
                         (option_poset itvT'_setoid itvT'_poset)
                         (option_setoidDec itvT'_setoid itvT'_setoidDec)
                         (option_setoidDec itvT'_setoid itvT'_setoidDec)
                         (option_posetDec itvT'_setoid itvT'_poset
                           itvT'_setoidDec itvT'_posetDec)
                         (option_posetDec itvT'_setoid itvT'_poset
                           itvT'_setoidDec itvT'_posetDec))
                       (option_posetDec itvT'_setoid itvT'_poset
                         itvT'_setoidDec itvT'_posetDec)))))
               (memT_maplike.mlookup loc lhs) (memT_maplike.mlookup loc rhs)))
           s true
       | Top -> false)
  | Coq_expr_addr_of lv0 -> use_le_lv fid lv0 lhs rhs
  | Coq_expr_cast (t, e1) -> use_le_expr fid e1 lhs rhs
  | Coq_expr_size_of e1 -> use_le_expr fid e1 lhs rhs

(** val use_le_lv : fidT -> lv -> memT -> memT -> bool **)

and use_le_lv fid l lhs rhs =
  match l with
  | Coq_lv_id x -> true
  | Coq_lv_deref e1 -> use_le_expr fid e1 lhs rhs
  | Coq_lv_array (e1, e2) ->
    (&&) (use_le_expr fid e1 lhs rhs) (use_le_expr fid e2 lhs rhs)
  | Coq_lv_struct (l0, pos) -> use_le_lv fid l0 lhs rhs

(** val use_le_exprs : fidT -> exprT list -> memT -> memT -> bool **)

let rec use_le_exprs fid es lhs rhs =
  match es with
  | [] -> true
  | e :: es1 ->
    (&&) (use_le_expr fid e lhs rhs) (use_le_exprs fid es1 lhs rhs)

(** val use_le_locs : fidT -> exprT -> memT -> memT -> bool **)

let use_le_locs fid e lhs rhs =
  match locs_of (sem_expr fid e lhs) with
  | Elt s ->
    fold
      (order_OrderedType locT_setoid locT_poset locT_setoidDec locT_posetDec)
      (coq_SetAVL_FSet
        (order_OrderedType locT_setoid locT_poset locT_setoidDec
          locT_posetDec)) (fun loc acc ->
      (&&) acc
        (leb valueT_setoid valueT_poset
          (pair_setoidDec
            (pair_setoid (option_setoid itvT'_setoid)
              (withtop_setoid
                (set_setoid locT_setoid locT_poset locT_setoidDec
                  locT_posetDec)))
            (withtop_setoid
              (map_setoid positive_setoid positive_poset positive_setoidDec
                positive_posetDec
                (pair_setoid
                  (pair_setoid (option_setoid itvT'_setoid)
                    (option_setoid itvT'_setoid))
                  (option_setoid itvT'_setoid)) ossT_poset ossT_hasBot))
            (pair_setoidDec (option_setoid itvT'_setoid)
              (withtop_setoid
                (set_setoid locT_setoid locT_poset locT_setoidDec
                  locT_posetDec))
              (option_setoidDec itvT'_setoid itvT'_setoidDec)
              (withtop_setoidDec
                (set_setoid locT_setoid locT_poset locT_setoidDec
                  locT_posetDec)
                (set_setoidDec locT_setoid locT_poset locT_setoidDec
                  locT_posetDec)))
            (withtop_setoidDec
              (map_setoid positive_setoid positive_poset positive_setoidDec
                positive_posetDec
                (pair_setoid
                  (pair_setoid (option_setoid itvT'_setoid)
                    (option_setoid itvT'_setoid))
                  (option_setoid itvT'_setoid)) ossT_poset ossT_hasBot)
              (map_setoidDec positive_setoid positive_poset
                positive_setoidDec positive_posetDec
                (pair_setoid
                  (pair_setoid (option_setoid itvT'_setoid)
                    (option_setoid itvT'_setoid))
                  (option_setoid itvT'_setoid)) ossT_poset ossT_hasBot
                (pair_setoidDec
                  (pair_setoid (option_setoid itvT'_setoid)
                    (option_setoid itvT'_setoid))
                  (option_setoid itvT'_setoid)
                  (pair_setoidDec (option_setoid itvT'_setoid)
                    (option_setoid itvT'_setoid)
                    (option_setoidDec itvT'_setoid itvT'_setoidDec)
                    (option_setoidDec itvT'_setoid itvT'_setoidDec))
                  (option_setoidDec itvT'_setoid itvT'_setoidDec)))))
          (pair_posetDec_pointwise
            (pair_setoid (option_setoid itvT'_setoid)
              (withtop_setoid
                (set_setoid locT_setoid locT_poset locT_setoidDec
                  locT_posetDec)))
            (withtop_setoid
              (map_setoid positive_setoid positive_poset positive_setoidDec
                positive_posetDec
                (pair_setoid
                  (pair_setoid (option_setoid itvT'_setoid)
                    (option_setoid itvT'_setoid))
                  (option_setoid itvT'_setoid)) ossT_poset ossT_hasBot))
            (pair_poset_pointwise (option_setoid itvT'_setoid)
              (withtop_setoid
                (set_setoid locT_setoid locT_poset locT_setoidDec
                  locT_posetDec)) (option_poset itvT'_setoid itvT'_poset)
              (withtop_poset
                (set_setoid locT_setoid locT_poset locT_setoidDec
                  locT_posetDec)
                (set_poset locT_setoid locT_poset locT_setoidDec
                  locT_posetDec))) arrvT_poset
            (pair_setoidDec (option_setoid itvT'_setoid)
              (withtop_setoid
                (set_setoid locT_setoid locT_poset locT_setoidDec
                  locT_posetDec))
              (option_setoidDec itvT'_setoid itvT'_setoidDec)
              (withtop_setoidDec
                (set_setoid locT_setoid locT_poset locT_setoidDec
                  locT_posetDec)
                (set_setoidDec locT_setoid locT_poset locT_setoidDec
                  locT_posetDec)))
            (withtop_setoidDec
              (map_setoid positive_setoid positive_poset positive_setoidDec
                positive_posetDec
                (pair_setoid
                  (pair_setoid (option_setoid itvT'_setoid)
                    (option_setoid itvT'_setoid))
                  (option_setoid itvT'_setoid)) ossT_poset ossT_hasBot)
              (map_setoidDec positive_setoid positive_poset
                positive_setoidDec positive_posetDec
                (pair_setoid
                  (pair_setoid (option_setoid itvT'_setoid)
                    (option_setoid itvT'_setoid))
                  (option_setoid itvT'_setoid)) ossT_poset ossT_hasBot
                (pair_setoidDec
                  (pair_setoid (option_setoid itvT'_setoid)
                    (option_setoid itvT'_setoid))
                  (option_setoid itvT'_setoid)
                  (pair_setoidDec (option_setoid itvT'_setoid)
                    (option_setoid itvT'_setoid)
                    (option_setoidDec itvT'_setoid itvT'_setoidDec)
                    (option_setoidDec itvT'_setoid itvT'_setoidDec))
                  (option_setoidDec itvT'_setoid itvT'_setoidDec))))
            (pair_posetDec_pointwise (option_setoid itvT'_setoid)
              (withtop_setoid
                (set_setoid locT_setoid locT_poset locT_setoidDec
                  locT_posetDec)) (option_poset itvT'_setoid itvT'_poset)
              (withtop_poset
                (set_setoid locT_setoid locT_poset locT_setoidDec
                  locT_posetDec)
                (set_poset locT_setoid locT_poset locT_setoidDec
                  locT_posetDec))
              (option_setoidDec itvT'_setoid itvT'_setoidDec)
              (withtop_setoidDec
                (set_setoid locT_setoid locT_poset locT_setoidDec
                  locT_posetDec)
                (set_setoidDec locT_setoid locT_poset locT_setoidDec
                  locT_posetDec))
              (option_posetDec itvT'_setoid itvT'_poset itvT'_setoidDec
                itvT'_posetDec)
              (withtop_posetDec
                (set_setoid locT_setoid locT_poset locT_setoidDec
                  locT_posetDec)
                (set_poset locT_setoid locT_poset locT_setoidDec
                  locT_posetDec)
                (set_setoidDec locT_setoid locT_poset locT_setoidDec
                  locT_posetDec)
                (set_posetDec locT_setoid locT_poset locT_setoidDec
                  locT_posetDec)))
            (withtop_posetDec
              (map_setoid positive_setoid positive_poset positive_setoidDec
                positive_posetDec
                (pair_setoid
                  (pair_setoid (option_setoid itvT'_setoid)
                    (option_setoid itvT'_setoid))
                  (option_setoid itvT'_setoid)) ossT_poset ossT_hasBot)
              (map_poset positive_setoid positive_poset positive_setoidDec
                positive_posetDec
                (pair_setoid
                  (pair_setoid (option_setoid itvT'_setoid)
                    (option_setoid itvT'_setoid))
                  (option_setoid itvT'_setoid)) ossT_poset ossT_hasBot)
              (map_setoidDec positive_setoid positive_poset
                positive_setoidDec positive_posetDec
                (pair_setoid
                  (pair_setoid (option_setoid itvT'_setoid)
                    (option_setoid itvT'_setoid))
                  (option_setoid itvT'_setoid)) ossT_poset ossT_hasBot
                (pair_setoidDec
                  (pair_setoid (option_setoid itvT'_setoid)
                    (option_setoid itvT'_setoid))
                  (option_setoid itvT'_setoid)
                  (pair_setoidDec (option_setoid itvT'_setoid)
                    (option_setoid itvT'_setoid)
                    (option_setoidDec itvT'_setoid itvT'_setoidDec)
                    (option_setoidDec itvT'_setoid itvT'_setoidDec))
                  (option_setoidDec itvT'_setoid itvT'_setoidDec)))
              (map_posetDec positive_setoid positive_poset positive_setoidDec
                positive_posetDec
                (pair_setoid
                  (pair_setoid (option_setoid itvT'_setoid)
                    (option_setoid itvT'_setoid))
                  (option_setoid itvT'_setoid)) ossT_poset ossT_hasBot
                (pair_setoidDec
                  (pair_setoid (option_setoid itvT'_setoid)
                    (option_setoid itvT'_setoid))
                  (option_setoid itvT'_setoid)
                  (pair_setoidDec (option_setoid itvT'_setoid)
                    (option_setoid itvT'_setoid)
                    (option_setoidDec itvT'_setoid itvT'_setoidDec)
                    (option_setoidDec itvT'_setoid itvT'_setoidDec))
                  (option_setoidDec itvT'_setoid itvT'_setoidDec))
                (pair_posetDec_pointwise
                  (pair_setoid (option_setoid itvT'_setoid)
                    (option_setoid itvT'_setoid))
                  (option_setoid itvT'_setoid) osT_poset
                  (option_poset itvT'_setoid itvT'_poset)
                  (pair_setoidDec (option_setoid itvT'_setoid)
                    (option_setoid itvT'_setoid)
                    (option_setoidDec itvT'_setoid itvT'_setoidDec)
                    (option_setoidDec itvT'_setoid itvT'_setoidDec))
                  (option_setoidDec itvT'_setoid itvT'_setoidDec)
                  (pair_posetDec_pointwise (option_setoid itvT'_setoid)
                    (option_setoid itvT'_setoid)
                    (option_poset itvT'_setoid itvT'_poset)
                    (option_poset itvT'_setoid itvT'_poset)
                    (option_setoidDec itvT'_setoid itvT'_setoidDec)
                    (option_setoidDec itvT'_setoid itvT'_setoidDec)
                    (option_posetDec itvT'_setoid itvT'_poset itvT'_setoidDec
                      itvT'_posetDec)
                    (option_posetDec itvT'_setoid itvT'_poset itvT'_setoidDec
                      itvT'_posetDec))
                  (option_posetDec itvT'_setoid itvT'_poset itvT'_setoidDec
                    itvT'_posetDec))))) (memT_maplike.mlookup loc lhs)
          (memT_maplike.mlookup loc rhs))) s true
  | Top -> false

(** val use_le_locss : fidT -> exprT list -> memT -> memT -> bool **)

let rec use_le_locss fid args lhs rhs =
  match args with
  | [] -> true
  | arg :: tl0 ->
    (&&) (use_le_locs fid arg lhs rhs) (use_le_locss fid tl0 lhs rhs)

(** val use_le_tmn : fidT -> tmnT -> memT -> memT -> bool **)

let use_le_tmn fid tmn lhs rhs =
  match tmn with
  | Coq_tmn_jump nextbids -> true
  | Coq_tmn_call (nextbid, nextfids, es, ret) ->
    (&&) (use_le_exprs fid es lhs rhs)
      (match ret with
       | Some ret' -> use_le_var fid ret' lhs rhs
       | None -> true)
  | Coq_tmn_exit e -> use_le_expr fid e lhs rhs

(** val use_le_instr : accT -> fidT -> instrT -> memT -> memT -> bool **)

let use_le_instr acc fid instr lhs rhs =
  match instr with
  | Coq_instr_assign (lv0, e) ->
    (&&) (use_le_lv fid lv0 lhs rhs) (use_le_expr fid e lhs rhs)
  | Coq_instr_alloc (lv0, a, site) ->
    let Coq_alloc_arr (sz, str) = a in
    (&&) ((&&) (use_le_lv fid lv0 lhs rhs) (use_le_expr fid sz lhs rhs))
      (use_le_expr fid str lhs rhs)
  | Coq_instr_assume e -> use_le_expr fid e lhs rhs
  | Coq_instr_libcall (ol, lc, args, site) ->
    (&&)
      ((&&)
        (match lc with
         | Coq_strcpy ->
           let ldst =
             locs_of
               (sem_expr fid (hd (Coq_expr_const Coq_lit_bot) args) lhs)
           in
           let rdst =
             locs_of
               (sem_expr fid (hd (Coq_expr_const Coq_lit_bot) args) rhs)
           in
           let lsrc =
             locs_of
               (sem_expr fid (hd (Coq_expr_const Coq_lit_bot) (tl args)) lhs)
           in
           let rsrc =
             locs_of
               (sem_expr fid (hd (Coq_expr_const Coq_lit_bot) (tl args)) rhs)
           in
           let lsrcValue = memory_lookup lsrc lhs in
           let rsrcValue = memory_lookup rsrc rhs in
           let lnewMem = memory_update acc ldst lsrcValue lhs in
           let rnewMem = memory_update acc rdst rsrcValue rhs in
           (match ol with
            | Some lv0 -> use_le_lv fid lv0 lnewMem rnewMem
            | None -> true)
         | Coq_scanf -> true
         | _ ->
           (match ol with
            | Some lv0 -> use_le_lv fid lv0 lhs rhs
            | None -> true))
        (match lc with
         | Coq_strcpy -> use_le_exprs fid args lhs rhs
         | Coq_scanf -> use_le_exprs fid args lhs rhs
         | Coq_builtin_alloca -> use_le_exprs fid args lhs rhs
         | Coq_builtin_choose_expr -> use_le_exprs fid args lhs rhs
         | _ -> true))
      (match lc with
       | Coq_strcpy -> use_le_locss fid args lhs rhs
       | _ -> true)
  | _ -> true

(** val use_le_instrs :
    accT -> fidT -> instrT list -> memT -> memT -> bool **)

let rec use_le_instrs acc fid instrs lhs rhs =
  match instrs with
  | [] -> true
  | instr :: next_instrs ->
    (&&) (use_le_instr acc fid instr lhs rhs)
      (let newlhs = sem_instr acc fid instr lhs in
       let newrhs = sem_instr acc fid instr rhs in
       use_le_instrs acc fid next_instrs newlhs newrhs)

(** val use_le_blk : accT -> fidT -> bidT -> cmdT -> memT -> memT -> bool **)

let use_le_blk acc fid bid cmd lin rin =
  let instrs = cinstrs cmd in
  let tmn = ctmn cmd in
  (&&) (use_le_instrs acc fid instrs lin rin)
    (use_le_tmn fid tmn (sem_instrs acc fid instrs lin)
      (sem_instrs acc fid instrs rin))

(** val use_le_func :
    accT -> fidT -> funcT -> func_stateT -> func_stateT -> bool **)

let use_le_func acc fid func lhs rhs =
  map_forall positive_setoid positive_poset positive_setoidDec
    positive_posetDec (fun bid cmd ->
    use_le_blk acc fid bid cmd (func_stateT_maplike.mlookup (bid, false) lhs)
      (func_stateT_maplike.mlookup (bid, false) rhs)) (fcmds func)

(** val eqb_inter_blk :
    fidT -> bidT -> cmdT -> func_stateT -> func_stateT -> bool **)

let eqb_inter_blk fid bid cmd lhs rhs =
  match ctmn cmd with
  | Coq_tmn_jump nextbids -> true
  | Coq_tmn_call (nbid, nextfids, args, retto) ->
    (&&)
      (eqb memT_setoid memT_setoidDec
        (func_stateT_maplike.mlookup (nbid, false) lhs)
        (func_stateT_maplike.mlookup (nbid, false) rhs))
      (eqb memT_setoid memT_setoidDec
        (func_stateT_maplike.mlookup (bid, true) lhs)
        (func_stateT_maplike.mlookup (bid, true) rhs))
  | Coq_tmn_exit e ->
    eqb memT_setoid memT_setoidDec
      (func_stateT_maplike.mlookup (bid, true) lhs)
      (func_stateT_maplike.mlookup (bid, true) rhs)

(** val eqb_inter_func :
    fidT -> funcT -> func_stateT -> func_stateT -> bool **)

let eqb_inter_func fid func lhs rhs =
  (&&)
    (eqb memT_setoid memT_setoidDec
      (func_stateT_maplike.mlookup ((fentry func), false) lhs)
      (func_stateT_maplike.mlookup ((fentry func), false) rhs))
    (map_forall positive_setoid positive_poset positive_setoidDec
      positive_posetDec (fun bid cmd -> eqb_inter_blk fid bid cmd lhs rhs)
      (fcmds func))

(** val postfix_jump :
    func_stateT -> fidT -> funcT -> bidT -> bidT set -> bool **)

let postfix_jump state fid func bid nbids =
  set_forall positive_setoid positive_poset positive_setoidDec
    positive_posetDec (fun nbid ->
    leb memT_setoid memT_poset memT_setoidDec memT_posetDec
      (func_stateT_maplike.mlookup (bid, true) state)
      (func_stateT_maplike.mlookup (nbid, false) state)) nbids

(** val postfix_call :
    progT -> accT -> func_stateT -> fidT -> funcT -> bidT -> bidT -> fidT ->
    exprT list -> bool **)

let postfix_call prog acc state fid func bid nbid nfid args =
  match find
          (order_OrderedType positive_setoid positive_poset
            positive_setoidDec positive_posetDec)
          (coq_MapAVL_FMap
            (order_OrderedType positive_setoid positive_poset
              positive_setoidDec positive_posetDec)) nfid (pfuncs prog) with
  | Some nfunc ->
    let locs =
      (map_maplike positive_setoid positive_poset positive_setoidDec
        positive_posetDec
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
        (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
        (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)).mlookup
        nfid acc.acclocs
    in
    let locs' =
      set_to_powerset locT_setoid locT_poset locT_setoidDec locT_posetDec
        locs
    in
    let mem = func_stateT_maplike.mlookup (bid, true) state in
    let mem_out =
      mrestrict locT_setoid locT_setoidDec valueT_setoid valueT_poset
        memT_setoid memT_poset
        (withtop_hasBot
          (map_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec
            valueT_setoid valueT_poset valueT_hasBot)
          (map_poset locT_setoid locT_poset locT_setoidDec locT_posetDec
            valueT_setoid valueT_poset valueT_hasBot)
          (map_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec
            valueT_setoid valueT_poset valueT_hasBot)) memT_maplike
        memT_restricting_maplike (powerset_neg locT_setoid locs') mem
    in
    (&&)
      (if (set_setlike (pair_setoid positive_setoid positive_setoid)
         (pair_poset positive_setoid positive_setoid positive_poset
           positive_poset)
         (pair_setoidDec positive_setoid positive_setoid positive_setoidDec
           positive_setoidDec)
         (pair_posetDec positive_setoid positive_setoid positive_poset
           positive_poset positive_setoidDec positive_setoidDec
           positive_posetDec positive_posetDec)).smem (fid, bid)
        ((map_maplike positive_setoid positive_poset positive_setoidDec
           positive_posetDec
           (set_setoid (pair_setoid positive_setoid positive_setoid)
             (pair_poset positive_setoid positive_setoid positive_poset
               positive_poset)
             (pair_setoidDec positive_setoid positive_setoid
               positive_setoidDec positive_setoidDec)
             (pair_posetDec positive_setoid positive_setoid positive_poset
               positive_poset positive_setoidDec positive_setoidDec
               positive_posetDec positive_posetDec))
           (set_poset (pair_setoid positive_setoid positive_setoid)
             (pair_poset positive_setoid positive_setoid positive_poset
               positive_poset)
             (pair_setoidDec positive_setoid positive_setoid
               positive_setoidDec positive_setoidDec)
             (pair_posetDec positive_setoid positive_setoid positive_poset
               positive_poset positive_setoidDec positive_setoidDec
               positive_posetDec positive_posetDec))
           (set_hasBot (pair_setoid positive_setoid positive_setoid)
             (pair_poset positive_setoid positive_setoid positive_poset
               positive_poset)
             (pair_setoidDec positive_setoid positive_setoid
               positive_setoidDec positive_setoidDec)
             (pair_posetDec positive_setoid positive_setoid positive_poset
               positive_poset positive_setoidDec positive_setoidDec
               positive_posetDec positive_posetDec))).mlookup nfid
          acc.callfroms)
       then true
       else (prerr_endline "postfix_call failed 1";false)
      )
      (if leb memT_setoid memT_poset memT_setoidDec memT_posetDec mem_out
        (func_stateT_maplike.mlookup (nbid, false) state)
       then true
       else (prerr_endline 
               ("postfix_call failed 2 "^(Pprinter.pprint_bidT bid)^" -> "^
                   (Pprinter.pprint_bidT nbid));
             false)
      )
  | None -> true

(** val postfix_block :
    progT -> accT -> func_stateT -> fidT -> funcT -> bidT -> cmdT -> bool **)

let postfix_block pos2str prog acc state fid func bid = function
| (instrs,tmn) ->
  (&&)
    (if leb memT_setoid memT_poset memT_setoidDec memT_posetDec
      (sem_instrs acc fid instrs
        (func_stateT_maplike.mlookup (bid, false) state))
      (func_stateT_maplike.mlookup (bid, true) state)
     then true
     else (prerr_endline ("postfix_block failed:inter node "^(Pprinter.pprint_bidT bid)); false))
    (match tmn with
     | Coq_tmn_jump nbids -> 
       if postfix_jump state fid func bid nbids
       then true
       else (prerr_endline ("postfix_block failed:jump "^(Pprinter.pprint_bidT bid)); false)         
     | Coq_tmn_call (nbid, nfids, args, retto) ->
       if set_forall positive_setoid positive_poset positive_setoidDec
         positive_posetDec (fun nfid ->
         postfix_call prog acc state fid func bid nbid nfid args) nfids
       then true
       else (prerr_endline ("postfix_block failed:call "^(Pprinter.pprint_bidT bid)); false)
     | Coq_tmn_exit ret -> true)

(** val postfix_func :
    progT -> accT -> func_stateT -> fidT -> funcT -> bool **)

let postfix_func pos2str prog acc state fid func =
  if
  map_forall positive_setoid positive_poset positive_setoidDec
    positive_posetDec (postfix_block pos2str prog acc state fid func) (fcmds func)
  then true
  else (prerr_endline "real postfix_func failed"; false)

(** val postfix_inter_call :
    progT -> accT -> func_stateT -> stateT' -> fidT -> funcT -> bidT -> bidT
    -> fidT -> exprT list -> bool **)

let postfix_inter_call pos2str prog acc func_state state fid func bid nbid nfid args =
  match find
          (order_OrderedType positive_setoid positive_poset
            positive_setoidDec positive_posetDec)
          (coq_MapAVL_FMap
            (order_OrderedType positive_setoid positive_poset
              positive_setoidDec positive_posetDec)) nfid (pfuncs prog) with
  | Some nfunc ->
    let locs =
      (map_maplike positive_setoid positive_poset positive_setoidDec
        positive_posetDec
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
        (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
        (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)).mlookup
        nfid acc.acclocs
    in
    let locs' =
      set_to_powerset locT_setoid locT_poset locT_setoidDec locT_posetDec
        locs
    in
    let mem = func_stateT_maplike.mlookup (bid, true) func_state in
    let argvals = sem_exprs fid args mem in
    let mem_in =
      mrestrict locT_setoid locT_setoidDec valueT_setoid valueT_poset
        memT_setoid memT_poset
        (withtop_hasBot
          (map_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec
            valueT_setoid valueT_poset valueT_hasBot)
          (map_poset locT_setoid locT_poset locT_setoidDec locT_posetDec
            valueT_setoid valueT_poset valueT_hasBot)
          (map_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec
            valueT_setoid valueT_poset valueT_hasBot)) memT_maplike
        memT_restricting_maplike locs' mem
    in
    let mem_bound = sem_bind_params acc nfid argvals (fparams nfunc) mem_in
    in
    leb memT_setoid memT_poset memT_setoidDec memT_posetDec mem_bound
      (func_stateT_maplike.mlookup ((fentry nfunc), false)
        (stateT'_maplike.mlookup nfid state))
  | None -> true

(** val postfix_inter_exit :
    progT -> accT -> func_stateT -> stateT' -> fidT -> funcT -> bidT -> exprT
    -> bool **)

let postfix_inter_exit pos2str prog acc func_state state fid func bid ret =
  let mem = func_stateT_maplike.mlookup (bid,true) func_state in
  let result = sem_expr fid ret mem in
  set_forall (pair_setoid positive_setoid positive_setoid)
    (pair_poset positive_setoid positive_setoid positive_poset
      positive_poset)
    (pair_setoidDec positive_setoid positive_setoid positive_setoidDec
      positive_setoidDec)
    (pair_posetDec positive_setoid positive_setoid positive_poset
      positive_poset positive_setoidDec positive_setoidDec positive_posetDec
      positive_posetDec) (fun callinfo ->
    match find
            (order_OrderedType positive_setoid positive_poset
              positive_setoidDec positive_posetDec)
            (coq_MapAVL_FMap
              (order_OrderedType positive_setoid positive_poset
                positive_setoidDec positive_posetDec)) (fst callinfo)
            (pfuncs prog) with
    | Some nfunc ->
      (match find
               (order_OrderedType positive_setoid positive_poset
                 positive_setoidDec positive_posetDec)
               (coq_MapAVL_FMap
                 (order_OrderedType positive_setoid positive_poset
                   positive_setoidDec positive_posetDec)) (snd callinfo)
               (fcmds nfunc) with
       | Some cmd ->
         (match ctmn cmd with
          | Coq_tmn_call (nbid, nextfids, args, retto) ->
            let mem_removed =
              if not_recursive acc fid
              then mrestrict locT_setoid locT_setoidDec valueT_setoid
                     valueT_poset memT_setoid memT_poset
                     (withtop_hasBot
                       (map_setoid locT_setoid locT_poset locT_setoidDec
                         locT_posetDec valueT_setoid valueT_poset
                         valueT_hasBot)
                       (map_poset locT_setoid locT_poset locT_setoidDec
                         locT_posetDec valueT_setoid valueT_poset
                         valueT_hasBot)
                       (map_hasBot locT_setoid locT_poset locT_setoidDec
                         locT_posetDec valueT_setoid valueT_poset
                         valueT_hasBot)) memT_maplike
                     memT_restricting_maplike (filter_fid fid) mem
              else mem
            in
            let mem_bound =
              sem_bind_ret acc (fst callinfo) result retto mem_removed
            in
            if leb memT_setoid memT_poset memT_setoidDec memT_posetDec mem_bound
              (func_stateT_maplike.mlookup (nbid,false)
                (stateT'_maplike.mlookup (fst callinfo) state))
            then true
            else (prerr_endline "postfix_inter_exit failed"; false)
          | _ -> true)
       | None -> true)
    | None -> true)
    ((map_maplike positive_setoid positive_poset positive_setoidDec
       positive_posetDec
       (set_setoid (pair_setoid positive_setoid positive_setoid)
         (pair_poset positive_setoid positive_setoid positive_poset
           positive_poset)
         (pair_setoidDec positive_setoid positive_setoid positive_setoidDec
           positive_setoidDec)
         (pair_posetDec positive_setoid positive_setoid positive_poset
           positive_poset positive_setoidDec positive_setoidDec
           positive_posetDec positive_posetDec))
       (set_poset (pair_setoid positive_setoid positive_setoid)
         (pair_poset positive_setoid positive_setoid positive_poset
           positive_poset)
         (pair_setoidDec positive_setoid positive_setoid positive_setoidDec
           positive_setoidDec)
         (pair_posetDec positive_setoid positive_setoid positive_poset
           positive_poset positive_setoidDec positive_setoidDec
           positive_posetDec positive_posetDec))
       (set_hasBot (pair_setoid positive_setoid positive_setoid)
         (pair_poset positive_setoid positive_setoid positive_poset
           positive_poset)
         (pair_setoidDec positive_setoid positive_setoid positive_setoidDec
           positive_setoidDec)
         (pair_posetDec positive_setoid positive_setoid positive_poset
           positive_poset positive_setoidDec positive_setoidDec
           positive_posetDec positive_posetDec))).mlookup fid acc.callfroms)

(** val postfix_inter_block :
    progT -> accT -> func_stateT -> stateT' -> fidT -> funcT -> bidT -> cmdT
    -> bool **)

let postfix_inter_block pos2str prog acc func_state state fid func bid cmd =
  match snd cmd with
  | Coq_tmn_jump nbids -> true
  | Coq_tmn_call (nbid, nfids, args, retto) ->
    if
    set_forall positive_setoid positive_poset positive_setoidDec
      positive_posetDec (fun nfid ->
      postfix_inter_call pos2str prog acc func_state state fid func bid nbid nfid
        args) nfids
    then true
    else (prerr_endline ("postfix_inter_block failed: call case "^(Pprinter.pprint_bidT bid)); false)
  | Coq_tmn_exit ret ->
    if
    postfix_inter_exit pos2str prog acc func_state state fid func bid ret
    then true
    else (prerr_endline ("postfix_inter_block failed: ret case "^(Pprinter.pprint_bidT bid)); false)

(** val postfix_inter_func :
    progT -> accT -> stateT' -> fidT -> funcT -> bool **)

let postfix_inter_func pos2str prog acc state fid func =
  map_forall positive_setoid positive_poset positive_setoidDec
    positive_posetDec
    (postfix_inter_block pos2str prog acc (stateT'_maplike.mlookup fid state) state
      fid func) (fcmds func)

(** val notop : func_stateT -> bool **)

let notop state =
  map_forall (pair_setoid positive_setoid bool_setoid)
    (pair_poset positive_setoid bool_setoid positive_poset bool_poset)
    (pair_setoidDec positive_setoid bool_setoid positive_setoidDec
      bool_setoidDec)
    (pair_posetDec positive_setoid bool_setoid positive_poset bool_poset
      positive_setoidDec bool_setoidDec positive_posetDec bool_posetDec)
    (fun x mem ->
    match mem with
    | Elt y -> true
    | Top -> false) state

(** val validate_func :
    progT -> accT -> fidT -> funcT -> func_stateT -> func_stateT -> stateT'
    -> bool **)

let validate_func pos2str prog acc fid func func_state propagated_func_state propagated_inter_state =
  (&&) (acc_correct_callfroms_func acc fid func)
    ((&&) (acc_correct_acclocs_star_func acc fid)
      ((&&) (acc_correct_acclocs_func pos2str acc fid func propagated_func_state)
        ((&&) (acc_correct_acclocs_callfroms_func acc fid)
          ((&&) (postfix_func pos2str prog acc propagated_func_state fid func)
            ((&&)
              (eqb_inter_func fid func propagated_func_state
                (stateT'_maplike.mlookup fid propagated_inter_state))
              (use_le_func acc fid func propagated_func_state func_state))))))

(** val validate : progT -> accT -> stateT' -> stateT' -> stateT' -> bool **)

let validate pos2str prog acc state propagated_state propagated_inter_state =
  (&&)
    (map_forall positive_setoid positive_poset positive_setoidDec
      positive_posetDec (fun fid func ->
      (&&)
        (validate_func pos2str prog acc fid func (stateT'_maplike.mlookup fid state)
          (stateT'_maplike.mlookup fid propagated_state)
          propagated_inter_state)
        (postfix_inter_func pos2str prog acc propagated_inter_state fid func))
      (pfuncs prog))
    ((&&)
      (map_forall positive_setoid positive_poset positive_setoidDec
        positive_posetDec (fun fid x ->
        isSome
          (find
            (order_OrderedType positive_setoid positive_poset
              positive_setoidDec positive_posetDec)
            (coq_MapAVL_FMap
              (order_OrderedType positive_setoid positive_poset
                positive_setoidDec positive_posetDec)) fid (pfuncs prog)))
        acc.callfroms)
      (map_forall positive_setoid positive_poset positive_setoidDec
        positive_posetDec (fun x func_state -> notop func_state)
        propagated_state))

