open DomAbsMem
open DomAbsValue
open DomAcc
open DomBasic
open DomBridge
open DomMap
open DomSet
open DomType
open List0
open SemAbsBlock
open SetAVLInstance
open SetInterface
open Syntax

(** val acc_correct_callfroms_blk : accT -> fidT -> bidT -> cmdT -> bool **)

let acc_correct_callfroms_blk acc fid bid cmd =
  match ctmn cmd with
  | Coq_tmn_call (nextbid, nfids, args, retto) ->
    set_forall positive_setoid positive_poset positive_setoidDec
      positive_posetDec (fun nfid ->
      (set_setlike (pair_setoid positive_setoid positive_setoid)
        (pair_poset positive_setoid positive_setoid positive_poset
          positive_poset)
        (pair_setoidDec positive_setoid positive_setoid positive_setoidDec
          positive_setoidDec)
        (pair_posetDec positive_setoid positive_setoid positive_poset
          positive_poset positive_setoidDec positive_setoidDec
          positive_posetDec positive_posetDec)).smem (fid, bid)
        ((map_maplike positive_setoid positive_poset positive_setoidDec
           positive_posetDec
           (set_setoid (pair_setoid positive_setoid positive_setoid)
             (pair_poset positive_setoid positive_setoid positive_poset
               positive_poset)
             (pair_setoidDec positive_setoid positive_setoid
               positive_setoidDec positive_setoidDec)
             (pair_posetDec positive_setoid positive_setoid positive_poset
               positive_poset positive_setoidDec positive_setoidDec
               positive_posetDec positive_posetDec))
           (set_poset (pair_setoid positive_setoid positive_setoid)
             (pair_poset positive_setoid positive_setoid positive_poset
               positive_poset)
             (pair_setoidDec positive_setoid positive_setoid
               positive_setoidDec positive_setoidDec)
             (pair_posetDec positive_setoid positive_setoid positive_poset
               positive_poset positive_setoidDec positive_setoidDec
               positive_posetDec positive_posetDec))
           (set_hasBot (pair_setoid positive_setoid positive_setoid)
             (pair_poset positive_setoid positive_setoid positive_poset
               positive_poset)
             (pair_setoidDec positive_setoid positive_setoid
               positive_setoidDec positive_setoidDec)
             (pair_posetDec positive_setoid positive_setoid positive_poset
               positive_poset positive_setoidDec positive_setoidDec
               positive_posetDec positive_posetDec))).mlookup nfid
          acc.callfroms)) nfids
  | _ -> true

(** val acc_correct_callfroms_func : accT -> fidT -> funcT -> bool **)

let acc_correct_callfroms_func acc fid fbody =
  map_forall positive_setoid positive_poset positive_setoidDec
    positive_posetDec (fun bid cmd ->
    acc_correct_callfroms_blk acc fid bid cmd) (fcmds fbody)

(** val acc_correct_acclocs_var : accT -> fidT -> varT -> memT -> bool **)

let acc_correct_acclocs_var pos2str acc fid x mem =
  let alocs =
    (map_maplike positive_setoid positive_poset positive_setoidDec
      positive_posetDec
      (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
      (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
      (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)).mlookup
      fid acc.acclocs
  in
  if 
  (set_setlike locT_setoid locT_poset locT_setoidDec locT_posetDec).smem
    (var_loc fid x []) alocs
  then true
  else 
    (prerr_endline ("acc_correct_acclocs_var: failed "^
                       (Pprinter.pprint_var pos2str x));
     false)

(** val acc_correct_acclocs_site : accT -> fidT -> allocT -> memT -> bool **)

let acc_correct_acclocs_site acc fid site mem =
  let alocs =
    (map_maplike positive_setoid positive_poset positive_setoidDec
      positive_posetDec
      (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
      (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
      (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)).mlookup
      fid acc.acclocs
  in
  (set_setlike locT_setoid locT_poset locT_setoidDec locT_posetDec).smem
    (lheap site []) alocs

(** val acc_correct_acclocs_expr : accT -> fidT -> exprT -> memT -> bool **)

let acc_correct_acclocs_expr pos2str acc =
  let rec acc_correct_acclocs_expr0 pos2str fid e mem =
    match e with
    | Coq_expr_const l -> true
    | Coq_expr_op2 (o, e1, e2) ->
      (&&) (acc_correct_acclocs_expr0 pos2str fid e1 mem)
        (acc_correct_acclocs_expr0 pos2str fid e2 mem)
    | Coq_expr_op1 (o, e1) -> acc_correct_acclocs_expr0 pos2str fid e1 mem
    | Coq_expr_value_of lv0 ->
      let alocs =
        (map_maplike positive_setoid positive_poset positive_setoidDec
          positive_posetDec
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)).mlookup
          fid acc.acclocs
      in
      (&&) (acc_correct_acclocs_lv0 pos2str fid lv0 mem)
        (match sem_expr_lv fid lv0 mem with
        | Elt locs ->
          (coq_SetAVL_FSet
             (order_OrderedType locT_setoid locT_poset locT_setoidDec
                locT_posetDec)).subset locs alocs
        | Top -> false)
    | Coq_expr_addr_of lv0 -> acc_correct_acclocs_lv0 pos2str fid lv0 mem
    | Coq_expr_cast (t, e1) -> acc_correct_acclocs_expr0 pos2str fid e1 mem
    | Coq_expr_size_of e1 -> acc_correct_acclocs_expr0 pos2str fid e1 mem
  and acc_correct_acclocs_lv0 pos2str fid l mem =
    match l with
    | Coq_lv_id x -> true
    | Coq_lv_deref e1 -> acc_correct_acclocs_expr0 pos2str fid e1 mem
    | Coq_lv_array (e1, e2) ->
      (&&) (acc_correct_acclocs_expr0 pos2str fid e1 mem)
        (acc_correct_acclocs_expr0 pos2str fid e2 mem)
    | Coq_lv_struct (l0, pos) -> acc_correct_acclocs_lv0 pos2str fid l0 mem
  in acc_correct_acclocs_expr0 pos2str

(** val acc_correct_acclocs_lv : accT -> fidT -> lv -> memT -> bool **)

let acc_correct_acclocs_lv pos2str acc =
  let rec acc_correct_acclocs_expr0 pos2str fid e mem =
    match e with
    | Coq_expr_const l -> true
    | Coq_expr_op2 (o, e1, e2) ->
      (&&) (acc_correct_acclocs_expr0 pos2str fid e1 mem)
        (acc_correct_acclocs_expr0 pos2str fid e2 mem)
    | Coq_expr_op1 (o, e1) -> acc_correct_acclocs_expr0 pos2str fid e1 mem
    | Coq_expr_value_of lv0 ->
      let alocs =
        (map_maplike positive_setoid positive_poset positive_setoidDec
          positive_posetDec
          (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
          (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)).mlookup
          fid acc.acclocs
      in
      (&&) (acc_correct_acclocs_lv0 pos2str fid lv0 mem)
        (match sem_expr_lv fid lv0 mem with
         | Elt locs ->
           (coq_SetAVL_FSet
             (order_OrderedType locT_setoid locT_poset locT_setoidDec
                locT_posetDec)).subset locs alocs
        | Top -> false)
    | Coq_expr_addr_of lv0 -> acc_correct_acclocs_lv0 pos2str fid lv0 mem
    | Coq_expr_cast (t, e1) -> acc_correct_acclocs_expr0 pos2str fid e1 mem
    | Coq_expr_size_of e1 -> acc_correct_acclocs_expr0 pos2str fid e1 mem
  and acc_correct_acclocs_lv0 pos2str fid l mem =
    match l with
    | Coq_lv_id x -> true
    | Coq_lv_deref e1 -> acc_correct_acclocs_expr0 pos2str fid e1 mem
    | Coq_lv_array (e1, e2) ->
      (&&) (acc_correct_acclocs_expr0 pos2str fid e1 mem)
        (acc_correct_acclocs_expr0 pos2str fid e2 mem)
    | Coq_lv_struct (l0, pos) -> acc_correct_acclocs_lv0 pos2str fid l0 mem
  in acc_correct_acclocs_lv0 pos2str

(** val acc_correct_acclocs_exprs :
    accT -> fidT -> exprT list -> memT -> bool **)

let rec acc_correct_acclocs_exprs pos2str acc fid es mem =
  match es with
  | [] -> true
  | e::es1 ->
    (&&) (acc_correct_acclocs_expr pos2str acc fid e mem)
      (acc_correct_acclocs_exprs pos2str acc fid es1 mem)

(** val acc_correct_acclocs_libcall_args :
    accT -> fidT -> exprT list -> memT -> bool **)

let rec acc_correct_acclocs_libcall_args acc fid args mem =
  match args with
  | [] -> true
  | arg :: tl0 ->
    (&&)
      (match locs_of (sem_expr fid arg mem) with
       | Elt locs ->
         let alocs =
           (map_maplike positive_setoid positive_poset positive_setoidDec
             positive_posetDec
             (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
             (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
             (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)).mlookup
             fid acc.acclocs
         in
         (coq_SetAVL_FSet
           (order_OrderedType locT_setoid locT_poset locT_setoidDec
             locT_posetDec)).subset locs alocs
       | Top -> false) (acc_correct_acclocs_libcall_args acc fid tl0 mem)

(** val acc_correct_acclocs_instr :
    accT -> fidT -> instrT -> memT -> bool **)

let acc_correct_acclocs_instr pos2str acc fid instr mem =
  match instr with
  | Coq_instr_assign (lv0, e) ->
    if
    (&&)
      ((&&) 
          (if acc_correct_acclocs_lv pos2str acc fid lv0 mem
           then true
           else (prerr_endline "acc_correct_acclocs_inst:assign lv"; false))
          (if acc_correct_acclocs_expr pos2str acc fid e mem
           then true
           else (prerr_endline "acc_correct_acclocs_inst:assign e"; false)))
      (if (let alocs =
            (map_maplike positive_setoid positive_poset positive_setoidDec
               positive_posetDec
               (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
               (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
               (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)).mlookup
              fid acc.acclocs
          in
          match sem_expr_lv fid lv0 mem with
          | Elt locs ->
            (if
            ((coq_SetAVL_FSet
               (order_OrderedType locT_setoid locT_poset locT_setoidDec
                  locT_posetDec)).subset locs alocs)
             then true
             else (prerr_endline "elt case"; 
                   prerr_endline 
                     (Orders.locTSetFold
                        (fun loc acc -> 
                          acc^" "^(Pprinter.pprint_locT pos2str loc)
                        )
                        locs "");
                   prerr_endline 
                     (Orders.locTSetFold
                        (fun loc acc -> 
                          acc^" "^(Pprinter.pprint_locT pos2str loc)
                        )
                        alocs "");
                   false))
          | Top -> 
            (prerr_endline "top case"; false))
       then true
       else (prerr_endline "acc_correct_acclocs_inst:assign sem_lv"; false))
    then true
    else (prerr_endline "acc_correct_acclocs_inst:assign"; false)
  | Coq_instr_alloc (lv0, a, site) ->
    let Coq_alloc_arr (sz, str) = a in
    (&&)
      ((&&)
        ((&&)
          ((&&) 
              (if acc_correct_acclocs_lv pos2str acc fid lv0 mem
               then true 
               else (prerr_endline "acc_correct_acclocs_inst:alloc lv"; false))
              (if acc_correct_acclocs_site acc fid site mem
               then true 
               else (prerr_endline "acc_correct_acclocs_inst:alloc site"; false)))
          (if acc_correct_acclocs_expr pos2str acc fid sz mem
           then true 
           else (prerr_endline "acc_correct_acclocs_inst:alloc e1"; false)))
        (if acc_correct_acclocs_expr pos2str acc fid str mem
         then true 
         else (prerr_endline "acc_correct_acclocs_inst:alloc e2"; false)))
      (let alocs =
         (map_maplike positive_setoid positive_poset positive_setoidDec
           positive_posetDec
           (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
           (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
           (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)).mlookup
           fid acc.acclocs
       in
       match sem_expr_lv fid lv0 mem with
       | Elt locs ->
         (if
         (coq_SetAVL_FSet
           (order_OrderedType locT_setoid locT_poset locT_setoidDec
             locT_posetDec)).subset locs alocs
          then true
          else (prerr_endline "acc_correct_acclocs_inst:locs"; 
                false))
       | Top -> 
         (prerr_endline "acc_correct_acclocs_inst:top"; false))
  | Coq_instr_assume e -> 
    if acc_correct_acclocs_expr pos2str acc fid e mem
    then true
    else (prerr_endline "acc_correct_acclocs_inst:assume"; false)
  | Coq_instr_libcall (ol, lc, args, site) ->
    if
    (&&)
      ((&&)
        ((&&)
          (match lc with
           | Coq_scanf -> true
           | _ ->
             (match lc with
              | Coq_strcpy ->
                let dstLocs =
                  sem_expr fid (hd (Coq_expr_const Coq_lit_bot) args) mem
                in
                let dst = locs_of dstLocs in
                let srcLocs =
                  sem_expr fid (hd (Coq_expr_const Coq_lit_bot) (tl args))
                    mem
                in
                let src = locs_of srcLocs in
                let srcValue = memory_lookup src mem in
                let newMem = memory_update acc dst srcValue mem in
                (match ol with
                 | Some lv0 -> acc_correct_acclocs_lv pos2str acc fid lv0 newMem
                 | None -> true)
              | _ ->
                (match ol with
                 | Some lv0 -> acc_correct_acclocs_lv pos2str acc fid lv0 mem
                 | None -> true)))
          (match lc with
           | Coq_strcpy -> acc_correct_acclocs_exprs pos2str acc fid args mem
           | Coq_scanf -> acc_correct_acclocs_exprs pos2str acc fid args mem
           | Coq_builtin_alloca -> acc_correct_acclocs_exprs pos2str acc fid args mem
           | Coq_builtin_choose_expr ->
             acc_correct_acclocs_exprs pos2str acc fid args mem
           | _ -> true))
        (match lc with
         | Coq_scanf -> true
         | _ ->
           (match lc with
            | Coq_strcpy ->
              let dstLocs =
                sem_expr fid (hd (Coq_expr_const Coq_lit_bot) args) mem
              in
              let dst = locs_of dstLocs in
              let srcLocs =
                sem_expr fid (hd (Coq_expr_const Coq_lit_bot) (tl args)) mem
              in
              let src = locs_of srcLocs in
              let srcValue = memory_lookup src mem in
              let newMem = memory_update acc dst srcValue mem in
              (match ol with
               | Some lv0 ->
                 let alocs =
                   (map_maplike positive_setoid positive_poset
                     positive_setoidDec positive_posetDec
                     (set_setoid locT_setoid locT_poset locT_setoidDec
                       locT_posetDec)
                     (set_poset locT_setoid locT_poset locT_setoidDec
                       locT_posetDec)
                     (set_hasBot locT_setoid locT_poset locT_setoidDec
                       locT_posetDec)).mlookup fid acc.acclocs
                 in
                 (match sem_expr_lv fid lv0 newMem with
                  | Elt locs ->
                    (coq_SetAVL_FSet
                      (order_OrderedType locT_setoid locT_poset
                        locT_setoidDec locT_posetDec)).subset locs alocs
                  | Top -> false)
               | None -> true)
            | _ ->
              (match ol with
               | Some lv0 ->
                 let alocs =
                   (map_maplike positive_setoid positive_poset
                     positive_setoidDec positive_posetDec
                     (set_setoid locT_setoid locT_poset locT_setoidDec
                       locT_posetDec)
                     (set_poset locT_setoid locT_poset locT_setoidDec
                       locT_posetDec)
                     (set_hasBot locT_setoid locT_poset locT_setoidDec
                       locT_posetDec)).mlookup fid acc.acclocs
                 in
                 (match sem_expr_lv fid lv0 mem with
                  | Elt locs ->
                    (coq_SetAVL_FSet
                      (order_OrderedType locT_setoid locT_poset
                        locT_setoidDec locT_posetDec)).subset locs alocs
                  | Top -> false)
               | None -> true))))
      (match lc with
       | Coq_strcpy -> acc_correct_acclocs_libcall_args acc fid args mem
       | Coq_scanf -> acc_correct_acclocs_libcall_args acc fid args mem
       | _ -> true)
    then true
    else (prerr_endline "acc_correct_acclocs_inst:libcall"; false)
  | Coq_instr_forget aloc ->
    let alocs =
      (map_maplike positive_setoid positive_poset positive_setoidDec
        positive_posetDec
        (set_setoid locT_setoid locT_poset locT_setoidDec locT_posetDec)
        (set_poset locT_setoid locT_poset locT_setoidDec locT_posetDec)
        (set_hasBot locT_setoid locT_poset locT_setoidDec locT_posetDec)).mlookup
        fid acc.acclocs
    in
    if (set_setlike locT_setoid locT_poset locT_setoidDec locT_posetDec).smem
      aloc alocs
    then true
    else (prerr_endline "acc_correct_acclocs_inst:forget"; false)
  | Coq_instr_skip -> true

(** val acc_correct_acclocs_instrs :
    accT -> fidT -> instrT list -> memT -> bool **)

let rec acc_correct_acclocs_instrs pos2str acc fid instrs mem =
  match instrs with
  | [] -> true
  | instr::next_instrs ->
    (&&) (if acc_correct_acclocs_instr pos2str acc fid instr mem
      then true
      else (prerr_endline 
            ("acc_correct_acclocs_instrs failed: "^
                (Pprinter.pprint_instrT pos2str instr));
          false))
      (let next_mem = sem_instr acc fid instr mem in
         acc_correct_acclocs_instrs pos2str acc fid next_instrs next_mem)

(** val acc_correct_acclocs_tmn : accT -> fidT -> tmnT -> memT -> bool **)

let acc_correct_acclocs_tmn pos2str acc fid tmn mem =
  match tmn with
  | Coq_tmn_jump nextbids -> true
  | Coq_tmn_call (nextbid, nextfids, es, ret) ->
    (&&) (acc_correct_acclocs_exprs pos2str acc fid es mem)
      (match ret with
       | Some ret' -> acc_correct_acclocs_var pos2str acc fid ret' mem
       | None -> true)
  | Coq_tmn_exit e -> acc_correct_acclocs_expr pos2str acc fid e mem
