open Datatypes
open List0
open OrderedType
open OrderedTypeEx

module SetList = 
 struct 
  type 'a set = 'a list
  
  (** val empty : 'a1 coq_OrderedType -> 'a1 set **)
  
  let empty comparedec =
    []
  
  (** val is_empty : 'a1 coq_OrderedType -> 'a1 set -> bool **)
  
  let is_empty comparedec = function
  | [] -> true
  | x :: x0 -> false
  
  (** val mem : 'a1 coq_OrderedType -> 'a1 -> 'a1 set -> bool **)
  
  let rec mem comparedec x = function
  | [] -> false
  | y :: l ->
    (match compare comparedec x y with
     | Eq -> true
     | Lt -> false
     | Gt -> mem comparedec x l)
  
  (** val add : 'a1 coq_OrderedType -> 'a1 -> 'a1 set -> 'a1 set **)
  
  let rec add comparedec x s = match s with
  | [] -> x :: []
  | y :: l ->
    (match compare comparedec x y with
     | Eq -> s
     | Lt -> x :: s
     | Gt -> y :: (add comparedec x l))
  
  (** val singleton : 'a1 coq_OrderedType -> 'a1 -> 'a1 set **)
  
  let singleton comparedec x =
    x :: []
  
  (** val remove : 'a1 coq_OrderedType -> 'a1 -> 'a1 set -> 'a1 set **)
  
  let rec remove comparedec x s = match s with
  | [] -> []
  | y :: l ->
    (match compare comparedec x y with
     | Eq -> l
     | Lt -> s
     | Gt -> y :: (remove comparedec x l))
  
  (** val union : 'a1 coq_OrderedType -> 'a1 set -> 'a1 set -> 'a1 set **)
  
  let rec union comparedec s = match s with
  | [] -> (fun s' -> s')
  | x :: l ->
    let rec union_aux s' = match s' with
    | [] -> s
    | x' :: l' ->
      (match compare comparedec x x' with
       | Eq -> x :: (union comparedec l l')
       | Lt -> x :: (union comparedec l s')
       | Gt -> x' :: (union_aux l'))
    in union_aux
  
  (** val inter : 'a1 coq_OrderedType -> 'a1 set -> 'a1 set -> 'a1 set **)
  
  let rec inter comparedec = function
  | [] -> (fun x -> [])
  | x :: l ->
    let rec inter_aux s' = match s' with
    | [] -> []
    | x' :: l' ->
      (match compare comparedec x x' with
       | Eq -> x :: (inter comparedec l l')
       | Lt -> inter comparedec l s'
       | Gt -> inter_aux l')
    in inter_aux
  
  (** val diff : 'a1 coq_OrderedType -> 'a1 set -> 'a1 set -> 'a1 set **)
  
  let rec diff comparedec s = match s with
  | [] -> (fun x -> [])
  | x :: l ->
    let rec diff_aux s' = match s' with
    | [] -> s
    | x' :: l' ->
      (match compare comparedec x x' with
       | Eq -> diff comparedec l l'
       | Lt -> x :: (diff comparedec l s')
       | Gt -> diff_aux l')
    in diff_aux
  
  (** val equal : 'a1 coq_OrderedType -> 'a1 set -> 'a1 set -> bool **)
  
  let rec equal comparedec s s' =
    match s with
    | [] ->
      (match s' with
       | [] -> true
       | e :: l -> false)
    | x :: l ->
      (match s' with
       | [] -> false
       | x' :: l' ->
         (match compare comparedec x x' with
          | Eq -> equal comparedec l l'
          | _ -> false))
  
  (** val subset : 'a1 coq_OrderedType -> 'a1 set -> 'a1 set -> bool **)
  
  let rec subset comparedec s s' =
    match s with
    | [] -> true
    | x :: l ->
      (match s' with
       | [] -> false
       | x' :: l' ->
         (match compare comparedec x x' with
          | Eq -> subset comparedec l l'
          | Lt -> false
          | Gt -> subset comparedec s l'))
  
  (** val fold :
      'a1 coq_OrderedType -> ('a1 -> 'a2 -> 'a2) -> 'a1 set -> 'a2 -> 'a2 **)
  
  let rec fold comparedec f s i =
    match s with
    | [] -> i
    | x :: l -> fold comparedec f l (f x i)
  
  (** val map_monotonic :
      'a1 coq_OrderedType -> 'a2 coq_OrderedType -> ('a1 -> 'a2) -> 'a1 set
      -> 'a2 set **)
  
  let map_monotonic comparedec h f s =
    map f s
  
  (** val filter :
      'a1 coq_OrderedType -> ('a1 -> bool) -> 'a1 set -> 'a1 set **)
  
  let rec filter comparedec f = function
  | [] -> []
  | x :: l ->
    if f x then x :: (filter comparedec f l) else filter comparedec f l
  
  (** val for_all :
      'a1 coq_OrderedType -> ('a1 -> bool) -> 'a1 set -> bool **)
  
  let rec for_all comparedec f = function
  | [] -> true
  | x :: l -> if f x then for_all comparedec f l else false
  
  (** val exists_ :
      'a1 coq_OrderedType -> ('a1 -> bool) -> 'a1 set -> bool **)
  
  let rec exists_ comparedec f = function
  | [] -> false
  | x :: l -> if f x then true else exists_ comparedec f l
  
  (** val partition :
      'a1 coq_OrderedType -> ('a1 -> bool) -> 'a1 set -> 'a1 set * 'a1 set **)
  
  let rec partition comparedec f = function
  | [] -> ([], [])
  | x :: l ->
    let (s1, s2) = partition comparedec f l in
    if f x then ((x :: s1), s2) else (s1, (x :: s2))
  
  (** val cardinal : 'a1 coq_OrderedType -> 'a1 set -> int **)
  
  let cardinal comparedec s =
    length s
  
  (** val elements : 'a1 coq_OrderedType -> 'a1 set -> 'a1 list **)
  
  let elements comparedec x =
    x
  
  (** val min_elt : 'a1 coq_OrderedType -> 'a1 set -> 'a1 option **)
  
  let min_elt comparedec = function
  | [] -> None
  | x :: l -> Some x
  
  (** val max_elt : 'a1 coq_OrderedType -> 'a1 set -> 'a1 option **)
  
  let rec max_elt comparedec = function
  | [] -> None
  | x :: l ->
    (match l with
     | [] -> Some x
     | e :: l0 -> max_elt comparedec l)
  
  (** val choose : 'a1 coq_OrderedType -> 'a1 set -> 'a1 option **)
  
  let choose comparedec =
    min_elt comparedec
  
  (** val map :
      'a1 coq_OrderedType -> 'a2 coq_OrderedType -> ('a1 -> 'a2) -> 'a1 set
      -> 'a2 set **)
  
  let map h h0 f s =
    fold h (fun a sb -> add h0 (f a) sb) s (empty h0)
  
  (** val set_compare :
      'a1 coq_OrderedType -> 'a1 set -> 'a1 set -> comparison **)
  
  let set_compare h =
    list_compare h
  
  (** val set_OrderedType :
      'a1 coq_OrderedType -> 'a1 set coq_OrderedType **)
  
  let set_OrderedType h =
    list_OrderedType h
 end

