Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.
Require Import List.
Require Import Setoid.
Require Import Morphisms.

Require Import LibTactics.
Require Import DomType.
Require Import DomBasic.
Require Import DomSet.
Require Import DomMap.
Require Import DomBridge.

Require Import Syntax.

Require Import DomAbsValue.
Require Import DomAcc.
Require Import DomAbsMem.
Require Import DomPidxSen.
Require Import DomPidxInsen.

Require Import SemAbsBlock.
Require Import SemAccSen.
Require Import SemAccInsen.

Require Import SemSenLocal.
Require Import SemInsenLocal.

Require Import TrsSen.

Require Import CorAccBasic.
Require Import CorAccExpr.
Require Import CorAccLibcall.
Require Import CorAccPruning.

Set Implicit Arguments.

Section Accs.
  Variables (prog:progT) (acc:accT) (state:stateT)
    (Hacc: SemAccInsen.acc_correct prog acc state).

  Definition tstate' := tstate prog state.

  Lemma tacc_correct_callfroms_prog :
    SemAccSen.acc_correct_callfroms_prog prog acc.
  Proof.
    move/and4P: Hacc=> [Hcf _ _ _].
    move/map_forallP: Hcf.
    refine (implies_right_turn (Proper _ _) _ _ _ _).
    refine (forall_implies _ _ _ _) => fid.
    refine (forall_implies _ _ _ _) => func.
    refine (forall_implies _ _ _ _) => Hfunc.
    move/map_forallP.
    refine (implies_right_turn (Proper _ _) _ _ _ _).
    refine (forall_implies _ _ _ _) => bid.
    refine (forall_implies _ _ _ _) => cmd.
    refine (forall_implies _ _ _ _) => Hcmd.
    rewrite/acc_correct_callfroms_blk/SemAccSen.acc_correct_callfroms_blk.
    by case: (ctmn cmd).
  Qed.  

  Lemma tacc_correct_callfroms_star :
    SemAccSen.acc_correct_callfroms_star acc.
  Proof.
    move/and4P: Hacc=> [_ Hcss _ _].
    move/map_forallP: Hcss.
    refine (implies_right_turn (Proper _ _) _ _ _ _).
    refine (forall_implies _ _ _ _) => fid.
    rewrite/mlookup/map_maplike.
    case Hcffid: ((callfroms acc)[fid])=> [cffid|]; last done.
    move=> Hcf.
    apply MF.find_mapsto_iff in Hcffid.
    move: {Hcf} (Hcf _ Hcffid).
    move/set_forallP.
    refine (implies_right_turn (Proper _ _) _ _ _ _).
    { move=> x y Hxy.
      usual_tac Hxy.
      exact: eq_refl.
    }
    refine (forall_implies _ _ _ _) => ci.
    refine (forall_implies _ _ _ _) => Hci.
    move/andP=> [Hin Hle]; split; first done.
    exact/lebP.
  Qed.

  Lemma tacc_correct_acclocs_assign :
    forall fid stack mem
           (Hpidx: pidx_consistent prog fid stack)
           var e
           (Hinstr: acc_correct_acclocs_instr acc fid (instr_assign var e) mem),
      SemAccSen.acc_correct_acclocs_instr acc fid (instr_assign var e) mem.
  Proof.
    intros.
    rewrite/SemAccSen.acc_correct_acclocs_instr/sem_instr=> mem'.
    apply/DomType.eq_le.
    move/andP: Hinstr=> [Hvare Hinstr].
    move/andP: Hvare=> [Hvar He].

    move: mem'=> [mem'|]; last first.
    { rewrite ? Top_top; dom_tac.
      rewrite/memT_hasJoin/withtop_hasJoin/withtop_join.
      reflexivity.
    }

    exploit tacc_correct_acclocs_lv; eauto.
    instantiate (1 := mem').
    move=> {Hvar} Hvar.
    rewrite Hvar.
    move: {Hvar} (sem_expr_lv _ _ _) Hinstr=> locs Hinstr.
  
    exploit tacc_correct_acclocs_expr; eauto.
    instantiate (1 := mem').
    move=> {He} He.
    rewrite He.
    move: {He} (sem_expr _ _ _)=> val.

    move: locs Hinstr=> [locs|] Hinstr; last first.
    { rewrite/memory_update/=.
      by case: (mrestrict_join _ _ _); case: mem; reflexivity.
    }
                           
    symmetry.
    apply memory_update_mrestrict_join_commutative_subset.
    move: (_[_]!) Hinstr=> alocs Hinstr.
    move=> loc; dom_tac=> Hloc.
    apply SetFacts.mem_iff.
    apply SetFacts.subset_iff in Hinstr.
    apply Hinstr.
    apply SetFacts.mem_iff.
    done.
  Qed.

  Lemma tacc_correct_acclocs_alloc :
    forall fid stack mem
           (Hpidx: pidx_consistent prog fid stack)
           l a site
           (Hinstr: acc_correct_acclocs_instr
                      acc fid (instr_alloc l a site) mem),
      SemAccSen.acc_correct_acclocs_instr acc fid (instr_alloc l a site) mem.
  Proof.
    intros.
    move: a Hinstr=> [sz str] Hinstr.
    { rewrite/SemAccSen.acc_correct_acclocs_instr/sem_instr.
      rewrite/= in Hinstr.
      move/andP: Hinstr=> [Hinstr Hlv'].
      move/andP: Hinstr=> [Hinstr Hstr].
      move/andP: Hinstr=> [Hinstr Hsz].
      move/andP: Hinstr=> [Hlv Hsite].
      move=> [mem'|]; last first.
      { rewrite ? Top_top.
        rewrite [X in le _ X]mem_mrestrict_join_top_rhs.
        exact: top_prop.
      }
      dom_tac.
      rewrite/acc_correct_acclocs_site in Hsite.
      exploit tacc_correct_acclocs_lv; eauto.
      instantiate (1 := mem').
      move=> {Hlv} Hlv.
      rewrite Hlv.
      move: (sem_expr_lv _ _ _) Hlv'=> {Hlv} locs Hlv'.
      move: locs Hlv'=> [locs|] Hlv'; last done.
      exploit tacc_correct_acclocs_expr; eauto.
      instantiate (1 := mem').
      move: (sem_expr _ _ _) (sem_expr _ _ _)=> vsz vsz' {Hsz} Hsz.
      exploit tacc_correct_acclocs_expr; eauto.
      instantiate (1 := mem').
      move: (sem_expr _ _ _) (sem_expr _ _ _)=> vstr vstr' {Hstr} Hstr.
      rewrite memory_update_mrestrict_join_commutative_subset; last first.
      { move=> loc /= => Hlocs.
        apply SetFacts.mem_iff.
        apply SetFacts.subset_iff in Hlv'.
        apply Hlv'.
        by apply SetFacts.mem_iff.
      }
      apply memory_update_le_mor; try reflexivity;
      last exact/DomType.eq_le/eq_refl.
      rewrite/eq/eq_/valueT_setoid/pair_setoid/pair_eq.
      split; simpl; first by reflexivity.
      dom_tac.
      rewrite/eq/eq_/map_setoid; move=> key.
      rewrite 2! map_mlookup_spec; rewrite 2! map_madd_spec.
      case: (eqb site key); last by reflexivity.
      assert (eq (itv_of vsz') (itv_of vsz)) as Heqsz.
      { apply itv_of_mor; done. }
      assert (eq (itv_of vstr') (itv_of vstr)) as Heqstr.
      { apply itv_of_mor; done. }
      rewrite{1}/eq/eq_.
      repeat split; simpl; done.
    }
  Qed.

  Lemma mem_mremove_spec (mem:memT) (loc rloc:locT):
    eq (A:=valueT)
       (mremove rloc mem)[loc]!
       (if mem is Elt _
        then (if eqb rloc loc
              then bot
              else mem[loc]!)
        else top).
  Proof.
    move: mem=> [mem|]; last done.
    rewrite/mremove/mlookup/memT_maplike/withtop_maplike.
    rewrite/mremove/mlookup/map_maplike MF.remove_o.
    rewrite is_compare_eqb.
    case: (eqb _ _); first by apply eq_refl.
    reflexivity.
  Qed.
  
  Lemma tacc_correct_acclocs_forget :
    forall fid stack mem
           (Hpidx: pidx_consistent prog fid stack)
           aloc
           (Hinstr: acc_correct_acclocs_instr acc fid (instr_forget aloc) mem),
      SemAccSen.acc_correct_acclocs_instr acc fid (instr_forget aloc) mem.
  Proof.
    move=> fid stack mem Hpidx aloc Hinstr.
    unfold acc_correct_acclocs_instr in Hinstr.
    move=> out; unfold sem_instr.
    move: mem=> [mem|]; last done.
    move: out=> [out|]; last done.
    apply mem_le_spec; [done|done|move=> loc].
    rewrite mem_mremove_spec.
    rewrite mem_mrestrict_join_def.
    rewrite mem_mrestrict_join_spec.
    rewrite {1}/mremove/memT_maplike/withtop_maplike.
    rewrite set_to_powerset_spec.
    case Hloc: (eqb _ _); first by apply bot_prop.
    rewrite{1}/mlookup/memT_maplike.
    rewrite/mrestrict_join map_join_spec.
    rewrite ? map_mrestrict_spec.
    rewrite powerset_neg_spec set_to_powerset_spec.
    case Hin: (_ !\in _); rewrite/negb; last first.
    { rewrite join_commutative join_bot.
      apply DomType.eq_le; apply eq_refl.
    }
    rewrite join_bot.
    rewrite mem_mremove_spec Hloc.
    apply DomType.eq_le; apply eq_refl.
  Qed.

  Lemma tacc_correct_acclocs_instr :
    forall fid stack instr mem
           (Hpidx: pidx_consistent prog fid stack)
           (Hinstr: acc_correct_acclocs_instr acc fid instr mem),
      SemAccSen.acc_correct_acclocs_instr
        acc fid instr mem.
  Proof.
    move=> fid stack instr mem Hc Hinstr.
    case: instr Hinstr=> [var e|l a site|test|ol lc le a|loc|site] Hinstr.
    { eapply tacc_correct_acclocs_assign; eauto. }
    { eapply tacc_correct_acclocs_alloc; eauto. }
    { eapply tacc_correct_acclocs_prune; eauto. }
    { eapply tacc_correct_acclocs_libcall; eauto. }
    { eapply tacc_correct_acclocs_forget; eauto. }
    { exact/DomType.eq_le/eq_refl. }
  Qed.

  Lemma tacc_correct_acclocs_tmn :
    forall fid cmd mem
      (Htmn : acc_correct_acclocs_tmn
                acc fid (ctmn cmd) mem),
      SemAccSen.acc_correct_acclocs_tmn
        acc fid
        (ctmn cmd)
        mem.
  Proof.
    intros. move: Htmn.
    case tcmd: (ctmn cmd) => [nbids|nbid nfids args retto|ret];
    rewrite/SemAccSen.acc_correct_acclocs_tmn/acc_correct_acclocs_tmn//.
    { move/andP => [Hexpr Hvar]; split.
      { by apply tacc_correct_acclocs_exprs. }
      { move: retto tcmd Hvar=> [retto|] tcmd Hvar; last done.
        apply Hvar.
      }
    }
    { by apply tacc_correct_acclocs_expr. }
  Qed.

  Lemma tacc_correct : SemAccSen.acc_correct prog acc tstate'.
  Proof.
    move/and4P: Hacc => [Hcf Hcss Hac Haccf].
    split.
    exact: tacc_correct_callfroms_prog.
    exact: tacc_correct_callfroms_star.
    { move/map_forallP: Hac.
      refine (implies_right_turn (Proper _ _) _ _ _ _).
      refine (forall_implies _ _ _ _) => fid.
      refine (forall_implies _ _ _ _) => func.
      refine (forall_implies _ _ _ _) => Hfunc.
      move/andP=> [Hac Hacp].
      split; last exact/list_forallP.
      move: Hac; move/map_forallP.
      refine (implies_right_turn (Proper _ _) _ _ _ _).
      refine (forall_implies _ _ _ _) => bid.
      refine (forall_implies _ _ _ _) => cmd.
      move=> H stack; move: H.
      refine (forall_implies _ _ _ _) => Hcmd.
      move/andP=> [Hinstrs Htmn].
      move=> Hc.
      rewrite/tstate'.
      { eapply tacc_correct_acclocs_tmn.
        rewrite tstate_lookup Hc /fst.
        by eauto.
      }
    }
    { move/map_forallP: Haccf.
      refine (implies_right_turn (Proper _ _) _ _ _ _).
      refine (forall_implies _ _ _ _) => fid.
      move=> Hyp ci; rewrite/fst.
      rewrite [(callfroms acc)[fid]!]/mlookup/map_maplike.
      case Hccfid: ((callfroms acc)[fid])=> [ccfid|]; last done.
      rewrite/coerce.
      apply MF.find_mapsto_iff in Hccfid.
      move: {Hyp} (Hyp _ Hccfid); move/set_forallP.
      refine (implies_right_turn (Proper _ _) _ _ _ _).
      { move=> x y Hxy.
        usual_tac Hxy.
        exact: eq_refl.
      }
      move=> Hyp Hin.
      by move/lebP: (Hyp _ Hin).
    }
  Qed.
End Accs.
