Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Utf8.

Require Import LibTactics.
Require Import DomBridge.
Require Import DomType.
Require Import DomBasic.
Require Import DomMap.

Require Import Syntax.

Require Import SemCon.
Require Import DomAbsValue.
Require Import DomAcc.
Require Import DomAbsMem.
Require Import DomPidx.
Require Import DomPidxSen.

Require Import SemAbsBlock.
Require Import SemAccSen.
Require Import SemSen.

Require Import DomZ.
Require Import DomSet.
Require Import CorBasic.
Require Import CorAbs.
Require Import CorItv.
Require Import CorConExpr.
Require Import CorConPruning.
Require Import CorConLibcall.

Set Implicit Arguments.

Obligation Tactic := Simpless.tactic.

Section Correctness.
Variable (prog:progT) (acc:accT).
  
Section MemoryAbs.
  Lemma sem_instr_proper2 : forall 
    (Hacc_prog: acc_correct_callfroms_prog prog acc)
    (Hacc_star: acc_correct_callfroms_star acc)
    (pidx:pidxT) (Hpidx_con: pidx_consistent prog (DomPidx.pifid pidx) pidx.2)
    (instr:instrT)
    (m:DomConMem.memT) (am:memT) (Hmem : m \abs am)
    (m':DomConMem.memT) (Hcon:SemCon.sem_instr pidx instr m m'),
    m' \abs sem_instr acc (DomPidx.pifid pidx) instr am.
  Proof.
    move=> Hcf Hcs pidx Hpidx i.
    destruct i=> cm am Hm cm' Hsem; inv Hsem.
    { rewrite <- Hsubst.
      eapply memory_update_sound; eauto.
      { by eapply sem_expr_lv_proper2_; eauto. }
      { by eapply sem_expr_proper2_; eauto. }
      { by rewrite Hsubst. }
    }
    { rewrite <- Hsubst.
      eapply memory_update_sound; eauto.
      { by eapply sem_expr_lv_proper2_; eauto. }
      { rewrite/abs_relation/option_abs.
        rewrite/DomConValue.vheap.
        rewrite/abs_relation/value_abs.
        rewrite/arrv_of/fst/snd.
        rewrite/mlookup/withtop_maplike.
        rewrite/mlookup/map_maplike.
        rewrite MF.add_o is_compare_eqb eqb_refl /coerce.
        (do ? split); rewrite/fst/snd.
        { by constructor; omega. }
        { by apply (sem_expr_proper2_ prog acc Hsz am). }
        { by apply (sem_expr_proper2_ prog acc Hst am). }
      }
      { by rewrite Hsubst. }
    }
    { eapply pruning_sound; eauto.
      by rewrite <- Heq.
      by rewrite <- Heq.
    }
    { by eapply sound_libcall; eauto. }
    { rewrite <- Hforget; unfold sem_instr.
      clear Hforget Hwf_newmem cm'.
      move: am Hm=> [am|] Hm; last done.
      unfold mremove, memT_maplike, withtop_maplike.
      move=> loc; dom_tac.
      case Hloc: (_ !\in _); unfold negb; first done.
      eapply abs_relation_eq_mor; [by apply eq_refl|idtac|eauto].
      rewrite/gamma_loc/= in Hloc.
      rewrite{2}/mlookup/memT_maplike/withtop_maplike.
      rewrite/mremove/mlookup/map_maplike.
      rewrite MapFacts.remove_o.
      rewrite is_compare_eqb eqb_sym Hloc.
      reflexivity.
    }
    { by rewrite <- Heq. }
  Qed.

  Lemma madd_wf (pidx:pidxT) (cm:DomConMem.memT) (loc:DomConValue.locT) (Hloc: DomConMem.loc_wf pidx loc) (v:DomConValue.valueT) :
    eq (DomConMem.mem_wf pidx cm)
       (DomConMem.mem_wf pidx (cm[loc <- Some v]!)).
  Proof.
    constructor; move=> Hwf loc' Hloc'.
    { rewrite morphism_madd_spec in Hloc'.
      case Heq: (eqb _ _) in Hloc'.
      { by move/eqbP in Heq; usual_tac Heq. }
      { by apply Hwf. }
    }
    { case Heq: (eqb loc' loc).
      { by move/eqbP in Heq; usual_tac Heq. }
      { exploit (Hwf loc'); last done.
        by rewrite morphism_madd_spec Heq.
      }
    }
  Qed.

  Lemma sem_bind_params_proper2 : forall 
    (Hacc_prog: acc_correct_callfroms_prog prog acc)
    (Hacc_star: acc_correct_callfroms_star acc)
    (pidx:pidxT) (Hpidx_con: pidx_consistent prog (DomPidx.pifid pidx) pidx.2)
    (cvs : list DomConValue.valueT)
    (avs : list valueT)
    (Hvs : (let (f, _) := list_abs in f) cvs avs)
    (params : list idT)
    (m m' : DomConMem.memT)
    (am : memT) (Hxy : m \abs am)
    (Hsem : SemCon.sem_bind_params pidx cvs params m m'),
    m' \abs sem_bind_params acc (DomPidx.pifid pidx) avs params am.
  Proof.
    move=> Hcf Hcs pidx Hpidx.
    elim=> [|cv cvs IH] [|av avs] // Hvs params cm cm' am Hm Hbind; inv Hbind; first by rewrite <- Heq.
    move: Hvs=> [Hv Hvs].
    eapply (IH avs); eauto.
    rewrite <- Hsubst.
    eapply memory_update_sound; eauto.
    { by rewrite/= SetFacts.singleton_b is_compare_eqb eqb_refl. }
    { apply madd_wf; auto.
      { rewrite/DomConMem.loc_wf/SemCon.var_loc
               /local/DomConValue.llocal.
        by rewrite DomConMem.nth_fid_length.
      }
      rewrite <- Hsubst in Hwf_mmid.
      rewrite madd_wf; eauto.
      rewrite/DomConMem.loc_wf/SemCon.var_loc
             /local/DomConValue.llocal.
      by rewrite DomConMem.nth_fid_length.
    }
  Qed.

  Lemma sem_bind_ret_proper2 : forall
    (Hacc_prog: acc_correct_callfroms_prog prog acc)
    (Hacc_star: acc_correct_callfroms_star acc)
    (pidx:pidxT) (Hpidx_con: pidx_consistent prog (DomPidx.pifid pidx) pidx.2)
    (x : DomConValue.valueT) (y : valueT) (Hxy : x \abs y)
    (retto : option varT) (m m': DomConMem.memT) (w : memT) (Hzw : m \abs w)
    (Hsem : SemCon.sem_bind_ret pidx x retto m m'),
    m' \abs sem_bind_ret acc (DomPidx.pifid pidx) y retto w.
  Proof.
    move=> cf cs pidx Hpidx x y Hxy retto cm cm' am Hm [Hbind Hwf].
    move: retto Hbind=> [retto|] Hbind; last by rewrite <- Hbind.
    rewrite <- Hbind.
    eapply memory_update_sound; eauto; last by rewrite Hbind.
    rewrite/abs_relation/loc_abs.
    rewrite/smem/set_setlike.
    rewrite SetFacts.singleton_b is_compare_eqb.
    apply/eqbP.
    rewrite/SemCon.var_loc/var_loc/=.
    by move: retto Hbind=> [gvar|lvar]; reflexivity.
  Qed.

  Lemma mrestrict_proper2 :
    Proper2 (Tlocs ===> abs_relation ===> abs_relation)
            (mrestrict (K:=DomConValue.locT) (E:=option DomConValue.valueT)
                       (M:=DomConMem.memT))
            (mrestrict (K:=locT) (E:=valueT) (M:=memT)).
  Proof.
    move=> llocs rlocs Hlocs cmem amem.
    rewrite/abs_relation/memory_abs.
    case: amem => [amemelt|] //.
    case Harmem: (mrestrict rlocs (Elt amemelt)) => [armemelt|] //.
    refine (forall_implies _ _ _ _) => loc.
    rewrite/Tlocs in Hlocs; rewrite Hlocs; clear Hlocs llocs.
    move=> Hyp.
    rewrite (morphism_mrestrict_spec (t_locs rlocs) cmem loc).
    rewrite <- Harmem.
    rewrite mem_mrestrict_spec.
    rewrite t_locs_spec.
    by case Hin: (_ !\in _).
  Qed.
End MemoryAbs.


Section StateAbs.
  Global Program Instance state_abs : abs SemCon.stateT SemAccSen.stateT := {
    abs_relation conState absState :=
      conState.2 \abs (absState[conState.1]!:memT)
  }.
  Next Obligation.
    move=> [lcp lcm] [rcp rcm] [Hcp Hcm] las ras Has.
    rewrite/fst/snd in Hcp Hcm *.
    rewrite <- Hcm.
    rewrite <- Hcp.
    by rewrite -> (Has lcp).
  Qed.
End StateAbs.


Lemma step_correct (cState cState':SemCon.stateT)
      (aState:SemAccSen.stateT)
      (Hcf: acc_correct_callfroms_prog prog acc)
      (Hcs: acc_correct_callfroms_star acc)
      (Hconsistent : pidx_consistent prog (DomPidx.pifid cState.1) cState.1.2)
      (Hwf : DomConMem.mem_wf cState.1 cState.2)
      (Hstep : SemCon.sem_step prog cState cState')
      (Habs : cState \abs aState)
      (Hpostfix : postfix prog acc aState) :
  cState' \abs aState.
Proof.
  inversion Hstep; clear Hstep;
  rewrite/abs_relation/state_abs/fst/snd; subst;
  [ rewrite/piinstrs/picmd/pifunc//= in Hpidx |
    rewrite/pitmn/picmd/pifunc//= in Htmn |
    rewrite/pitmn/picmd/pifunc//= in Htmn |
    rewrite/pitmn/picmd/pifunc//= in Hcmd
  ];
  remember ((pfuncs prog) [fid]) as ffunc;
  [ case: ffunc Heqffunc Hpidx => [ffunc|] Heqffunc Hpidx // |
    case: ffunc Heqffunc Htmn => [ffunc|] Heqffunc Htmn // |
    case: ffunc Heqffunc Htmn => [ffunc|] Heqffunc Htmn // |
    case: ffunc Heqffunc Hcmd => [ffunc|] Heqffunc Hcmd //
  ];
  have Hfmapsto: (MapsTo fid ffunc (pfuncs prog)) by apply MF.find_mapsto_iff.

  { remember ((fcmds ffunc) [bid]) as cmd.
    case: cmd Heqcmd Hpidx => [[cinstrs tmn]|] Heqcmd Hpidx //.
    inversion Hpidx as [Hspidx]; clear Hpidx;
    rewrite/Syntax.cinstrs/fst in Hspidx; subst.
    have Hcmapsto: (MapsTo bid (instrs, tmn) (fcmds ffunc)) by apply MF.find_mapsto_iff.
    pose (Hpostfix fid ffunc Hfmapsto bid stack (instrs, tmn) Hcmapsto) as Hpostfixb.
    rewrite/= in Hconsistent.
    rewrite -> Hconsistent in Hpostfixb.
    case: Hpostfixb => [Hpostfixb Hpostfixt]; clear Hpostfixt.
    usual_tac Hstack.
    rewrite/abs_relation/state_abs/fst/snd in Habs.
    elim: {Heqcmd Hcmapsto} instrs mnew Hwf_mnew {Hwf} mold (aState[_]!) (aState[_]!) Hinstrs Habs Hpostfixb=> [|istr istrs IH] mnew Hwf_mnew mold mnew' mold' Hinstrs Habs Hpostfixb.
    { inv Hinstrs. inv Hpostfixb.
      rewrite <- Heq. move:Habs. apply abs_relation_proper.
      reflexivity. assumption.
    }
    { inv Hinstrs. inv Hpostfixb.
      exploit IH; eauto.
      assert (mmid \abs sem_instr acc (pifid (fid, bid, false, nstack)) istr mold').
      apply sem_instr_proper2 with mold; auto.
      move:H. apply abs_relation_proper. reflexivity. apply Hsem.
    }
  }

  { remember ((fcmds ffunc) [bid]) as cmd.
    case: cmd Heqcmd Htmn => [[cinstrs tmn]|] Heqcmd Htmn //.
    inversion Htmn as [Hstmn]; clear Htmn;
    rewrite/Syntax.cinstrs/fst in Hstmn; subst.
    have Hcmapsto: (MapsTo bid (cinstrs, tmn) (fcmds ffunc)) by apply MF.find_mapsto_iff.
    pose (Hpostfix fid ffunc Hfmapsto bid stack (cinstrs, tmn) Hcmapsto) as Hpostfixb.
    rewrite/= in Hconsistent.
    rewrite -> Hconsistent in Hpostfixb.
    case: Hpostfixb => [Hpostfixb Hpostfixt]; clear Hpostfixb.
    rewrite/ctmn//= in Hstmn; subst.
    rewrite <- (Hpostfixt nbid Hnbid).
    rewrite <- Heq.
    exact Habs.
  }

  { remember ((fcmds ffunc) [bid]) as cmd.
    case: cmd Heqcmd Htmn => [[cinstrs tmn]|] Heqcmd Htmn //.
    inversion Htmn as [Hstmn]; clear Htmn;
    rewrite/Syntax.cinstrs/fst in Hstmn; subst.
    have Hcmapsto: (MapsTo bid (cinstrs, tmn) (fcmds ffunc)) by apply MF.find_mapsto_iff.
    pose (Hpostfix fid ffunc Hfmapsto bid stack (cinstrs, tmn) Hcmapsto) as Hpostfixb.
    rewrite/= in Hconsistent.
    rewrite -> Hconsistent in Hpostfixb.
    case: Hpostfixb => [Hpostfixb Hpostfixt]; clear Hpostfixb.
    rewrite/ctmn//= in Hstmn; subst.
    rewrite/postfix_call in Hpostfixt.
    have Hnfmem := (Hpostfixt _ _ _ Hfunc); clear Hpostfixt.
    rewrite <- Hnfmem.
    exploit sem_bind_params_proper2; auto; last first.
    { instantiate (4 := (nfid, fentry nfunc, false, (fid, bid) :: stack)).
      move=> Habs'; exact Habs'.
    }
    eauto.
    eauto.
    exploit sem_exprs_proper2; last first.
    { move=> Hexprs'.
      instantiate (3 := (fid, bid, true, stack)) in Hexprs'.
      exact Hexprs'.
    }
    eauto.
    eauto.
    reflexivity.
    reflexivity.
    eauto.
    eauto.
    rewrite/pifid/pidxT_pidx; simpl. apply/andP. split; auto.
    rewrite <- Heqffunc. rewrite <- Heqcmd.
    done.
    done.
  }

  { remember ((fcmds ffunc) [bid]) as cmd.
    case: cmd Heqcmd Hcmd => [[cinstrs tmn]|] Heqcmd Hcmd //.
    inversion Hcmd as [Hscmd]; clear Hcmd;
    rewrite/Syntax.cinstrs/fst in Hscmd; subst.
    have Hcmapsto: (MapsTo bid (cinstrs, tmn) (fcmds ffunc)) by apply MF.find_mapsto_iff.
    pose (Hpostfix fid ffunc Hfmapsto bid ((pfid, pbid)::stack) (cinstrs, tmn) Hcmapsto) as Hpostfixb.
    rewrite/pifid/pibid/pidxT_pidx/fst/snd in Hconsistent.
    rewrite -> Hconsistent in Hpostfixb.
    case: Hpostfixb => [Hpostfixb Hpostfixt]; clear Hpostfixb.
    rewrite/ctmn//= in Hscmd; subst.
    rewrite/postfix_exit in Hpostfixt.
    remember ((fcmds pfunc) [pbid]) as pcmd.
    have Hpfunceq: ((pfuncs prog) [pfid] = Some pfunc) by apply MF.find_mapsto_iff.
    case: pcmd Heqpcmd => [pcmd |] Heqpcmd.     
    { have Hpcmd: (MapsTo pbid pcmd (fcmds pfunc)) by apply MF.find_mapsto_iff.
      rewrite/pitmn/picmd/pifunc//= in Hcall.
      rewrite Hpfunceq in Hcall.
      rewrite <- Heqpcmd in Hcall.
      inversion Hcall as [Hcallno].
      pose (Hpostfixt pfunc Hpfunc pcmd Hpcmd) as Hpostfixtt.
      rewrite -> Hcallno in Hpostfixtt.
      rewrite <- Hpostfixtt.
      exploit sem_bind_ret_proper2; try reflexivity; try eauto.
      rewrite/pifid/=.
      rewrite simpl_pidx_consistent in Hconsistent.
      by move/andP: Hconsistent=> [_ Hc].
      exploit sem_expr_proper2; try reflexivity; try eauto; last first.
      exact: Habs.

      move: {Hpostfixtt Hret Hbind} Habs.
      rewrite {1}/abs_relation/state_abs/fst/snd.
      rewrite/fst/snd in Hwf.
      clear -Hcf Hcs Hwf Hconsistent=> Habs.
      move: (_[_]!) Habs=> [mem|] Habs; last by case: (not_recursive _ _).
      rewrite {2}/mrestrict/memT_restricting_maplike/withtop_restricting_maplike.
      rewrite if_Elt=> loc.
      move: (Habs loc)=> {Habs} Habs.
      rewrite {2}/mlookup/memT_maplike/withtop_maplike.
      dom_tac.
      case Hin: (_ !\in _); last done.
      case Hrec: (not_recursive _ _); last done.
      dom_tac.
      case Hin': (_ !\in _); first done.
      move: (Hwf loc)=> {Hwf}.
      move: (_[_]!)=> [val|]; last done.
      move: {Habs} loc Hin Hin'.
      move=> [[[[[a u] o]|gvar]|[[fvar vvar] dvar]] fs]; try done.
      move=> Hle Heq Hwf.
      rewrite/= in Hle Heq Hwf.
      move/Bool.negb_false_iff in Heq.
      move/eqbP in Heq; usual_tac Heq.
      move/is_true_iff in Hle.
      move/leb_iff in Hle.
      exploit Hwf; [done|move=> {Hwf} Hwf].
      usual_tac Hwf.
      exploit acc_correct_consistent; eauto.
      move=> Hfind; elimtype False.
      clear -Hle Hwf Hfind.
      exploit DomConMem.nth_fid_none; eauto.
      move=> [Heq _]; subst.
      clear -Hle.
      rewrite/= in Hle.
      omega.
    }

    { rewrite/pitmn/picmd/pifunc//= in Hcall.
      rewrite Hpfunceq in Hcall.
      rewrite <- Heqpcmd in Hcall.
      inversion Hcall.
    }
  }
Qed.

Lemma sem_wf n conState (Hsem: SemCon.sem prog n conState) :
  DomConMem.mem_wf conState.1 conState.2.
Proof.
  elim: Hsem.
  { move=> pidx pidx' mem Hpidx' Hpidx Hmem Hmem_wf.
    rewrite/fst/snd.
    by rewrite Hmem.
  }
  { auto. }
Qed.
  
Lemma correct (n:nat) (conState:SemCon.stateT)
      (absState:SemAccSen.stateT)
      (Hcf: acc_correct_callfroms_prog prog acc)
      (Hcs: acc_correct_callfroms_star acc)
      (Hsem: SemCon.sem prog n conState)
      (Hpostfix: SemSen.postfix prog acc absState) :
  conState \abs absState.
Proof.
  elim: Hsem.
  { move=> pidx pidxi mem Hinitial Hpidx Hmem.
    rewrite/abs_relation/state_abs/fst/snd.
    rewrite Hmem.
    rewrite/abs_relation/memory_abs//=.
    by case: (proj1_sig absState pidx).
  }
  { move=> step midstate newstate Hsem Habs Hstep Hwf Hwf_mem.
    eapply step_correct; auto.
    by eapply sem_consistent; eauto.
    by eapply sem_wf; eauto.
    done.
    done.
  }
Qed.

End Correctness.
