Require Import ssreflect.
Require Import ssrfun.
Require Import ssrbool.

Require Import Tactics.
Require Import Classes.Equivalence.
Require Import List.


Ltac inv H := inversion H; clear H; subst.

Ltac exploit x :=
    refine ((fun x y => y x) (x _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _ _ _ _ _ _ _ _ _ _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _ _ _ _ _ _ _ _ _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _ _ _ _ _ _ _ _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _ _ _ _ _ _ _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _ _ _ _ _ _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _ _ _ _ _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _ _ _ _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _ _ _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _ _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _ _) _)
 || refine ((fun x y => y x) (x _ _ _) _)
 || refine ((fun x y => y x) (x _ _) _)
 || refine ((fun x y => y x) (x _) _).


Module Simpless.
  Ltac program_simplify :=
    intros ; Tactics.destruct_all_rec_calls ; repeat (Tactics.destruct_conjs; simpl proj1_sig in * );
    subst*; Tactics.autoinjections ; try Tactics.discriminates ;
    try (solve [ red ; intros ; Tactics.destruct_conjs ; Tactics.autoinjections ; Tactics.discriminates ]).

  Ltac program_simpl := program_simplify ; try typeclasses eauto with program ; try Tactics.program_solve_wf.

  Ltac tactic := program_simpl; try done.
End Simpless.

Global Ltac obligation_tactic := Tactics.program_simpl; try done.


Section Props.
  Lemma is_true_iff : forall (b:bool), (b = true) <-> b.
  Proof. by case. Qed.

  Lemma forall_implies
        A (P Q : A -> Prop) (Himplies : forall a : A, P a -> Q a) :
    (forall a : A, P a) -> (forall a : A, Q a).
  Proof. by move=> Hcond a; apply/Himplies/Hcond. Qed.

  Lemma forall_iff
        A (P Q : A -> Prop) (Hiff : forall a : A, P a <-> Q a) :
    (forall a : A, P a) <-> (forall a : A, Q a).
  Proof. by split=> Hcond a; apply/Hiff/Hcond. Qed.

  Lemma implies_right_turn P Q R :
    P -> (Q -> R) -> ((P -> Q) -> R).
  Proof.
    move=> p QR PQ.
    apply: QR.
    apply: PQ.
    exact: p.
  Qed.
End Props.


Section Lists.
  Fixpoint list_forall {A} (test: A -> bool) (l : list A) :=
    if l is (a :: l')
    then andb (test a) (list_forall test l')
    else true.

  Lemma list_forallP {A} (test: A -> bool) (l : list A) :
    reflect
      (forall i (Hi : List.In i l), test i)
      (list_forall test l).
  Proof.
    elim: l=> [|a l' IH]; first by constructor.
    rewrite/list_forall -/list_forall.
    case Ha: (test a).
    { apply/equivP; first by apply IH.
      split; refine (forall_implies _ _ _ _) => b Hyp Hin.
      case: Hin; last exact: Hyp.
      by move=> Hab; subst.
      by apply Hyp; right.
    }
    { constructor=> Hc.
      rewrite Hc in Ha; first done.
      by left.
    }
  Qed.

  Fixpoint nth' {A} (n:nat) (l:list A) {struct l} : option A :=
    if l is a::l'
    then if n is S n'
         then nth' n' l'
         else Some a
    else None.

  Lemma nth'_length {A} (l:list A) :
    nth' (length l) l = None.
  Proof. by elim: l=> [|a l IH]. Qed.

  Lemma find_none_rev A test (l:list A) (Hnone : List.find test l = None) :
    List.find test (rev l) = None.
  Proof.
    elim: l Hnone=> [|a l IH] Hnone; first done.
    rewrite/= in Hnone *.
    case Ha: (test a) in Hnone; first done.
    exploit IH; auto.
    elim: {l IH Hnone} (rev l)=> [|a' l' IH] /=; first by rewrite Ha.
    by case: (test a').
  Qed.

  Lemma existsb_rev A (f: A -> bool) (l: list A) :
    existsb f l = existsb f (rev l).
  Proof.
    elim: l=> [|a l IH]; first done.
    rewrite/= existsb_app IH.
    rewrite [existsb _ (_::nil)]/existsb /orb.
    by case: (f a); case: (existsb _ _).
  Qed.
End Lists.


Section Options.
  Definition liftM {A B} (f : A -> B) :=
    [fun oa : option A =>
       if oa is Some a
       then Some (f a)
       else None].
End Options.
